IMPLEMENTATION MODULE M2Constants ;

FROM M2ALU IMPORT Equ ;
FROM SymbolTable IMPORT MakeConstLit, PushValue, IsConst, GetType, IsValueSolved ;
FROM NameKey IMPORT MakeKey ;


VAR
   ZeroCard, ZeroReal,
   OneCard, OneReal,
   TwoCard, TwoReal  : CARDINAL ;


(*
   IsSame - returns TRUE if the symbols, s1, and, s2, are the same or
            have the same constant value and same type.
*)

PROCEDURE IsSame (s1, s2: CARDINAL) : BOOLEAN ;
BEGIN
   IF s1=s2
   THEN
      RETURN( TRUE )   (* easy decision.. *)
   ELSE
      IF IsConst(s1) AND IsConst(s2) AND (GetType(s1)=GetType(s2)) AND
         IsValueSolved(s1) AND IsValueSolved(s2)
      THEN
         PushValue(s1) ;
         PushValue(s2) ;
         RETURN( Equ() )
      ELSE
         RETURN( FALSE )    (* not constant or a different type therefore we do not know *)
      END
   END
END IsSame ;


(*
   IsZero - returns TRUE if symbol, s, represents the value zero.
*)

PROCEDURE IsZero (s: CARDINAL) : BOOLEAN ;
BEGIN
   IF (s=ZeroCard) OR (s=ZeroReal)
   THEN
      RETURN( TRUE )
   ELSIF IsConst(s)
   THEN
      PushValue(s) ;
      PushValue(ZeroCard) ;
      RETURN( Equ() )
   ELSE
      RETURN( FALSE )
   END
END IsZero ;


(*
   IsOne - returns TRUE if symbol, s, represents the value one.
*)

PROCEDURE IsOne (s: CARDINAL) : BOOLEAN ;
BEGIN
   IF (s=OneCard) OR (s=OneReal)
   THEN
      RETURN( TRUE )
   ELSIF IsConst(s)
   THEN
      PushValue(s) ;
      PushValue(OneCard) ;
      RETURN( Equ() )
   ELSE
      RETURN( FALSE )
   END
END IsOne ;


(*
   IsTwo - returns TRUE if symbol, s, represents the value two.
*)

PROCEDURE IsTwo (s: CARDINAL) : BOOLEAN ;
BEGIN
   IF (s=TwoCard) OR (s=TwoReal)
   THEN
      RETURN( TRUE )
   ELSIF IsConst(s)
   THEN
      PushValue(s) ;
      PushValue(TwoCard) ;
      RETURN( Equ() )
   ELSE
      RETURN( FALSE )
   END
END IsTwo ;


(*
   Init - initialize the constants.
*)

PROCEDURE Init ;
BEGIN
   ZeroCard := MakeConstLit(MakeKey('0')) ;
   OneCard  := MakeConstLit(MakeKey('1')) ;
   TwoCard  := MakeConstLit(MakeKey('2')) ;
   ZeroReal := MakeConstLit(MakeKey('0.0')) ;
   OneReal  := MakeConstLit(MakeKey('1.0')) ;
   TwoReal  := MakeConstLit(MakeKey('2.0'))
END Init ;


BEGIN
   Init
END M2Constants.
