IMPLEMENTATION MODULE M2Gen487 ;


FROM ASCII IMPORT nul ;
FROM M2Math IMPORT pi, sin, cos, tan, atan, sqrt ;
FROM M2Base IMPORT LongInt, Integer, Cardinal, LongReal, Real, Trunc ;
FROM M2System IMPORT Address ;
FROM NameKey IMPORT MakeKey, GetKey ;
FROM M2Error IMPORT BeginError, EndError ;
FROM M2Lexical IMPORT WriteError, NearToken, InternalError ;
FROM M2ALU IMPORT PopInt, PopCard, PopLongReal, PushInt, CoerseLongRealToCard, ConvertRealToInt, LessEqu ;
FROM M2GDB IMPORT LineInfo ;
FROM M2Quads IMPORT GetLastQuadNo, QuadToTokenNo ;
FROM NumberIO IMPORT WriteInt, WriteCard, IntToStr ;
FROM FpuIO IMPORT WriteLongReal, LongRealToStr ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StdIO IMPORT Write ;
FROM M2AsmUtil IMPORT WriteAsmName ;
FROM NameKey IMPORT WriteKey ;
FROM M2Options IMPORT MathCoprocessor ;
FROM M2Constants IMPORT IsZero, IsOne ;

FROM SymbolTable IMPORT NulSym, MakeConstLit, GetType, GetSymName,
                        IsProcedure, GetVarFather, IsConstLit, PutConst,
                        IsConst, PushValue, PushOffset, GetMode, ModeOfAddr ;

FROM Lists IMPORT List,
                  InitList, IncludeItemIntoList, ForeachItemInListDo ;


CONST
   MaxRunTimeStack =    8 ;
   MaxInstruction  = 1000 ;

TYPE
   Instruction     = (fadd, fsub, fstore, fpop, fpush, fdiv, fmul, fdup,
                      fchange, fabs, fmod, fcompare,
                      fround, fsin, fcos, ftan, fatan, fsqrt,
                      fnone) ;

   InstructionInfo = RECORD
                        Inst   : Instruction ;
                        Operand: CARDINAL ;   (* Symbol number       *)
                        LineNo : CARDINAL ;   (* Current source line *)
                     END ;

VAR
   NoOfInstructions: CARDINAL ;   (* Number of instructions in list *)
   Instructions    : ARRAY [1..MaxInstruction] OF InstructionInfo ;
   Constants       : List ;
   RunTimeStack    : CARDINAL ;
   CurrentLine     : CARDINAL ;
   
(* %%%FORWARD%%%
PROCEDURE Pusho (o: CARDINAL) ; FORWARD ;
PROCEDURE add ; FORWARD ;
PROCEDURE sub ; FORWARD ;
PROCEDURE div (t: CARDINAL) ; FORWARD ;
PROCEDURE mult ; FORWARD ;
PROCEDURE mod (t: CARDINAL) ; FORWARD ;
PROCEDURE neg ; FORWARD ;
PROCEDURE compare ; FORWARD ;
PROCEDURE push (o: CARDINAL) ; FORWARD ;
PROCEDURE pop (o: CARDINAL) ; FORWARD ;
PROCEDURE InstructionOperand (i: ARRAY OF CHAR; o: CARDINAL) ; FORWARD ;
PROCEDURE AddInstruction (f: Instruction; o: CARDINAL) ; FORWARD ;
PROCEDURE Popo (o: CARDINAL) ; FORWARD ;
PROCEDURE LastOperandWasZero () : BOOLEAN ; FORWARD ;
PROCEDURE LastOperandWasOne () : BOOLEAN ; FORWARD ;
PROCEDURE LastInstructionWas (f: Instruction) : BOOLEAN ; FORWARD ;
PROCEDURE LastOperandWas (o: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE RemoveLastInstruction () : CARDINAL ; FORWARD ;
PROCEDURE LastButOneInstructionWas (f: Instruction) : BOOLEAN ; FORWARD ;
PROCEDURE AlterLastInstructionTo (f: Instruction) ; FORWARD ;
PROCEDURE IncRunTimeStack ; FORWARD ;
PROCEDURE DecRunTimeStack ; FORWARD ;
PROCEDURE WriteInstruction (f: Instruction; o: CARDINAL) ; FORWARD ;
PROCEDURE CodeStore (o: CARDINAL) ; FORWARD ;
PROCEDURE CodePop (o: CARDINAL) ; FORWARD ;
PROCEDURE CodePush (o: CARDINAL) ; FORWARD ;
PROCEDURE WriteOperand (o: CARDINAL) ; FORWARD ;
PROCEDURE WriteConstantName (o: CARDINAL) ; FORWARD ;
PROCEDURE WriteConstantAndValue (Sym: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   IsCoprocessorUsed - returns TRUE if one of these operands requires
                       use of the coprocessor. If a variable has a GetMode
                       of ValueAddr then it does not qualify as a coprocessor
                       variable as it is computing addresses.
*)

PROCEDURE IsCoprocessorUsed (Op1, Op2, Op3: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (IsCoprocessorType(Op1) AND (GetMode(Op1)#ValueAddr)) OR
          (IsCoprocessorType(Op2) AND (GetMode(Op2)#ValueAddr)) OR
          (IsCoprocessorType(Op3) AND (GetMode(Op3)#ValueAddr))
         )
END IsCoprocessorUsed ;


(*
   IsCoprocessorType - returns TRUE if the operand has a coprocessor supported type.
*)

PROCEDURE IsCoprocessorType (Op: CARDINAL) : BOOLEAN ;
VAR
   t: CARDINAL ;
BEGIN
   t := GetType(Op) ;
   RETURN( (t=LongInt) OR (t=Real) OR (t=LongReal) )
END IsCoprocessorType ;


(*
   CoprocessorLineInfo - places the line information into the coprocessor
                         instruction buffer.
*)

PROCEDURE CoprocessorLineInfo (l: CARDINAL) ;
BEGIN
   CurrentLine := l
END CoprocessorLineInfo ;


(*
   CoprocessorMath - codes the inbuilt mathematical operators.
                     Op2 := Op1(Op3)

                     where:

                     Op1    function
                     Op2    return variable
                     Op3    parameter
*)

PROCEDURE CoprocessorMath (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   push(Op3) ;
   IF Op1=Trunc
   THEN
      FlushCoprocessor ;
      WriteString('    fnstcw  CQfpu1  # TRUNC') ; WriteLn ;
      WriteString('    movl    CQfpu1,%eax') ; WriteLn ;
      WriteString('    movb    $12,%ah') ; WriteLn ;
      WriteString('    movl    %eax,CQfpu2') ; WriteLn ;
      WriteString('    fldcw   CQfpu2') ; WriteLn ;
      pop(Op2) ;
      FlushCoprocessor ;
      WriteString('    fldcw   CQfpu1') ; WriteLn
   ELSIF Op1=sin
   THEN
      AddInstruction(fsin, 0) ;
      pop(Op2)
   ELSIF Op1=cos
   THEN
      AddInstruction(fcos, 0) ;
      pop(Op2)
   ELSIF Op1=tan
   THEN
      AddInstruction(ftan, 0) ;
      pop(Op2)
   ELSIF Op1=atan
   THEN
      AddInstruction(fatan, 0) ;
      pop(Op2)
   ELSIF Op1=sqrt
   THEN
      AddInstruction(fsqrt, 0) ;
      pop(Op2)
   ELSE
      InternalError('unknown math operator', __FILE__, __LINE__)
   END
END CoprocessorMath ;


(*
   CoprocessorAdd - generates coprocessor code for:
                    Op1 := Op2 + Op3
*)

PROCEDURE CoprocessorAdd (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   Pusho(Op3) ;
   Pusho(Op2) ;
   add ;
   Popo(Op1)
END CoprocessorAdd ;


(*
   CoprocessorSub - generates coprocessor code for:
                    Op1 := Op2 - Op3
*)

PROCEDURE CoprocessorSub (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   Pusho(Op3) ;
   Pusho(Op2) ;
   sub ;
   Popo(Op1)
END CoprocessorSub ;


(*
   CoprocessorMult - generates coprocessor code for:
                     Op1 := Op2 * Op3
*)

PROCEDURE CoprocessorMult (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   Pusho(Op3) ;
   Pusho(Op2) ;
   mult ;
   Popo(Op1)
END CoprocessorMult ;


(*
   CoprocessorDiv - generates coprocessor code for:
                    Op1 := Op2 DIV Op3
*)

PROCEDURE CoprocessorDiv (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   Pusho(Op3) ;
   Pusho(Op2) ;
   div(GetType(Op1)) ;
   Popo(Op1)
END CoprocessorDiv ;


(*
   CoprocessorMod - generates coprocessor code for:
                    Op1 := Op2 MOD Op3
*)

PROCEDURE CoprocessorMod (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   Pusho(Op3) ;
   Pusho(Op2) ;
   mod(GetType(Op1)) ;
   Popo(Op1)
END CoprocessorMod ;


(*
   CoprocessorNeg - generates coprocessor code for:
                    Op1 := - Op2
*)

PROCEDURE CoprocessorNeg (Op1, Op2: CARDINAL) ;
BEGIN
   Pusho(Op2) ;
   neg ;
   Popo(Op1)
END CoprocessorNeg ;


(*
   CoprocessorBecomes - performs Op1 := Op2 for any coprocessor supported type.
*)

PROCEDURE CoprocessorBecomes (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op2) ;
   pop(Op1)
END CoprocessorBecomes ;


(*
   CoprocessorCopy - performs Op1 copy Op2 for any coprocessor supported type.
                     This is required because the CSE will map redundant :=
                     and arithmetic expressions into Op1 copy Op2 (particularly
                     when Op2 is a constant.

                     Here are the copy rules:

------------------------------------------------------------------------------
   Copy Operator
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]

*)

PROCEDURE CoprocessorCopy (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op2) ;
   IF GetMode(Op1)=ValueAddr
   THEN
      FlushCoprocessor ;
      InternalError('should not be using the coprocessor to perform address arithmetic - check common subexpression elimiation',
                    __FILE__, __LINE__)
   ELSE
      pop(Op1)
   END
END CoprocessorCopy ;


(*
   CoprocessorConvert - Converts Operand3 to type Operand2 placing the result into
                        Operand1.
                        Convert will, if need be, alter the machine representation
                        of Operand3 to comply with TYPE Operand2.

                        Assumes that we never have to convert a coprocessor type
                        into a type whose size is < TSIZE(WORD).
                        So converting a CHAR into a REAL is done in two steps:

                        (i)  CHAR -> INTEGER
                        (ii) INTEGER -> REAL

                        So this routine only converts:
                        TSIZE(INTEGER) types <-> Coprocessor type
*)

PROCEDURE CoprocessorConvert (Op1, Op2, Op3: CARDINAL) ;
BEGIN
   push(Op3) ;
   FlushCoprocessor ;
   IF Op2=LongInt
   THEN
      InstructionOperand('fistpq', Op1)
   ELSIF Op2=LongReal
   THEN
      InstructionOperand('fstpl', Op1)
   ELSIF Op2=Real
   THEN
      InstructionOperand('fstps', Op1)
   ELSE
      InstructionOperand('fistpl', Op1)
   END ;
   DecRunTimeStack
END CoprocessorConvert ;


(*
   CoprocessorIfEqu - generates code for the = operator
*)

PROCEDURE CoprocessorIfEqu (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op1) ;
   push(Op2) ;
   (* lets see whether any optimization occurred ? *)
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      (*
         no optimization - therefore lets try once more this time
         with the operands in a different order
      *)
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op2 := RemoveLastInstruction() ;
      Op1 := RemoveLastInstruction() ;
      push(Op2) ;
      push(Op1)
   END ;
   compare ;
   WriteString('   andb $69,%ah') ; WriteLn ;
   WriteString('   cmpb $64,%ah') ; WriteLn
END CoprocessorIfEqu ;


(*
   CoprocessorIfNotEqu - generates code for the # operator
*)

PROCEDURE CoprocessorIfNotEqu (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op1) ;
   push(Op2) ;
   (* lets see whether any optimization occurred ? *)
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      (*
         no optimization - therefore lets try once more this time
         with the operands in a different order
      *)
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op2 := RemoveLastInstruction() ;
      Op1 := RemoveLastInstruction() ;
      push(Op2) ;
      push(Op1)
   END ;
   compare ;
   WriteString('   andb $68,%ah') ; WriteLn ;
   WriteString('   xorb $64,%ah') ; WriteLn
END CoprocessorIfNotEqu ;


(*
   CoprocessorIfLess - generates code for the < operator
*)

PROCEDURE CoprocessorIfLess (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op1) ;
   push(Op2) ;
   compare ;
   WriteString('   andb $69,%ah') ; WriteLn ;
   WriteString('   decb %ah') ; WriteLn ;
   WriteString('   cmpb $64,%ah') ; WriteLn
END CoprocessorIfLess ;


(*
   CoprocessorIfLessEqu - generates code for the <= operator
*)

PROCEDURE CoprocessorIfLessEqu (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op1) ;
   push(Op2) ;
   compare ;
   WriteString('    andb $69,%ah') ; WriteLn ;
   WriteString('    cmpb $1,%ah') ; WriteLn
END CoprocessorIfLessEqu ;


(*
   CoprocessorIfGre - generates code for the > operator
*)

PROCEDURE CoprocessorIfGre (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op1) ;
   push(Op2) ;
   compare ;
   WriteString('    andb $5,%ah') ; WriteLn
END CoprocessorIfGre ;


(*
   CoprocessorIfGreEqu - generates code for the >= operator
*)

PROCEDURE CoprocessorIfGreEqu (Op1, Op2: CARDINAL) ;
BEGIN
   push(Op1) ;
   push(Op2) ;
   compare ;
   WriteString('    andb $69,%ah') ; WriteLn
END CoprocessorIfGreEqu ;


(*
   CoprocessorParam - pushes constant, Sym, of type, Type onto the stack.
*)

PROCEDURE CoprocessorParam (Sym: CARDINAL; Type: CARDINAL) ;
BEGIN
   IF Sym=pi
   THEN
      IF Type=LongReal
      THEN
         WriteString('    pushl   $0x400921FB  #   pi  2nd word') ; WriteLn ;
         WriteString('    pushl   $0x54442D18  #   pi  1st word') ; WriteLn
      ELSIF Type=Real
      THEN
         WriteString('    pushl   $0x40490FDB  #   pi') ; WriteLn
      ELSE
         WriteError('cannot pass pi into this paramater')
      END
   ELSIF Type=LongInt
   THEN
      WriteString('    pushl  $0') ; WriteLn ;
      PushValue(Sym) ;
      WriteString('    pushl  $') ; WriteInt(PopInt(), 0) ; WriteLn
   ELSIF Type=Real
   THEN
      PushValue(Sym) ;
      CoerseLongRealToCard ;
      WriteString('    pushl  $') ; WriteCard(PopCard(), 0) ; WriteLn
   ELSIF Type=LongReal
   THEN
      WriteString('    pushl  4+') ; WriteOperand(Sym) ; WriteLn ;
      WriteString('    pushl  ') ; WriteOperand(Sym) ; WriteLn
   ELSE
      WriteError('coprocessor constant type currently unknown in CoprocessorParam')
   END
END CoprocessorParam ;


(*
   CoprocessorFunctValue - retrieves the function value from the coprocessor stack.
*)

PROCEDURE CoprocessorFunctValue (Sym: CARDINAL) ;
BEGIN
   IncRunTimeStack ;
   pop(Sym)
END CoprocessorFunctValue ;


(*
   CoprocessorReturnValue - pushes the symbol, Sym, onto the
                            coprocessor stack before the function returns.
*)

PROCEDURE CoprocessorReturnValue (Sym: CARDINAL; Proc: CARDINAL) ;
BEGIN
   push(Sym) ;
   DecRunTimeStack
END CoprocessorReturnValue ;


(*
   Pusho - pushes a symbol onto the coprocessor stack. It checks to
           see that the symbol is not a value.
*)

PROCEDURE Pusho (o: CARDINAL) ;
BEGIN
   IF GetMode(o)=ValueAddr
   THEN
      BeginError ;
      WriteString('processing quad') ; WriteCard(GetLastQuadNo(), 4) ; WriteLn ;
      InternalError('why are we using coprocessor for address calculation?', __FILE__, __LINE__) ;
      EndError
   END ;
   push(o)
END Pusho ;


(*
   push - push operand, o.
*)

PROCEDURE push (o: CARDINAL) ;
BEGIN
   IF LastInstructionWas(fpop) AND
      LastOperandWas(o)
   THEN
      IF GetType(o)=LongInt
      THEN
         (* Intel bodged this - no fstore for LONGINT :-( *)
         (*
            so we:

            dup
            pop

            instead of

            pop
            push
         *)
         AlterLastInstructionTo(fdup) ;
         IncRunTimeStack ;
         AddInstruction(fpop, o) ;
         DecRunTimeStack
      ELSE
         AlterLastInstructionTo(fstore)
      END
   ELSE
      AddInstruction(fpush, o)
   END ;
   IncRunTimeStack
END push ;


(*
   Popo - pops a symbol from the coprocessor stack and checks
          to see whether it has a value mode.
*)

PROCEDURE Popo (o: CARDINAL) ;
BEGIN
   IF GetMode(o)=ValueAddr
   THEN
      BeginError ;
      WriteString('processing quad') ; WriteCard(GetLastQuadNo(), 4) ; WriteLn ;
      InternalError('why are we using the coprocessor to calculate addresses?', __FILE__, __LINE__) ;
      EndError
   END ;
   pop(o)
END Popo ;


(*
   pop - pops stack into operand, o.
*)

PROCEDURE pop (o: CARDINAL) ;
BEGIN
   IF LastInstructionWas(fpush) AND
      LastOperandWas(o)
   THEN
      o := RemoveLastInstruction()
   ELSE
      AddInstruction(fpop, o)
   END ;
   DecRunTimeStack
END pop ;


(*
   add - adds the top two elements on the stack
*)

PROCEDURE add ;
VAR
   Op1, Op2: CARDINAL ;
BEGIN
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op1 := RemoveLastInstruction() ;
      Op2 := RemoveLastInstruction() ;
      IF IsZero(Op1)
      THEN
         Pusho(Op2)
      ELSIF IsZero(Op2)
      THEN
         Pusho(Op1)
      ELSE
         (* reverse order of Op1, Op2 *)
         Pusho(Op1) ;
         Pusho(Op2) ;
         AddInstruction(fadd, 0) ;
         DecRunTimeStack
      END
   ELSE
      IF LastInstructionWas(fpush) AND
         LastOperandWasZero()
      THEN
         Op1 := RemoveLastInstruction()
      ELSE
         AddInstruction(fadd, 0)
      END ;
      DecRunTimeStack
   END
END add ;


(*
   sub - subtracts the top two elements on the stack
*)

PROCEDURE sub ;
VAR
   Op1, Op2: CARDINAL ;
BEGIN
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op1 := RemoveLastInstruction() ;
      Op2 := RemoveLastInstruction() ;
      IF IsZero(Op1)
      THEN
         Pusho(Op2) ;
         AddInstruction(fchange, 0)
      ELSIF IsZero(Op2)
      THEN
         Pusho(Op1)
      ELSE
         (* preserve order of Op1, Op2 *)
         Pusho(Op2) ;
         Pusho(Op1) ;
         AddInstruction(fsub, 0) ;
         DecRunTimeStack
      END
   ELSE
      IF LastInstructionWas(fpush) AND
         LastOperandWasZero()
      THEN
         Op1 := RemoveLastInstruction() ;
         AddInstruction(fchange, 0)
      ELSE
         AddInstruction(fsub, 0)
      END ;
      DecRunTimeStack
   END
END sub ;


(*
   mult - multiplies the top two elements on the stack
*)

PROCEDURE mult ;
VAR
   Op1, Op2: CARDINAL ;
BEGIN
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op1 := RemoveLastInstruction() ;
      Op2 := RemoveLastInstruction() ;
      IF IsZero(Op1)
      THEN
         Pusho(Op1)
      ELSIF IsZero(Op2)
      THEN
         Pusho(Op2)
      ELSIF IsOne(Op1)
      THEN
         Pusho(Op2)
      ELSIF IsOne(Op2)
      THEN
         Pusho(Op1)
      ELSE
         (* swap order of Op1, Op2 *)
         Pusho(Op1) ;
         Pusho(Op2) ;
         AddInstruction(fmul, 0) ;
         DecRunTimeStack
      END
   ELSE
      IF LastInstructionWas(fpush) AND
         LastOperandWasOne()
      THEN
         Op1 := RemoveLastInstruction()
      ELSE
         AddInstruction(fmul, 0)
      END ;
      DecRunTimeStack
   END
END mult ;


(*
   div - divides the top two elements on the stack.
         The coprocessor type is passed in, t.
*)

PROCEDURE div (t: CARDINAL) ;
VAR
   Op1, Op2: CARDINAL ;
BEGIN
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op1 := RemoveLastInstruction() ;
      Op2 := RemoveLastInstruction() ;
      IF IsZero(Op2)
      THEN
         Pusho(Op2)
      ELSIF IsOne(Op1)
      THEN
         Pusho(Op2)
      ELSE
         (* preserve order of Op1, Op2 *)
         Pusho(Op2) ;
         Pusho(Op1) ;
         AddInstruction(fdiv, 0) ;
         IF t=LongInt
         THEN
            AddInstruction(fround, 0)
         END ;
         DecRunTimeStack
      END
   ELSE
      IF LastInstructionWas(fpush) AND
         LastOperandWasOne()
      THEN
         Op1 := RemoveLastInstruction()
      ELSE
         AddInstruction(fdiv, 0) ;
         IF t=LongInt
         THEN
            AddInstruction(fround, 0)
         END
      END ;
      DecRunTimeStack
   END
END div ;


(*
   mod - calulates the modulus of the two elements on the stack.
         The coprocessor type is passed in, t.

         Note that fprem replaces the top of stack with the result.
*)

PROCEDURE mod (t: CARDINAL) ;
VAR
   Op1, Op2: CARDINAL ;
BEGIN
   IF LastInstructionWas(fpush) AND
      LastButOneInstructionWas(fpush)
   THEN
      DecRunTimeStack ;
      DecRunTimeStack ;
      Op1 := RemoveLastInstruction() ;
      Op2 := RemoveLastInstruction() ;
      IF IsZero(Op2)
      THEN
         Pusho(Op2)
      ELSE
         (* preserve order of Op1, Op2 *)
         Pusho(Op2) ;
         Pusho(Op1) ;
         AddInstruction(fmod, 0) ;
         IF t=LongInt
         THEN
            AddInstruction(fround, 0)
         END ;
         DecRunTimeStack
      END
   ELSE
      AddInstruction(fmod, 0) ;
      IF t=LongInt
      THEN
         AddInstruction(fround, 0)
      END ;
      DecRunTimeStack
   END
END mod ;


(*
   neg - negates the top element on the stack
*)

PROCEDURE neg ;
VAR
   Op1: CARDINAL ;
BEGIN
   IF LastInstructionWas(fchange)
   THEN
      Op1 := RemoveLastInstruction()
   ELSE
      AddInstruction(fchange, 0)
   END
END neg ;


(*
   compare - 
*)

PROCEDURE compare ;
BEGIN
   AddInstruction(fcompare, 0) ;
   DecRunTimeStack ;
   DecRunTimeStack ;
   FlushCoprocessor
END compare ;


(*
   AddInstruction - adds an instruction to the list
*)

PROCEDURE AddInstruction (f: Instruction; o: CARDINAL) ;
BEGIN
   INC(NoOfInstructions) ;
   IF NoOfInstructions=MaxInstruction
   THEN
      WriteError('too many fpu instructions in basic block - increase MaxInstruction in M2Gen487.mod')
   END ;
   WITH Instructions[NoOfInstructions] DO
      Inst := f ;
      Operand := o ;
      LineNo := CurrentLine
   END
END AddInstruction ;


(*
   LastOperandWasZero - returns TRUE if the last instruction operand was 0
*)

PROCEDURE LastOperandWasZero () : BOOLEAN ;
BEGIN
   IF NoOfInstructions>0
   THEN
      RETURN( IsZero(Instructions[NoOfInstructions].Operand) )
   ELSE
      RETURN( FALSE )
   END
END LastOperandWasZero ;


(*
   LastOperandWasOne - returns TRUE if the last instruction operand was 1
*)

PROCEDURE LastOperandWasOne () : BOOLEAN ;
BEGIN
   IF NoOfInstructions>0
   THEN
      RETURN( IsOne(Instructions[NoOfInstructions].Operand) )
   ELSE
      RETURN( FALSE )
   END
END LastOperandWasOne ;


(*
   LastInstructionWas - returns TRUE if the last instruction was, f.
*)

PROCEDURE LastInstructionWas (f: Instruction) : BOOLEAN ;
BEGIN
   IF NoOfInstructions>0
   THEN
      RETURN( Instructions[NoOfInstructions].Inst=f )
   ELSE
      RETURN( FALSE )
   END
END LastInstructionWas ;


(*
   LastOperandWas - returns TRUE if the last operand was, o.
*)

PROCEDURE LastOperandWas (o: CARDINAL) : BOOLEAN ;
BEGIN
   IF NoOfInstructions>0
   THEN
      RETURN( Instructions[NoOfInstructions].Operand=o )
   ELSE
      RETURN( FALSE )
   END
END LastOperandWas ;


(*
   RemoveLastInstruction - removes the last instruction and returns the operand.
*)

PROCEDURE RemoveLastInstruction () : CARDINAL ;
BEGIN
   IF NoOfInstructions>0
   THEN
      DEC(NoOfInstructions) ;
      RETURN( Instructions[NoOfInstructions+1].Operand )
   ELSE
      RETURN( NulSym )
   END
END RemoveLastInstruction ;


(*
   LastButOneInstructionWas - returns TRUE if the last but one instruction was, f.
*)

PROCEDURE LastButOneInstructionWas (f: Instruction) : BOOLEAN ;
BEGIN
   IF NoOfInstructions>1
   THEN
      RETURN( Instructions[NoOfInstructions-1].Inst=f )
   ELSE
      RETURN( FALSE )
   END
END LastButOneInstructionWas ;


(*
   AlterLastInstructionTo - alters the last instruction to f.
*)

PROCEDURE AlterLastInstructionTo (f: Instruction) ;
BEGIN
   IF NoOfInstructions>0
   THEN
      Instructions[NoOfInstructions].Inst := f
   ELSE
      WriteError('NoOfInstruction = 0  inconsistancy in M2Gen487')
   END
END AlterLastInstructionTo ;


(*
   IncRunTimeStack - increments the run time stack count and checks for
                     overflow. It could be altered to free up the stack
                     upon overflow. However at present it just gives an
                     error message.
*)

PROCEDURE IncRunTimeStack ;
BEGIN
   INC(RunTimeStack) ;
   IF RunTimeStack=MaxRunTimeStack
   THEN
      NearToken('fpu run time stack overflow', QuadToTokenNo(GetLastQuadNo())) ;
      WriteError('fpu run time stack overflow - time to recode the code generator Gaius')
   END
END IncRunTimeStack ;


(*
   DecRunTimeStack - decrements the run time stack
*)

PROCEDURE DecRunTimeStack ;
BEGIN
   IF RunTimeStack=0
   THEN
      WriteError('RunTimeStack should never = 0  bug in M2Gen487.mod')
   ELSE
      DEC(RunTimeStack)
   END
END DecRunTimeStack ;


(*
   CoprocessorInUse - returns TRUE if the coprocessor is in use
*)

PROCEDURE CoprocessorInUse () : BOOLEAN ;
BEGIN
   RETURN( NoOfInstructions>0 )
END CoprocessorInUse ;


(*
   FlushCoprocessor - flushes all coprocessor instructions
*)

PROCEDURE FlushCoprocessor ;
VAR
   l, i: CARDINAL ;
BEGIN
   IF CoprocessorInUse()
   THEN
      IF MathCoprocessor
      THEN
         WriteString('# flushing coprocessor instructions') ; WriteLn ;
         l := 0 ;
         i := 1 ;
         WHILE i<=NoOfInstructions DO
            WITH Instructions[i] DO
               WriteInstruction(Inst, Operand) ;
               IF LineNo#l
               THEN
                  LineInfo(LineNo)
               END ;
               l := Instructions[i].LineNo
            END ;
            INC(i)
         END ;
         NoOfInstructions := 0 ;
         CurrentLine := 0
      ELSE
         WriteError('user inconsistancy you have specified the -nomath option and you are generating coprocessor instructions')
      END
   END
END FlushCoprocessor ;


(*
   WriteInstruction - 
*)

PROCEDURE WriteInstruction (f: Instruction; o: CARDINAL) ;
BEGIN
   CASE f OF

   fadd    :  WriteString('    faddp  %st,%st(1)') ; WriteLn |
   fsub    :  WriteString('    fsubp  %st,%st(1)') ; WriteLn |
   fmul    :  WriteString('    fmulp  %st,%st(1)') ; WriteLn |
   fdiv    :  WriteString('    fdivp  %st,%st(1)') ; WriteLn |
   fmod    :  WriteString('    fprem') ; WriteLn ; WriteString('    fstp   %st(1)') ; WriteLn |
   fpush   :  CodePush(o) |
   fpop    :  CodePop(o) |
   fstore  :  CodeStore(o) |
   fdup    :  WriteString('    fld    %st') ; WriteLn |
   fchange :  WriteString('    fchs') ; WriteLn |
   fround  :  WriteString('    frndint') ; WriteLn |
   fabs    :  WriteString('    fabs   %st') ; WriteLn |
   fcompare:  WriteString('    fucompp') ; WriteLn ; WriteString('    fnstsw %ax') ; WriteLn |
   fsin    :  WriteString('    fsin') ; WriteLn |
   fcos    :  WriteString('    fcos') ; WriteLn |
   ftan    :  WriteString('    ftan') ; WriteLn |
   fatan   :  WriteString('    fatan') ; WriteLn |
   fsqrt   :  WriteString('    fsqrt') ; WriteLn |
   fnone   :

   ELSE
      WriteError('instruction not known - M2Gen487.mod')
   END
END WriteInstruction ;


(*
   CodeStore - codes the store coprocessor instruction. It depends on which
               coprocessor type we are using.
*)

PROCEDURE CodeStore (o: CARDINAL) ;
BEGIN
   IF IsConst(o)
   THEN
      InternalError('cannot store into a constant', __FILE__, __LINE__)
   ELSIF (GetType(o)=Integer) OR (GetType(o)=Cardinal) OR (GetType(o)=Address)
   THEN
      InstructionOperand('fistl', o)
   ELSIF GetType(o)=LongInt
   THEN
      InstructionOperand('fistq', o)
   ELSIF GetType(o)=Real
   THEN
      InstructionOperand('fsts', o)
   ELSIF GetType(o)=LongReal
   THEN
      InstructionOperand('fstl', o)
   ELSE
      WriteError('M2Gen487 does not understand this type')
   END
END CodeStore ;


(*
   InstructionOperand - writes instruction, i, and operand, o, but it will use
                        the indirect addressing of, o, if it is a ValueAddr.
*)

PROCEDURE InstructionOperand (i: ARRAY OF CHAR; o: CARDINAL) ;
BEGIN
   IF GetMode(o)=ValueAddr
   THEN
      WriteString('    movl   ') ; WriteOperand(o) ; WriteString(',%eax') ; WriteLn ;
      WriteString('    ') ; WriteString(i) ; WriteString('   (%eax)') ; WriteLn
   ELSE
      WriteString('    ') ; WriteString(i) ; Write(' ') ; WriteOperand(o) ; WriteLn
   END
END InstructionOperand ;


(*
   CodePop - codes the pop coprocessor instruction. It depends on which
             coprocessor type we are using.
*)

PROCEDURE CodePop (o: CARDINAL) ;
BEGIN
   IF IsConst(o)
   THEN
      InternalError('cannot pop into a constant', __FILE__, __LINE__)
   ELSIF (GetType(o)=Integer) OR (GetType(o)=Cardinal) OR (GetType(o)=Address)
   THEN
      InstructionOperand('fistpl', o)
   ELSIF GetType(o)=LongInt
   THEN
      InstructionOperand('fistpq', o)
   ELSIF GetType(o)=Real
   THEN
      InstructionOperand('fstps', o)
   ELSIF GetType(o)=LongReal
   THEN
      InstructionOperand('fstpl', o)
   ELSE
      WriteError('M2Gen487 does not understand this type')
   END
END CodePop ;


(*
   HasNoFraction - returns TRUE if constant, Sym, has no fractional component.
*)

PROCEDURE HasNoFraction (Sym: CARDINAL) : BOOLEAN ;
VAR
   a: ARRAY [0..200] OF CHAR ;
   r: LONGREAL ;
   i: CARDINAL ;
BEGIN
   IF IsConstLit(Sym)
   THEN
      GetKey(GetSymName(Sym), a)
   ELSE
      PushValue(Sym) ;
      r := PopLongReal() ;
      LongRealToStr(r, 200, 100, a)
   END ;
   i := 0 ;
   WHILE (i<HIGH(a)) AND (a[i]#'.') DO
      INC(i)
   END ;
   IF (i<HIGH(a)) AND (a[i]='.')
   THEN
      INC(i) ;
      WHILE (i<HIGH(a)) AND (a[i]#nul) AND (a[i]='0') DO
         INC(i)
      END ;
      IF (i=HIGH(a)) OR (a[i]=nul)
      THEN
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( TRUE )
   END
END HasNoFraction ;


(*
   IsWithinIntegerLimits - returns TRUE if constant Sym has a value within INTEGER limits
*)

PROCEDURE IsWithinIntegerLimits (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   PushValue(Sym) ;
   PushInt(2147483647) ;
   IF LessEqu()
   THEN
      PushInt(-2147483647) ;
      PushValue(Sym) ;
      RETURN( LessEqu() )
   ELSE
      RETURN( FALSE )
   END
END IsWithinIntegerLimits ;


(*
   IsWithinLongIntegerLimits - returns TRUE if constant Sym has a value within
                               LONGINT value.
*)

PROCEDURE IsWithinLongIntegerLimits (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( TRUE )
END IsWithinLongIntegerLimits ;


(*
   ConvertToInt - converts a constant into an integer constant.
*)

PROCEDURE ConvertToInt (Sym: CARDINAL) : CARDINAL ;
VAR
   New  : CARDINAL ;
   Value: INTEGER ;
   a    : ARRAY [0..80] OF CHAR ;
BEGIN
   PushValue(Sym) ;
   ConvertRealToInt ;
   Value := PopInt() ;
   IntToStr(Value, 0, a) ;
   New := MakeConstLit(MakeKey(a)) ;
   RETURN( New )
END ConvertToInt ;


(*
   CodePush - pushes operand, o, it checks for pushing special case values such
              as 0 and 1. It also decides which type push instruction to use.
*)

PROCEDURE CodePush (o: CARDINAL) ;
BEGIN
   IF IsZero(o)
   THEN
      WriteString('    fldz') ; WriteLn
   ELSIF IsOne(o)
   THEN
      WriteString('    fld1') ; WriteLn
   ELSIF o=pi
   THEN
      WriteString('    fldpi') ; WriteLn
   ELSIF IsConst(o)
   THEN
      IF GetType(o)=LongReal
      THEN
         IF HasNoFraction(o)
         THEN
            (* no point introducing rounding errors when we can avoid it *)
            InstructionOperand('fildq', ConvertToInt(o))
         ELSE
            InstructionOperand('fldl', o)
         END
      ELSE
         InstructionOperand('fildq', o)
      END
   ELSIF (GetType(o)=Integer) OR (GetType(o)=Cardinal) OR (GetType(o)=Address)
   THEN
      InstructionOperand('fildl', o)
   ELSIF GetType(o)=LongInt
   THEN
      InstructionOperand('fildq', o)
   ELSIF GetType(o)=Real
   THEN
      InstructionOperand('flds', o)
   ELSIF GetType(o)=LongReal
   THEN
      InstructionOperand('fldl', o)
   ELSE
      WriteError('M2Gen487 does not understand this type')
   END
END CodePush ;


(*
   WriteOperand - write the assembly operand for, o. It ignores the GetMode of o
                  and treats all operands as being OffsetAddr. Thus we can use it
                  as a building block to access a ValueAddr.
*)

PROCEDURE WriteOperand (o: CARDINAL) ;
VAR
   offset: INTEGER ;
BEGIN
   IF IsConst(o)
   THEN
      IncludeItemIntoList(Constants, o) ;
      WriteConstantName(o)
   ELSE
      IF IsProcedure(GetVarFather(o))
      THEN
         (* local variable *)
         PushOffset(o) ;
         WriteInt(PopInt(),0) ;
         WriteString('(%ebp)')
      ELSE
         WriteAsmName(o)
      END
   END
END WriteOperand ;


(*
   WriteConstantName - writes the constant name
*)

PROCEDURE WriteConstantName (o: CARDINAL) ;
BEGIN
   WriteString('CQ_') ; WriteCard(o, 0)
END WriteConstantName ;


(*
   WriteAllUsedConstants - write out all constants used within the module
*)

PROCEDURE WriteAllUsedConstants ;
BEGIN
   ForeachItemInListDo(Constants, WriteConstantAndValue)
END WriteAllUsedConstants ;


(*
   WriteConstantAndValue - writes a constant name and value
*)

PROCEDURE WriteConstantAndValue (Sym: CARDINAL) ;
BEGIN
   WriteConstantName(Sym) ;
   IF GetType(Sym)=LongReal
   THEN
      WriteString(':  .double  ') ;
      IF IsConstLit(Sym)
      THEN
         WriteKey(GetSymName(Sym))
      ELSE
         PushValue(Sym) ;
         WriteLongReal(PopLongReal(), 24, 12)
      END
   ELSE
      WriteString(':  .quad') ;
      PushValue(Sym) ;
      WriteInt(PopInt(), 8)
   END ;
   WriteLn
END WriteConstantAndValue ;


(*
   CoprocessorTerminate - writes all constants used within the compilation
*)

PROCEDURE CoprocessorTerminate ;
BEGIN 
   WriteString('.lcomm  CQfpu1,4') ; WriteLn ;   (* storage for fpu status words *)
   WriteString('.lcomm  CQfpu2,4') ; WriteLn ;
   WriteAllUsedConstants
END CoprocessorTerminate ;


(*
   Init - initialize the global variables
*)

PROCEDURE Init ;
BEGIN
   RunTimeStack := 0 ;
   NoOfInstructions := 0 ;
   CurrentLine := 0 ;
   InitList(Constants)
END Init ;


BEGIN
   Init
END M2Gen487.
