IMPLEMENTATION MODULE M2GenOpt486 ;


FROM SYSTEM IMPORT WORD ;
FROM StdIO IMPORT Write, PushOutput, PopOutput ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard, IntToStr ;

FROM NameKey IMPORT MakeKey, WriteKey ;
FROM M2Batch IMPORT MakeDefinitionSource ;
FROM M2Debug IMPORT Assert ;
FROM M2System IMPORT WordSize, Word ;
FROM M2AsmUtil IMPORT WriteAsmName ;
FROM M2Batch IMPORT MakeDefinitionSource ;
FROM M2Quads IMPORT QuadOperator, GetQuad, GetNextQuad, DisplayQuad ;
FROM M2Error IMPORT BeginError, EndError ;
FROM M2Lexical IMPORT InternalError, WriteError ;
FROM M2GDB IMPORT LineInfo ;
FROM M2Options IMPORT GenerateDebugging, OptimizeLength ;

FROM M2AS IMPORT InstRegConstant, InstRegReg, InstRegMD, InstMDReg, InstMDConstant,
                 WriteConstant, BinInstReg, BinInstMD ;

IMPORT M2Time ;
IMPORT IO ;

FROM M2Alias IMPORT MaxAlias, AliasSymbol, AliasList, AliasType,
                    InitAlias, KillAlias, DuplicateAlias,
                    CollectAllAliases, CollectAliases, RemoveAliasTo,
                    MakeClean,
                    IsDirty, IsClean, IsAliased, IsAbleToBeCleaned,
                    MakeDirty ;

FROM i486 IMPORT Register, Instruction, KindOfScope, DatumPrefix,
                 GetVarTypeSize, EqualSizes, FitsInRegister,
                 GetOffset, GetSize, GetScope ;

FROM SymbolTable IMPORT NulSym, ModeOfAddr, StartScope, EndScope,
                        MakeConstLit,
                        MakeTemporary, PutVar,
                        GetVarQuads, GetMode, GetStringLength,
                        GetSymName, GetType, GetScopeAuthor,
                        IsModule, IsTemporary, IsExported,
                        IsConst, IsConstString, IsProcedure, IsConstLit,
                        PushValue, PopValue, PushSize ;

FROM M2ALU IMPORT PopCard, Add, PopInt ;
FROM M2Gen487 IMPORT IsCoprocessorType, IsCoprocessorUsed ;


CONST
   MaxCost      = 10000 ;
   DebugOn      = FALSE ;
   MaxNumChars  =    20 ;
   MaxStack     =   500 ;
   MaxParam     =     6 ;

TYPE
   Procedures = (noneproc, codeline,
                 instmdreg, instregmd, instregreg, instregconst,
                 instmdconst, instbinmd, instbinreg) ;

   Parameters = RECORD
                   CurrentQuad: CARDINAL ;
                   w          : ARRAY [0..MaxParam] OF WORD ;
                END ;

   ProcedureDescription = RECORD
                             WhichProcedure : Procedures ;
                             WhichParameters: Parameters ;
                             Level          : CARDINAL ;
                          END ;

   ProcedureList = RECORD
                      ListPtr         : CARDINAL ;
                      ListOfProcedures: ARRAY [0..MaxStack] OF ProcedureDescription ;
                   END ;

VAR
   CurrentAssemblyCode: ProcedureList ;
   StartQuad, EndQuad,
   EAX, EBX, ECX, EDX,
   ESP, EBP, ESI, EDI,
   CurrentAlias       : AliasList ;


(*
   Strategy is to produce reasonable i486 code by aliasing

   maintain one list, CurrentAssemblyCode which contains the current assembly
   code we are working on.
*)

(* %%%FORWARD%%%
    PROCEDURE PutAllQuadsIntoMethods (Start, End, Count: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE DebugMethods (BestCost, CurrentCost: CARDINAL) ; FORWARD ;
    PROCEDURE CheckLeafIsCleanAndSaveIfBetter (a: AliasList; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakeFlushAllDirtySymbolsMethod ; FORWARD ;
    PROCEDURE DisplayMethod (p: ProcedureList; h: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE WriteNoneProc (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteForRegSym (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMakeClean (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE IsAllClean (a: AliasList) : BOOLEAN ; FORWARD ;
    PROCEDURE RemoveCurrentLevel ; FORWARD ;
    PROCEDURE CodeFlushAllDirtySymbols (a: AliasList; h: CARDINAL;
                                        BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE SubFromDone (h: CARDINAL) ; FORWARD ;
    PROCEDURE AddToDone (h: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeMoveThroughRegister (a: AliasList; h: CARDINAL;
                                       BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeInstRegConst (a: AliasList; h: CARDINAL;
                                BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeInstRegReg (a: AliasList; h: CARDINAL;
                              BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeLine (a: AliasList; h: CARDINAL;
                        BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeCopy (a: AliasList; h: CARDINAL;
                        BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeMoveFromTo (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeBecm (a: AliasList; h: CARDINAL;
                        BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeAddr (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeMakeClean (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeSymToSym (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE SymToSymMachineWord (a: AliasList; h: CARDINAL;
                                   BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeMoveThroughRegister (a: AliasList; h: CARDINAL;
                                       BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakeMoveThroughRegisterMethod (from: CARDINAL; frommode: ModeOfAddr;
                                             to  : CARDINAL; tomode  : ModeOfAddr;
                                             reg: CARDINAL) ; FORWARD ;
    PROCEDURE CodeSymToReg (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeExtToReg (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeRegToSym (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeRegToExt (a: AliasList; h: CARDINAL;
                            BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeRegToReg (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE NoOfRegisters () : CARDINAL ; FORWARD ;
    PROCEDURE GetRegister (i: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE IsRegister (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
    PROCEDURE ToCard (w: WORD) : CARDINAL ; FORWARD ;
    PROCEDURE ToScope (w: WORD) : KindOfScope ; FORWARD ;
    PROCEDURE ToInst (w: WORD) : Instruction ; FORWARD ;
    PROCEDURE ToDatum (w: WORD) : DatumPrefix ; FORWARD ;
    PROCEDURE ToReg (w: WORD) : Register ; FORWARD ;
    PROCEDURE ToMode (w: WORD) : ModeOfAddr ; FORWARD ;
    PROCEDURE TimeOf (h: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE EmitBestInstructions (Cost, Calls: CARDINAL) ; FORWARD ;
    PROCEDURE MakeQuadMethod (quad: CARDINAL) ; FORWARD ;
    PROCEDURE DoNextProcedure (a: AliasList; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE IsAbleToCodeQuad (q: CARDINAL) : BOOLEAN ; FORWARD ;
    PROCEDURE PushMethod (p: Procedures;
                          w0, w1, w2, w3, w4, w5, w6, w7: WORD) ; FORWARD ;
    PROCEDURE DisplayMethods (p: ProcedureList) ; FORWARD ;
    PROCEDURE WriteNSpaces (n: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMethod (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteSymToSym (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteRegToSym (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteSymToReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteRegToExt (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteExtToReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteRegToReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteCodeCopy (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteCodeAddr (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteCodeBecm (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteCodeLine (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteSymbol (Sym: CARDINAL; Mode: ModeOfAddr) ; FORWARD ;
    PROCEDURE WriteRegister (Sym: CARDINAL; Mode: ModeOfAddr) ; FORWARD ;
    PROCEDURE WriteRegisterName (Sym: CARDINAL) ; FORWARD ;
    PROCEDURE MakeQuadMethod (CurrentQuad, EndQuad: CARDINAL) ; FORWARD ;
    PROCEDURE SaveCurrentIntoBest (a: AliasList; BestCost, CurrentCost: CARDINAL) ; FORWARD ;
    PROCEDURE DumpCurrentInstructions ; FORWARD ;
    PROCEDURE ForeachRegisterDoSymToSym (a: AliasList; h: CARDINAL;
                                         BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE SymToSymMachineWord (a: AliasList; h: CARDINAL;
                                   BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE GenerateOptimalCode (Head, Start, End: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstBinMD (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstBinReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMakeChanged (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMoveFrom (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMakeWillBeClean (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMakeFlush (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMakeAssign (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteMoveViaReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstMDReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstRegMD (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstRegReg (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstRegConst (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteInstMDConst (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE DumpCurrentInstructions ; FORWARD ;
    PROCEDURE IsUsedOutSide (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
    PROCEDURE CodeAssign (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeWillBeClean (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeNone (a: AliasList; h: CARDINAL;
                        BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE DumpInstructions (p: ProcedureList) ; FORWARD ;
    PROCEDURE CodeInstMDConst (a: AliasList; h: CARDINAL;
                               BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeInstMDReg (a: AliasList; h: CARDINAL;
                             BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeInstRegMD (a: AliasList; h: CARDINAL;
                             BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeInstBinReg (a: AliasList; h: CARDINAL;
                              BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE CodeInstBinMD (a: AliasList; h: CARDINAL;
                             BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakeCleanMoveIntoMethod (from: CARDINAL; frommode: ModeOfAddr;
                                       to  : CARDINAL; tomode  : ModeOfAddr) ; FORWARD ;
    PROCEDURE CodeCleanMoveInto (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE WriteCleanMoveInto (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE MakeCodeAddrMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
    PROCEDURE WriteCodeAdd (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE CodeAdd (a: AliasList; h: CARDINAL;
                       BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakePeepAddMethod (op1: CARDINAL; m1: ModeOfAddr;
                                 op2: CARDINAL; m2: ModeOfAddr;
                                 op3: CARDINAL; m3: ModeOfAddr) ; FORWARD ;
    PROCEDURE CodePeepAdd (a: AliasList; h: CARDINAL;
                           BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE WritePeepAdd (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE WriteCodeInc (p: ProcedureList; m: CARDINAL) ; FORWARD ;
    PROCEDURE MakeIncMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ; FORWARD ;
    PROCEDURE CodeInc (a: AliasList; h: CARDINAL;
                       BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakeChanged (sym: CARDINAL; mode: ModeOfAddr) ; FORWARD ;
    PROCEDURE CodeMakeChanged (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakeForeachRegisterDoSymToSym (from: CARDINAL; frommode: ModeOfAddr;
                                             to: CARDINAL; tomode: ModeOfAddr) ; FORWARD ;
    PROCEDURE CodeForeachRegisterDoSymToSym (a: AliasList; h: CARDINAL;
                                             BestCost, CurrentCost: CARDINAL) : CARDINAL ; FORWARD ;
    PROCEDURE MakeCodeAddMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)



(*
   IsAbleToCode - is the dynamic code generator able to encode these quadruples?
*)

PROCEDURE IsAbleToCode (Start, End: CARDINAL) : BOOLEAN ;
BEGIN
   WHILE Start#End DO
      IF IsAbleToCodeQuad(Start)
      THEN
         Start := GetNextQuad(Start)
      ELSE
         RETURN( FALSE )
      END
   END ;
   RETURN( IsAbleToCodeQuad(Start) )
END IsAbleToCode ;


(*
   IsAbleToCodeQuad - returns TRUE if we can encode this quadruple.
*)

PROCEDURE IsAbleToCodeQuad (q: CARDINAL) : BOOLEAN ;
VAR
   Op : QuadOperator ;
   Op1,
   Op2,
   Op3: CARDINAL ;
BEGIN
   GetQuad(q, Op, Op1, Op2, Op3) ;
   CASE Op OF

   AddrOp        :  RETURN( TRUE ) |
   CopyOp        :  RETURN( TRUE ) |
   BecomesOp     :  RETURN( TRUE ) |
   AddOp         :  RETURN( NOT IsCoprocessorUsed(Op1, Op2, Op3) ) |
   MultOp        :  RETURN( TRUE ) |
   BaseOp        :  RETURN( TRUE ) |
   GotoOp        :  RETURN( TRUE ) |
   OffsetOp      :  RETURN( TRUE ) |
   NewLocalVarOp :  RETURN( TRUE ) |
   KillLocalVarOp:  RETURN( TRUE ) |
   LineNumberOp  :  RETURN( TRUE )

   ELSE
      WriteLn ;
      WriteString('cannot encode: ') ;
      DisplayQuad(q) ;
      RETURN( FALSE )
   END
END IsAbleToCodeQuad ;


(*
   GenCode - takes the list of quadruples from Head and generates
             assembly langauge between quadruples Start and End.
*)

PROCEDURE GenCode (Head, Start, End: CARDINAL) ;
BEGIN
   PushOutput(IO.Write) ; (* debugging purposes *)
   GenerateOptimizedCode(Head, Start, End) ;
   PopOutput
END GenCode ;


(*
   GenerateOptimizedCode - 
*)

PROCEDURE GenerateOptimizedCode (Head, Start, End: CARDINAL) ;
VAR
   Cost: CARDINAL ;
   a   : AliasList ;
BEGIN
   (* initialize alias list *)
   a := InitAlias() ;
   StartQuad := Start ;
   EndQuad := End ;
   REPEAT
      BestAlias := NIL ;
      CurrentLevel := 0 ;
      NoOfCalls := 0 ;
      (* zap the assembly language lists *)
      CurrentAssemblyCode.ListPtr := 0 ;
      WriteString('coding the following quads:') ; WriteLn ;
      Start := PutAllQuadsIntoMethods(Start, End, OptimizeLength) ;
      Cost := DoNextProcedure(a, MaxCost, 0) ;
      CodeFlushAllDirtySymbolsMethod ; (* must clean up the dirty symbols *)
      EmitBestInstructions(Cost, NoOfCalls) ;
      IF a=NIL
      THEN
         InternalError('alias list should not be NIL', __FILE__, __LINE__)
      END ;
      a := KillAlias(a) ;
      a := BestAlias ;
      IF a=NIL
      THEN
         InternalError('alias list should not be NIL', __FILE__, __LINE__)
      END
   UNTIL Start=0 ;
   a := KillAlias(a)
END GenerateOptimalCode ;


(*
   NumberOfQuads - returns the number of quads between, Start, and, End.
*)

PROCEDURE NumberOfQuads (Start, End: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE (Start#End) AND (Start#0) DO
      Start := GetNextQuad(Start) ;
      INC(i)
   END ;
   RETURN( i )
END NumberOfQuads ;


(*
   PutAllQuadsIntoMethods - 
*)

PROCEDURE PutAllQuadsIntoMethods (Start, End, Count: CARDINAL) : CARDINAL ;
VAR
   old: CARDINAL ;
   op : QuadOperator ;
   op1,
   op2,
   op3: CARDINAL ;
BEGIN
   IF Count=0
   THEN
      RETURN( Start )
   ELSE
      DisplayQuad(Start) ;
      GetQuad(Start, op, op1, op2, op3) ;
      old := Start ;
      IF Start=End
      THEN
         Start := 0  (* signifies end *)
      ELSE
         IF op=LineNumberOp
         THEN
            Start := PutAllQuadsIntoMethods(GetNextQuad(Start), End, Count)
         ELSE
            Start := PutAllQuadsIntoMethods(GetNextQuad(Start), End, Count-1)
         END
      END ;
      IF op#LineNumberOp  (* dont bother to put line number ops in the list (takes too long) *)
      THEN
         MakeQuadMethod(old) ;
         ToDoProcedures.ListOfProcedures[ToDoProcedures.ListPtr].WhichParameters.CurrentQuad := old
      END ;
      RETURN( Start )
   END
END PutAllQuadsIntoMethods ;


(*
   FitsIntoRegister - returns TRUE if a symbol, Sym, fill fit into a register.
*)

PROCEDURE FitsIntoRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym)
   THEN
      IF IsConstString(Sym)
      THEN
         RETURN( WordSize>GetStringLength(Sym) )
      ELSE
         (* we assume all constants can fit into a register, this will NOT work for LONGREAL *)
         RETURN( TRUE )
      END
   ELSE
      PushSize(Sym) ;
      RETURN( WordSize>=PopCard() )
   END
END FitsIntoRegister ;


(*
   TimeOf - returns the number of cycles that the following instructions consume.
*)

PROCEDURE TimeOf (h: CARDINAL) : CARDINAL ;
VAR
   i: CARDINAL ;
BEGIN
   WITH DoneProcedures DO
      WITH ListOfProcedures[h].WhichParameters DO
         CASE ListOfProcedures[h].WhichProcedure OF

         instregconst: RETURN( M2Time.InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) ) |
         instregreg  : RETURN( M2Time.InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])) ) |
         instregmd   : RETURN( M2Time.InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), w[3], w[4], ToScope(w[5]), w[6]) ) |
         instmdreg   : RETURN( M2Time.InstMDReg(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToReg(w[4]), ToScope(w[5]), w[6]) ) |
         instmdconst : RETURN( M2Time.InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]) ) |
         instbinmd   : RETURN( M2Time.BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]) ) |
         instbinreg  : RETURN( M2Time.BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])) )

         ELSE
            InternalError('case statement should only be dealing with assembly language instructions', __FILE__, __LINE__)
         END
      END
   END
END TimeOf ;


(*
   EmitBestInstructions - 
*)

PROCEDURE EmitBestInstructions (Cost, Calls: CARDINAL) ;
BEGIN
   WriteString('BEST Instructions (cost = ') ; WriteCard(Cost, 0) ;
   WriteString(', calls = ') ; WriteCard(Calls, 0) ; Write(')') ; WriteLn ;
   DumpInstructions(BestAssemblyCode)
END EmitBestInstructions ;


(*
   ToCard - 
*)

PROCEDURE ToCard (w: WORD) : CARDINAL ;
BEGIN
   RETURN( VAL(CARDINAL, w) )
END ToCard ;


(*
   ToScope - returns the KindOfScope represented by, w.
*)

PROCEDURE ToScope (w: WORD) : KindOfScope ;
BEGIN
   RETURN( VAL(KindOfScope, w) )
END ToScope ;


(*
   ToInst - convert a word into an instruction
*)

PROCEDURE ToInst (w: WORD) : Instruction ;
BEGIN
   RETURN( VAL(Instruction, w) )
END ToInst ;


(*
   ToDatum - convert a word into a DatumPrefix
*)

PROCEDURE ToDatum (w: WORD) : DatumPrefix ;
BEGIN
   CASE CARDINAL(w) OF

   1: RETURN( PByte ) |
   2: RETURN( PWord ) |
   4: RETURN( PDWord )

   ELSE
      InternalError('size unexpected in ToDatum', __FILE__, __LINE__)
   END
END ToDatum ;


(*
   ToReg - converts a WORD into a Register.
*)

PROCEDURE ToReg (w: WORD) : Register ;
BEGIN
   IF CARDINAL(w)=EAX
   THEN
      RETURN( RegEAX )
   ELSIF CARDINAL(w)=EBX
   THEN
      RETURN( RegEBX )
   ELSIF CARDINAL(w)=ECX
   THEN
      RETURN( RegECX )
   ELSIF CARDINAL(w)=EDX
   THEN
      RETURN( RegEDX )
   ELSIF CARDINAL(w)=EBP
   THEN
      RETURN( RegEBP )
   ELSIF CARDINAL(w)=ESP
   THEN
      RETURN( RegESP )
   ELSIF CARDINAL(w)=ESI
   THEN
      RETURN( RegESI )
   ELSIF CARDINAL(w)=EDI
   THEN
      RETURN( RegEDI )
   ELSE
      InternalError('unknown register', __FILE__, __LINE__)
   END
END ToReg ;


(*
   ToMode - 
*)

PROCEDURE ToMode (w: WORD) : ModeOfAddr ;
BEGIN
   RETURN( VAL(ModeOfAddr, w) )
END ToMode ;




(*
   DebugMethods - 
*)

PROCEDURE DebugMethods (BestCost, CurrentCost: CARDINAL) ;
BEGIN
   WriteLn ;
   WriteString('=================================================================') ; WriteLn ;
   WriteString('                            (Current cost:') ; WriteCard(CurrentCost, 4) ; Write(')') ; WriteLn ;
   WriteString('                            (Best    cost:') ; WriteCard(BestCost, 4) ; Write(')') ; WriteLn ;
   WriteString('current assembly code:') ; WriteLn ;
   DisplayMethods(CurrentAssemblyCode) ;
   WriteLn
END DebugMethods ;


(*
   CheckLeafIsCleanAndSaveIfBetter - 
*)

PROCEDURE CheckLeafIsCleanAndSaveIfBetter (a: AliasList; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   IF DebugOn
   THEN
      WriteString(' (leaf)')
   END ;
   IF IsAllClean(a)
   THEN
      IF CurrentCost<BestCost
      THEN
         (*
            at this point all symbols are clean and we are at the end of
            the basic block. We also know that the current cost is better
            than our previous best and so we can update our best score.
         *)
         IF DebugOn
         THEN
            WriteString(' best so far...') ; WriteLn
         END ;
         SaveCurrentIntoBest(a, BestCost, CurrentCost) ;
         RETURN( CurrentCost )
      ELSE
         IF DebugOn
         THEN
            WriteString(' [not as good as the best]') ; WriteLn
         END ;
         RETURN( BestCost )
      END
   ELSE
      IF DebugOn
      THEN
         WriteString(' [symbols are not clean]') ; WriteLn
      END ;
      RETURN( BestCost )
   END
END CheckLeafIsCleanAndSaveIfBetter ;


(*
   IsAllClean - 
*)

PROCEDURE IsAllClean (a: AliasList) : BOOLEAN ;
VAR
   Start, End: CARDINAL ;
   n, i      : CARDINAL ;
   al        : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   WITH DoneProcedures.ListOfProcedures[DoneProcedures.ListPtr].WhichParameters DO
      n := CollectAllAliases(a, al) ;
      i := 0 ;
      WHILE i<n DO
         IF NOT IsConst(al[i].Sym)
         THEN
            IF IsDirty(a, al[i].Sym, al[i].Mode) AND IsUsedOutSide(al[i].Sym)
            THEN
               RETURN( FALSE )
            END
         END ;
         INC(i)
      END
   END ;
   RETURN( TRUE )
END IsAllClean ;


(*
   AddToDone - adds a ToDoProcedure to the DoneProcedure list and returns the old
               ListPtr before this procedure is added to this list.
*)

PROCEDURE AddToDone (h: CARDINAL) : CARDINAL ;
BEGIN
   WITH DoneProcedures DO
      INC(ListPtr) ;
      IF ListPtr>MaxStack
      THEN
         InternalError('increase MaxStack', __FILE__, __LINE__)
      ELSE
         ListOfProcedures[ListPtr] := ToDoProcedures.ListOfProcedures[h] ;
         RETURN( ListPtr-1 )
      END
   END
END AddToDone ;


(*
   SubFromDone - removes all procedures above, h, on the DoneProcedures list.
*)

PROCEDURE SubFromDone (h: CARDINAL) ;
BEGIN
   WITH DoneProcedures DO
      ListPtr := h
   END
END SubFromDone ;


(*
   AddToCurrentCode - adds procedure, h, to the current code pile of assembly instructions.
*)

PROCEDURE AddToCurrentCode (h: CARDINAL) ;
BEGIN
   WITH CurrentAssemblyCode DO
      INC(ListPtr) ;
      IF ListPtr>MaxStack
      THEN
         InternalError('increase MaxStack', __FILE__, __LINE__)
      ELSE
         ListOfProcedures[ListPtr] := DoneProcedures.ListOfProcedures[h]
      END
   END
END AddToCurrentCode ;


(*
   MakeInstBinMethod - 
*)

PROCEDURE MakeInstBinMethod (inst: Instruction; size: CARDINAL;
                             sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   IF IsRegister(sym)
   THEN
      PushMethod(instbinreg,
                 inst, size, sym, mode, 0, 0, 0, 0)
   ELSE
      PushMethod(instbinmd,
                 inst, size, GetOffset(sym), 0, GetScope(sym), sym, mode, 0)
   END
END MakeInstBinMethod ;


(*
   MakeInstRegConstantMethod - 
*)

PROCEDURE MakeInstRegConstantMethod (inst: Instruction; size: CARDINAL;
                                     reg: CARDINAL; sym: CARDINAL) ;
BEGIN
   PushMethod(instregconst,
              inst, size, reg, sym, 0, 0, 0, 0)
END MakeInstRegConstantMethod ;


(*
   MakeInstRegRegMethod - 
*)

PROCEDURE MakeInstRegRegMethod (inst: Instruction; size: CARDINAL;
                                fromreg, toreg: CARDINAL) ;
BEGIN
   PushMethod(instregreg,
              inst, size, toreg, fromreg, 0, 0, 0, 0)
END MakeInstRegRegMethod ;


(*
   MakeInstRegMDMethod - 
*)

PROCEDURE MakeInstRegMDMethod (inst: Instruction; size: CARDINAL;
                               reg: CARDINAL; offset1, offset2: CARDINAL;
                               scope: KindOfScope; sym: CARDINAL) ;
BEGIN
   PushMethod(instregmd,
              inst, size, reg, offset1, offset2, scope, sym, 0)
END MakeInstRegMDMethod ;


(*
   MakeInstMDRegMethod - 
*)

PROCEDURE MakeInstMDRegMethod (inst: Instruction; size: CARDINAL;
                               offset1, offset2: CARDINAL;
                               scope: KindOfScope; sym: CARDINAL; reg: CARDINAL) ;
BEGIN
   PushMethod(instmdreg,
              inst, size, offset1, offset2, reg, scope, sym, 0)
END MakeInstMDRegMethod ;


(*
   MakeInstMDConstantMethod - 
*)

PROCEDURE MakeInstMDConstantMethod (inst: Instruction; size: CARDINAL;
                                    offset1, offset2: CARDINAL;
                                    scope: KindOfScope; sym: CARDINAL; const: CARDINAL) ;
BEGIN
   PushMethod(instmdconst,
              inst, size, offset1, offset2, scope, sym, const, 0)
END MakeInstMDConstantMethod ;


(*
   MakeFlushAllDirtySymbolsMethod - makes a FlushAllDirtySymbols method.
*)

PROCEDURE MakeFlushAllDirtySymbolsMethod ;
BEGIN
   PushMethod(makeflush,
              0, 0, 0, 0, 0, 0, 0, 0)
END MakeFlushAllDirtySymbolsMethod ;


(*
   MakeCleanSymbolMethod - makes a clean symbol method.
*)

PROCEDURE MakeCleanSymbolMethod (sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   PushMethod(makeclean,
              sym, mode, 0, 0, 0, 0, 0, 0)
END MakeCleanSymbolMethod ;


(*
   MakeWillBeCleanMethod - 
*)

PROCEDURE MakeWillBeCleanMethod (sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   PushMethod(makewillbeclean,
              sym, mode, 0, 0, 0, 0, 0, 0)
END MakeWillBeCleanMethod ;



(*
   MakeMoveFromToMethod -
*)

PROCEDURE MakeMoveFromToMethod (from: CARDINAL; frommode: ModeOfAddr;
                                to  : CARDINAL; tomode  : ModeOfAddr) ;
BEGIN
   (* no point moving a symbol into itself *)
   IF (from#to) OR (frommode#tomode)
   THEN
      PushMethod(movefrom,
                 from, frommode, to, tomode, 0, 0, 0, 0)
   END
END MakeMoveFromToMethod ;


(*
   MakeSymToSymMethod - 
*)

PROCEDURE MakeSymToSymMethod (from: CARDINAL; frommode: ModeOfAddr;
                              to  : CARDINAL; tomode  : ModeOfAddr) ;
BEGIN
   IF (from#to) OR (frommode#tomode)
   THEN
      PushMethod(symtosym,
                 from, frommode, to, tomode, 0, 0, 0, 0)
   END
END MakeSymToSymMethod ;


(*
   MakeRegToSymMethod - 
*)

PROCEDURE MakeRegToSymMethod (reg: CARDINAL;
                              to : CARDINAL; tomode: ModeOfAddr) ;
BEGIN
   IF (reg#to) OR (tomode#OffsetAddr)
   THEN
      PushMethod(regtosym,
                 reg, to, tomode, 0, 0, 0, 0, 0)
   END
END MakeRegToSymMethod ;


(*
   MakeSymToRegMethod - 
*)

PROCEDURE MakeSymToRegMethod (from: CARDINAL; frommode: ModeOfAddr;
                              reg : CARDINAL) ;
BEGIN
   IF (from#reg) OR (frommode#OffsetAddr)
   THEN
      PushMethod(symtoreg,
                 from, frommode, reg, 0, 0, 0, 0, 0)
   END
END MakeSymToRegMethod ;


(*
   MakeRegToRegMethod - 
*)

PROCEDURE MakeRegToRegMethod (fromreg, toreg: CARDINAL; size: CARDINAL) ;
BEGIN
   IF fromreg#toreg
   THEN
      PushMethod(regtoreg,
                 fromreg, toreg, size, 0, 0, 0, 0, 0)
   END
END MakeRegToRegMethod ;


(*
   MakeRegToExtMethod - 
*)

PROCEDURE MakeRegToExtMethod (reg: CARDINAL;
                              to : CARDINAL; tomode : ModeOfAddr) ;
BEGIN
   IF (reg#to) OR (tomode#OffsetAddr)
   THEN
      PushMethod(regtoext,
                 reg, to, tomode, 0, 0, 0, 0, 0)
   END
END MakeRegToExtMethod ;


(*
   MakeExtToRegMethod - 
*)

PROCEDURE MakeExtToRegMethod (from: CARDINAL; frommode: ModeOfAddr;
                              reg : CARDINAL) ;
BEGIN
   IF (from#reg) OR (frommode#OffsetAddr)
   THEN
      PushMethod(exttoreg,
                 from, frommode, reg, 0, 0, 0, 0, 0)
   END
END MakeExtToRegMethod ;


(*
   MakeAssign - 
*)

PROCEDURE MakeAssign (from: CARDINAL; frommode: ModeOfAddr;
                      to: CARDINAL; tomode: ModeOfAddr) ;
BEGIN
   PushMethod(makeassign,
              from, frommode, to, tomode, 0, 0, 0, 0)
END MakeAssign ;


(*
   MakeCodeBecmMethod - 
*)

PROCEDURE MakeCodeBecmMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
VAR
   Size1, Size3: CARDINAL ;
BEGIN
   (* check for assignment errors *)
   IF IsConst(op1) AND IsConst(op3)
   THEN
   ELSIF IsConstString(op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
   ELSIF IsProcedure(op3)
   THEN
   ELSE
      Size1 := GetVarTypeSize(op1) ;
      IF IsConst(op3)
      THEN
         Size3 := Size1
      ELSE
         Size3 := GetVarTypeSize(op3)
      END ;
      IF (Size1#Size3) AND (NOT (IsCoprocessorType(op1) AND (IsCoprocessorType(op3))))
      THEN
         BeginError ;
         WriteLn ;
         WriteKey(GetSymName(op1)) ; WriteString(' := ') ; WriteKey(GetSymName(op3)) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(op1)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(op1))) ; WriteLn ;
         WriteString('Type of ') ; WriteKey(GetSymName(op3)) ;
         WriteString(' is ') ; WriteKey(GetSymName(GetType(op3))) ; WriteLn ;
         WriteError('Different sizes in :=')
      END
   END ;
   (* end of checking *)
   PushMethod(codebecm,
              op1, GetMode(op1), op3, GetMode(op3), 0, 0, 0, 0)
END MakeCodeBecmMethod ;


(*
   MakeCodeLineMethod - 
*)

PROCEDURE MakeCodeLineMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   PushMethod(codeline,
              op3, 0, 0, 0, 0, 0, 0, 0)
END MakeCodeLineMethod ;


(*
   MakeCodeAddrMethod - 
*)

PROCEDURE MakeCodeAddrMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   PushMethod(codeaddr,
              op1, GetMode(op1), op3, GetMode(op3), 0, 0, 0, 0)
END MakeCodeAddrMethod ;


(*
   MakeCodeCopyMethod - 
*)

PROCEDURE MakeCodeCopyMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   PushMethod(codecopy,
              op1, GetMode(op1), op3, GetMode(op3), 0, 0, 0, 0)
END MakeCodeCopyMethod ;


(*
   PushMethod - 
*)

PROCEDURE PushMethod (p: Procedures;
                      w0, w1, w2, w3, w4, w5, w6, w7: WORD) ;
BEGIN
   WITH ToDoProcedures DO
      IF ListPtr=MaxStack
      THEN
         InternalError('internal error - increase MaxStack in M2GenDyn486', __FILE__, __LINE__)
      ELSE
         INC(ListPtr) ;
         WITH ListOfProcedures[ListPtr] DO
            WhichProcedure := p ;
            IF ListPtr>1
            THEN
               WhichParameters.CurrentQuad := ListOfProcedures[ListPtr-1].WhichParameters.CurrentQuad
            ELSE
               WhichParameters.CurrentQuad := StartQuad
            END ;
            WITH WhichParameters DO
               w[0] := w0 ;
               w[1] := w1 ;
               w[2] := w2 ;
               w[3] := w3 ;
               w[4] := w4 ;
               w[5] := w5 ;
               w[6] := w6 ;
               w[7] := w7
            END ;
            Level := CurrentLevel
         END
      END
   END
END PushMethod ;


(*
   DisplayMethods - 
*)

PROCEDURE DisplayMethods (p: ProcedureList) ;
VAR
   h: CARDINAL ;
BEGIN
   h := DisplayMethod(p, 1) ;
   WriteLn
END DisplayMethods ;


(*
   DisplayMethod - 
*)

PROCEDURE DisplayMethod (p: ProcedureList; h: CARDINAL) : CARDINAL ;
BEGIN
   WHILE h<=p.ListPtr DO
      Write('(') ;
      WriteMethod(p, h) ;
      IF (h<p.ListPtr) AND (p.ListOfProcedures[h].Level<p.ListOfProcedures[h+1].Level)
      THEN
         h := DisplayMethod(p, h+1)
      ELSE
         INC(h)
      END ;
      Write(')')
   END ;
   RETURN( h )
END DisplayMethod ;


(*
   WriteNSpaces - writes, n, spaces.
*)

PROCEDURE WriteNSpaces (n: CARDINAL) ;
BEGIN
   WHILE n>0 DO
      Write(' ') ;
      DEC(n)
   END
END WriteNSpaces ;


(*
   WriteMethod - writes a method to the default output stream.
*)

PROCEDURE WriteMethod (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m] DO
      CASE WhichProcedure OF

      moveinto        :  WriteCleanMoveInto(p, m) |
      makeassign      :  WriteMakeAssign(p, m) |
      makewillbeclean :  WriteMakeWillBeClean(p, m) |
      forregsym       :  WriteForRegSym(p, m) |
      noneproc        :  WriteNoneProc(p, m) |
      symtosym        :  WriteSymToSym(p, m) |
      symtoreg        :  WriteSymToReg(p, m) |
      regtosym        :  WriteRegToSym(p, m) |
      regtoreg        :  WriteRegToReg(p, m) |
      exttoreg        :  WriteExtToReg(p, m) |
      regtoext        :  WriteRegToExt(p, m) |
      codecopy        :  WriteCodeCopy(p, m) |
      codeaddr        :  WriteCodeAddr(p, m) |
      codebecm        :  WriteCodeBecm(p, m) |
      codeline        :  WriteCodeLine(p, m) |
      codeadd         :  WriteCodeAdd(p, m) |
      peepadd         :  WritePeepAdd(p, m) |
      codeinc         :  WriteCodeInc(p, m) |
      movefrom        :  WriteMoveFrom(p, m) |
      makeclean       :  WriteMakeClean(p, m) |
      makechanged     :  WriteMakeChanged(p, m) |
      makeflush       :  WriteMakeFlush(p, m) |
      moveviareg      :  WriteMoveViaReg(p, m) |
      instmdreg       :  WriteInstMDReg(p, m) |
      instregmd       :  WriteInstRegMD(p, m) |
      instregreg      :  WriteInstRegReg(p, m) |
      instregconst    :  WriteInstRegConst(p, m) |
      instmdconst     :  WriteInstMDConst(p, m) |
      instbinmd       :  WriteInstBinMD(p, m) |
      instbinreg      :  WriteInstBinReg(p, m)

      ELSE
         InternalError('unknown procedure in case', __FILE__, __LINE__)
      END
   END
END WriteMethod ;


(*
   WriteNoneProc - 
*)

PROCEDURE WriteNoneProc (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WriteString('(noneproc)')
END WriteNoneProc ;


(*
   WriteInstBinMD - 
*)

PROCEDURE WriteInstBinMD (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]), 0) ;
      WriteString('>" ') ;
      BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]) ;
      WriteString('" ')
   END
END WriteInstBinMD ;


(*
   WriteInstBinReg - 
*)

PROCEDURE WriteInstBinReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])), 0) ;
      WriteString('>" ') ;
      BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])) ;
      WriteString('" ')
   END
END WriteInstBinReg ;


(*
   WriteForRegSym - 
*)

PROCEDURE WriteForRegSym (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('ForRegSym( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ANYREG -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteForRegSym ;


(*
   WriteSymToSym - 
*)

PROCEDURE WriteSymToSym (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('SymToSym( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteSymToSym ;


(*
   WriteRegToSym - 
*)

PROCEDURE WriteRegToSym (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('RegToSym( ') ;
      WriteSymbol(ToCard(w[0]), OffsetAddr) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[1]), ToMode(w[2])) ;
      WriteString(' )')
   END
END WriteRegToSym ;


(*
   WriteSymToReg - 
*)

PROCEDURE WriteSymToReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('SymToReg( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), OffsetAddr) ;
      WriteString(' )')
   END
END WriteSymToReg ;


(*
   WriteRegToExt - 
*)

PROCEDURE WriteRegToExt (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('RegToExt( ') ;
      WriteSymbol(ToCard(w[0]), OffsetAddr) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[1]), ToMode(w[2])) ;
      WriteString(' )')
   END
END WriteRegToExt ;


(*
   WriteExtToReg - 
*)

PROCEDURE WriteExtToReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('ExtToReg( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), OffsetAddr) ;
      WriteString(' )')
   END
END WriteExtToReg ;


(*
   WriteRegToReg - 
*)

PROCEDURE WriteRegToReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('RegToReg( ') ;
      WriteSymbol(ToCard(w[0]), OffsetAddr) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[1]), OffsetAddr) ;
      WriteString(' )')
   END
END WriteRegToReg ;


(*
   WriteCodeCopy - 
*)

PROCEDURE WriteCodeCopy (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CodeCopy( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCodeCopy ;


(*
   WriteCodeAddr - 
*)

PROCEDURE WriteCodeAddr (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CodeAddr( ADR(') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(') -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCodeAddr ;


(*
   WriteCodeBecm - 
*)

PROCEDURE WriteCodeBecm (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CodeBecm( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' := ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCodeBecm ;


(*
   WriteCodeLine - 
*)

PROCEDURE WriteCodeLine (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('(line ') ;
      WriteCard(ToCard(w[0]), 0) ;
      Write(')')
   END
END WriteCodeLine ;


(*
   WriteMoveFrom - 
*)

PROCEDURE WriteMoveFrom (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MoveFrom( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteMoveFrom ;


(*
   WriteMakeClean - 
*)

PROCEDURE WriteMakeClean (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MakeClean( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteMakeClean ;


(*
   WriteMakeChanged - 
*)

PROCEDURE WriteMakeChanged (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MakeChanged( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteMakeChanged ;


(*
   WriteMakeWillBeClean - 
*)

PROCEDURE WriteMakeWillBeClean (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('isclean( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteMakeWillBeClean ;


(*
   WriteMakeFlush - 
*)

PROCEDURE WriteMakeFlush (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WriteString('MakeFlush() ')
END WriteMakeFlush ;


(*
   WriteMakeAssign - 
*)

PROCEDURE WriteMakeAssign (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('assign( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteMakeAssign ;


(*
   WriteMoveViaReg - 
*)

PROCEDURE WriteMoveViaReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('MoveViaReg( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteRegister(ToCard(w[4]), OffsetAddr) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteMoveViaReg ;


(*
   WriteInstMDReg - 
*)

PROCEDURE WriteInstMDReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstMDReg(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToReg(w[4]), ToScope(w[5]), w[6]), 0) ;
      WriteString('>" ') ;
      InstMDReg(ToInst(w[0]), ToDatum(w[1]), ToCard(w[2]), ToCard(w[3]),
                ToReg(w[4]), ToScope(w[5]), ToCard(w[6])) ;
      WriteString('" ')
   END
END WriteInstMDReg ;


(*
   WriteInstRegMD - 
*)

PROCEDURE WriteInstRegMD (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), w[3], w[4], ToScope(w[5]), w[6]), 0) ;
      WriteString('>" ') ;
      InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3]),
                ToCard(w[4]), ToScope(w[5]), ToCard(w[6])) ;
      WriteString('" ')
   END
END WriteInstRegMD ;


(*
   WriteInstRegReg -
*)

PROCEDURE WriteInstRegReg (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])), 0) ;
      WriteString('>" ') ;
      InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])) ;
      WriteString('" ')
   END
END WriteInstRegReg ;


(*
   WriteInstRegConst -
*)

PROCEDURE WriteInstRegConst (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])), 0) ;
      WriteString('>" ') ;
      InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) ;
      WriteString('" ')
   END
END WriteInstRegConst ;


(*
   WriteInstMDConst -
*)

PROCEDURE WriteInstMDConst (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString(' <') ;
      WriteCard(M2Time.InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]), 0) ;
      WriteString('>" ') ;
      InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]) ;
      WriteString('" ')
   END
END WriteInstMDConst ;


(*
   WriteSymbol - 
*)

PROCEDURE WriteSymbol (Sym: CARDINAL; Mode: ModeOfAddr) ;
BEGIN
   IF IsRegister(Sym)
   THEN
      WriteRegister(Sym, Mode)
   ELSE
      IF (Mode=ImmediateAddr) OR IsConst(Sym)
      THEN
         Write('#') ;
         IF IsConst(Sym)
         THEN
            WriteConstant(Sym)
         ELSE
            WriteAsmName(Sym)
         END
      ELSIF Mode=OffsetAddr
      THEN
         WriteAsmName(Sym)
      ELSIF Mode=ValueAddr
      THEN
         Write('[') ; WriteAsmName(Sym) ; Write(']')
      ELSE
         InternalError('incorrect mode', __FILE__, __LINE__)
      END
   END
END WriteSymbol ;


(*
   WriteRegister - 
*)

PROCEDURE WriteRegister (Sym: CARDINAL; Mode: ModeOfAddr) ;
BEGIN
   IF Mode=ValueAddr
   THEN
      Write('[') ; WriteRegisterName(Sym) ; Write(']')
   ELSIF Mode=OffsetAddr
   THEN
      WriteRegisterName(Sym)
   ELSE
      InternalError('incorrect mode for register', __FILE__, __LINE__)
   END
END WriteRegister ;


(*
   WriteRegisterName - 
*)

PROCEDURE WriteRegisterName (Sym: CARDINAL) ;
BEGIN
   IF Sym=EAX
   THEN
      WriteString('%eax')
   ELSIF Sym=EBX
   THEN
      WriteString('%ebx')
   ELSIF Sym=ECX
   THEN
      WriteString('%ecx')
   ELSIF Sym=EDX
   THEN
      WriteString('%edx')
   ELSIF Sym=ESP
   THEN
      WriteString('%esp')
   ELSIF Sym=EBP
   THEN
      WriteString('%ebp')
   ELSIF Sym=EDI
   THEN
      WriteString('%edi')
   ELSIF Sym=ESI
   THEN
      WriteString('%esi')
   ELSE
      InternalError('no such register', __FILE__, __LINE__)
   END
END WriteRegisterName ;


(*
   MakeQuadMethod
*)

PROCEDURE MakeQuadMethod (quad: CARDINAL) ;
VAR
   Op: QuadOperator;
   Op1, Op2, Op3: CARDINAL ;
BEGIN
   GetQuad(quad, Op, Op1, Op2, Op3) ;
   CASE Op OF

(*   GotoOp      : MakeCodeGotoMethod(Op, Op1, Op2, Op3) | got to here *)
   CopyOp      : MakeCodeCopyMethod(Op, Op1, Op2, Op3) |
   AddrOp      : MakeCodeAddrMethod(Op, Op1, Op2, Op3) |
   LineNumberOp: MakeCodeLineMethod(Op, Op1, Op2, Op3) |
   BecomesOp   : MakeCodeBecmMethod(Op, Op1, Op2, Op3) |
   AddOp       : MakeCodeAddMethod(Op, Op1, Op2, Op3)

   ELSE
      InternalError('unknown quadruple in case', __FILE__, __LINE__)
   END
END MakeQuadMethod ;


(*
   SaveCurrentIntoBest - saves the current assembly instructions into the best
                         list of assembly instructions.
*)

PROCEDURE SaveCurrentIntoBest (a: AliasList; BestCost, CurrentCost: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   WriteString('------------------------------------------------------') ; WriteLn ;
   WriteString('current list of assembly instructions:  (Cost=') ; WriteCard(CurrentCost, 0) ; Write(')') ; WriteLn ;
   WriteString('                                     :  (Best=') ; WriteCard(BestCost, 0) ; Write(')') ; WriteLn ;
   DisplayMethods(DoneProcedures) ; WriteLn ;
   WriteString('yield the following instructions:') ; WriteLn ;
   DumpCurrentInstructions ;
   (* save alias *)
   IF BestAlias#NIL
   THEN
      BestAlias := KillAlias(BestAlias)
   END ;
   BestAlias := DuplicateAlias(a) ;
   (* save instructions *)
   i := 0 ;
   WITH CurrentAssemblyCode DO
      i := ListPtr ;
      BestAssemblyCode.ListPtr := i ;
      WHILE i>0 DO
         BestAssemblyCode.ListOfProcedures[i] := ListOfProcedures[i] ;
         DEC(i)
      END
   END
END SaveCurrentIntoBest ;


(*
   DumpCurrentInstructions - dumps the current assembly instructions to the standard output
*)

PROCEDURE DumpCurrentInstructions ;
BEGIN
   DumpInstructions(CurrentAssemblyCode)
END DumpCurrentInstructions ;


(*
   DumpInstructions - emits all instructions in, p.
*)

PROCEDURE DumpInstructions (p: ProcedureList) ;
VAR
   i: CARDINAL ;
BEGIN
   WITH p DO
      i := 1 ;
      WHILE i<=ListPtr DO
         WITH ListOfProcedures[i].WhichParameters DO
            CASE ListOfProcedures[i].WhichProcedure OF

            instbinmd   : BinInstMD(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5]) |
            instbinreg  : BinInstReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2])) |
            instmdconst : InstMDConstant(ToInst(w[0]), ToDatum(w[1]), w[2], w[3], ToScope(w[4]), w[5], w[6]) |
            instregconst: InstRegConstant(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3])) |
            instregreg  : InstRegReg(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToReg(w[3])) |
            instregmd   : InstRegMD(ToInst(w[0]), ToDatum(w[1]), ToReg(w[2]), ToCard(w[3]),
                                    ToCard(w[4]), ToScope(w[5]), ToCard(w[6])) |
            instmdreg   : InstMDReg(ToInst(w[0]), ToDatum(w[1]), ToCard(w[2]), ToCard(w[3]),
                                    ToReg(w[4]), ToScope(w[5]), ToCard(w[6])) |
            codeline    : LineInfo(ToCard(w[0]))

            ELSE
               InternalError('case statement should only be dealing with assembly language instructions', __FILE__, __LINE__)
            END
         END ;
         INC(i)
      END
   END
END DumpInstructions ;


(*
   CodeFlushAllDirtySymbols - 
*)

PROCEDURE CodeFlushAllDirtySymbols (a: AliasList; h: CARDINAL;
                                    BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   n, i: CARDINAL ;
   al  : ARRAY [0..MaxAlias] OF AliasType ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      n := CollectAllAliases(a, al) ;
      i := 0 ;
      WHILE i<n DO
         IF NOT IsConst(al[i].Sym)
         THEN
            IF IsDirty(a, al[i].Sym, al[i].Mode) AND IsUsedOutSide(al[i].Sym)
            THEN
               IF DebugOn
               THEN
                  Write('{') ; WriteSymbol(al[i].Sym, al[i].Mode) ; WriteString(' is dirty}')
               END ;
               (*
                  symbol is dirty and it is used outside this basic block
                  - hence we must flush it to memory
               *)
               MakeFlushAllDirtySymbolsMethod ;
               MakeCleanSymbolMethod(al[i].Sym, al[i].Mode) ;
               BestCost := DoNextProcedure(a, BestCost, CurrentCost)
            END
         END ;
         INC(i)
      END
   END ;
   BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeFlushAllDirtySymbols ;


(*
   IsUsedOutSide - returns TRUE if symbol is used outside the current basic block.
*)

PROCEDURE IsUsedOutSide (Sym: CARDINAL) : BOOLEAN ;
VAR
   Father,
   Start, End: CARDINAL ;
BEGIN
   Father := GetScopeAuthor(Sym) ;
   IF IsModule(Father) AND IsExported(Father, Sym)
   THEN
      RETURN( TRUE )
   ELSE
      GetVarQuads(Sym, Start, End) ;
      IF (Start<StartQuad) OR (EndQuad<End)
      THEN
         (* is used outside *)
         RETURN( TRUE )
      ELSE
         (* not used outside *)
         IF IsTemporary(Sym)
         THEN
            (* user couldn't care less about temporaries *)
            RETURN( FALSE )
         ELSE
            (* if we are debugging then user may want to see variable anyhow (via gdb) *)
            RETURN( GenerateDebugging )
         END
      END
   END
END IsUsedOutSide ;


(*
   CodeCleanSymbol - 
*)

PROCEDURE CodeCleanSymbol (a: AliasList; h: CARDINAL;
                           BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   c, n, i: CARDINAL ;
   al     : ARRAY [0..MaxAlias] OF AliasType ;
   sym    : CARDINAL ;
   mode   : ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      sym := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      IF IsDirty(a, sym, mode)
      THEN
         c := 0 ;
         n := CollectAliases(a, al, sym, mode) ;
         i := 0 ;
         WHILE i<n DO
            IF IsClean(a, al[i].Sym, al[i].Mode) AND ((al[i].Sym#sym) OR (al[i].Mode#mode))
            THEN
               (* ok not moving (a -> a) .. *)
               INC(c) ;
               MakeSymToSymMethod(al[i].Sym, al[i].Mode, sym, mode) ;
               BestCost := DoNextProcedure(a, BestCost, CurrentCost)
            END ;
            INC(i)
         END ;
         IF c=0
         THEN
            DebugMethods(BestCost, CurrentCost) ;
            InternalError('failed to find a clean symbol in alias list', __FILE__, __LINE__)
         END
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeCleanSymbol ;


(*
   CodeAssign - updates the Alias list to the effect that an assignment has taken place.
*)

PROCEDURE CodeAssign (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   i, n,
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
   al              : ARRAY [0..MaxAlias] OF AliasType ;
   Found           : BOOLEAN ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         n := CollectAliases(a, al, from, frommode) ;
         RemoveAliasTo(a, to, tomode) ;
         AliasSymbol(a, from, frommode, to, tomode) ;
         MakeClean(a, to, tomode) ;
         (* inherit all from's aliases *)
         i := 0 ;
         WHILE i<n DO
            IF (al[i].Sym#to) OR (al[i].Mode#tomode)
            THEN
               AliasSymbol(a, al[i].Sym, al[i].Mode, to, tomode)
            END ;
            INC(i)
         END
      END
   END ;
   BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeAssign ;


(*
   CodeWillBeClean - 
*)

PROCEDURE CodeWillBeClean (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   sym : CARDINAL ;
   mode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      sym := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      MakeClean(a, sym, mode) ;
      BestCost := DoNextProcedure(a, BestCost, CurrentCost)
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeWillBeClean ;


(*
   CodeNone -
*)

PROCEDURE CodeNone (a: AliasList; h: CARDINAL;
                    BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   RETURN( DoNextProcedure(a, BestCost, CurrentCost) )
END CodeNone ;


(*
   CodeInstRegConst -
*)

PROCEDURE CodeInstRegConst (a: AliasList; h: CARDINAL;
                            BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstRegConst ;


(*
   CodeInstMDConst -
*)

PROCEDURE CodeInstMDConst (a: AliasList; h: CARDINAL;
                           BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstMDConst ;


(*
   CodeInstMDReg -
*)

PROCEDURE CodeInstMDReg (a: AliasList; h: CARDINAL;
                         BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstMDReg ;


(*
   CodeInstRegMD -
*)

PROCEDURE CodeInstRegMD (a: AliasList; h: CARDINAL;
                         BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstRegMD ;


(*
   CodeInstRegReg -
*)

PROCEDURE CodeInstRegReg (a: AliasList; h: CARDINAL;
                          BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstRegReg ;


(*
   CodeInstBinReg -
*)

PROCEDURE CodeInstBinReg (a: AliasList; h: CARDINAL;
                          BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstBinReg ;


(*
   CodeInstBinMD -
*)

PROCEDURE CodeInstBinMD (a: AliasList; h: CARDINAL;
                         BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost + TimeOf(h)) )
END CodeInstBinMD ;


(*
   CodeLine - 
*)

PROCEDURE CodeLine (a: AliasList; h: CARDINAL;
                    BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   AddToCurrentCode(h) ;   (* assembly code onto current pile *)
   RETURN( DoNextProcedure(a, BestCost, CurrentCost) )
END CodeLine ;


(*
------------------------------------------------------------------------------
   Copy Operator
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
*)

PROCEDURE CodeCopy (a: AliasList; h: CARDINAL;
                    BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   op1, op3: CARDINAL ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op3 := ToCard(w[2]) ;
      IF GetMode(op3)=OffsetAddr
      THEN
         (*
            Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
            Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
         *)
         MakeMoveFromToMethod(op3, OffsetAddr, op1, OffsetAddr) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         (*
            Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
            Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
         *)
         MakeMoveFromToMethod(op3, ValueAddr, op1, OffsetAddr) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeCopy ;


(*
   CodeMoveFromTo - should only be called via DoNextProcedure
*)
(****************************************************************
            T O O      S L O W
PROCEDURE CodeMoveFromTo (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   i, n            : CARDINAL ;
   al              : ARRAY [0..MaxAlias] OF AliasType ;
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         IF (from=to) AND (frommode=tomode)
         THEN
            (* nothing to do *)
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         ELSE
            IF IsAliased(a, from, frommode, to, tomode)
            THEN
               IF IsDirty(a, from, frommode)
               THEN
                  (*
                     choices:

                     (i)   clean up dirty symbol
                     (ii)  do nothing
                  *)
                  (* try (i) clean up from symbol *)

                  MakeCleanSymbolMethod(from, frommode) ;
                  BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;

                  IF IsAbleToBeCleaned(a, from, frommode)
                  THEN
                     (* now explore (ii) *)
                     BestCost := DoNextProcedure(a, BestCost, CurrentCost)
                  END
               ELSE
                  (* it is clean *)
                  BestCost := DoNextProcedure(a, BestCost, CurrentCost)
               END
            ELSE
               (*
                  not aliased therefore choices are:

                  (i)   alias it
                  (ii)  move it across
                  (iii) spill, to, into a previous dirty alias

                  (in fact to achieve (i) and (ii) all we need do is to alias it, make it dirty
                   and then call MoveFromTo again).

                  But before we do that we should check to see whether we should
                  copy, to, into a previous alias.
               *)
               MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;
               IF IsClean(a, to, tomode)
               THEN
                  n := CollectAliases(a, al, to, tomode) ;
                  i := 0 ;
                  WHILE i<n DO
                     IF IsDirty(a, al[i].Sym, al[i].Mode) AND
                        (NOT IsAbleToBeCleaned(a, al[i].Sym, al[i].Mode))
                     THEN
                        MakeCleanSymbolMethod(al[i].Sym, al[i].Mode)
                     END ;
                     INC(i)
                  END
               ELSE
                  MakeCleanSymbolMethod(to, tomode)
               END ;
               BestCost := DoNextProcedure(a, BestCost, CurrentCost)
            END
         END
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeMoveFromTo ;
****************************************************************)


(*
   CodeMoveFromTo - should only be called via DoNextProcedure
*)

PROCEDURE CodeMoveFromTo (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   i, n            : CARDINAL ;
   al              : ARRAY [0..MaxAlias] OF AliasType ;
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
         IF (from=to) AND (frommode=tomode)
         THEN
            (* nothing to do *)
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         ELSE
            IF IsAliased(a, from, frommode, to, tomode)
            THEN
               (* nothing to do *)
               BestCost := DoNextProcedure(a, BestCost, CurrentCost)
            ELSE
               (*
                  not aliased therefore:

                  - check to see whether "to" IsClean
                  - if so then check to see whether any dirty
                    symbol is dependent upon this symbol for cleaning
                    then
                       clean symbol
                    fi
                  - alias from -> to
               *)

               (*
                  actually we require more intelligence as we only
                  need to clean the symbol if someone uses the result
                  outside this BB or later on inside the BB - but
                  we cannot tell what will happen later on inside this
                  BB. So we will just try and if we get a non infinite
                  cost then it will work anyway...
               *)
               MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;
               BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
               IF BestCost<MaxCost
               THEN
                  (* how about that - it worked anyhow *)
               ELSE
                  WriteString('no go..') ;
                  (* no go we must clean critical symbols *)
                  MakeCleanMoveIntoMethod(from, frommode, to, tomode) ;
                  IF IsClean(a, to, tomode)
                  THEN
                     MakeDirty(a, to, tomode) ;  (* just for a test *)
                     n := CollectAliases(a, al, to, tomode) ;
                     i := 0 ;
                     WHILE i<n DO
                        IF IsDirty(a, al[i].Sym, al[i].Mode) AND
                           (NOT IsAbleToBeCleaned(a, al[i].Sym, al[i].Mode))
                        THEN
                           MakeCleanSymbolMethod(al[i].Sym, al[i].Mode)
                        END ;
                        INC(i)
                     END ;
                     MakeClean(a, to, tomode)   (* restore orig state *)
                  ELSE
                     InternalError('CodeMoveFromTo should never get here', __FILE__, __LINE__)
                  END ;
                  BestCost := DoNextProcedure(a, BestCost, CurrentCost)
               END
            END
         END
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeMoveFromTo ;


(*
   MakeCleanMoveIntoMethod - 
*)

PROCEDURE MakeCleanMoveIntoMethod (from: CARDINAL; frommode: ModeOfAddr;
                                   to  : CARDINAL; tomode  : ModeOfAddr) ;
BEGIN
   PushMethod(moveinto,
              from, frommode, to, tomode, 0, 0, 0, 0)
END MakeCleanMoveIntoMethod ;


(*
   CodeCleanMoveInto - the symbol [to:tomode] is now clean, we simply remove
                       all aliases to [to:tomode] and then alias [from:frommode]
                       to this symbol after which we make [to:tomode] dirty.
                       We then call MakeMoveFrom again.
*)

PROCEDURE CodeCleanMoveInto (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures DO
      WITH ListOfProcedures[h].WhichParameters DO
         from     := ToCard(w[0]) ;
         frommode := ToMode(w[1]) ;
         to       := ToCard(w[2]) ;
         tomode   := ToMode(w[3]) ;
(* T O O     S L O W
         MakeMoveFromToMethod(from, frommode, to, tomode) ;
*)
         RemoveAliasTo(a, to, tomode) ;   (* remove all previous aliases to [to:tomode] *)
         AliasSymbol(a, from, frommode, to, tomode) ;
         MakeDirty(a, to, tomode) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeCleanMoveInto ;


(*
   WriteCleanMoveInto -
*)

PROCEDURE WriteCleanMoveInto (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('CleanMoveInto( ') ;
      WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' -> ') ;
      WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' )')
   END
END WriteCleanMoveInto ;


(*
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym2<I>           := produces a constant
   Sym1<O> := Sym2<O>           := has the effect Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O> := Sym2<V>           := has the effect
                                             Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V> := Sym2<O>           := has the effect
                                             Mem[Mem[Sym1<I>]] := Mem[Sym2<I>]
   Sym1<V> := Sym2<V>           := has the effect
                                        Mem[Mem[Sym1<I>]] := Mem[Mem[Sym2<I>]]
*)

PROCEDURE CodeBecm (a: AliasList; h: CARDINAL;
                    BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   op1, op3: CARDINAL ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op3 := ToCard(w[2]) ;

      IF IsConst(op1) AND IsConst(op3)
      THEN
         PushValue(op3) ;
         PopValue(op1) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsConstString(op3)
      THEN
         InternalError('not yet implemented MoveStringFromTo', __FILE__, __LINE__) ;
         (* MoveStringFromTo(Operand3, Operand1) ; *)
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsProcedure(op3)
      THEN
         MakeMoveFromToMethod(op3, GetMode(op3),
                              op1, GetMode(op1)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         IF IsCoprocessorType(op1) OR IsCoprocessorType(op3)
         THEN
            InternalError('not yet implemented floating point for this optimized code generator', __FILE__, __LINE__) ;
            (* CoprocessorBecomes(Operand1, Operand3) *)
         ELSE
            MakeMoveFromToMethod(op3, GetMode(op3),
                                 op1, GetMode(op1)) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         END
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeBecm ;


(*
   Addr Operator  - contains the address of a variable.

   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<O>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<V>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<O>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
*)

PROCEDURE CodeAddr (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
BEGIN
   RETURN( 0 )
(**********************************************
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   (*
      Follow the Quadruple rules:
   *)
   WITH DoneProcedures DO
      WITH ListOfProcedures[h] DO
         IF IsConst(Operand3)
         THEN
            WriteError('Not expecting a constant')
         ELSE
            IF GetMode(Operand3)=OffsetAddr
            THEN
               (*
                  Mem[Op1] := Op3
               *)
               IF IsProcedure(GetVarFather(Operand3))
               THEN
                  InstRegReg(mov, MaxDatum(), RegEAX, RegEBP) ;
                  IF GetOffset(Operand3)<0
                  THEN
                     InstRegInt(sub, MaxDatum(), RegEAX, -GetOffset(Operand3))
                  ELSE
                     InstRegInt(add, MaxDatum(), RegEAX, GetOffset(Operand3))
                  END ;
                  InstMDReg(mov, MaxDatum(), GetOffset(Operand1), 0,
                            RegEAX, GetScope(Operand1), Operand1)
               ELSE
                  InstMDOffset(mov, MaxDatum(), Operand1, Operand3)
               END
            ELSE
               (*
                  Mem[Op1] := Mem[Op3]
               *)
               Assert(GetMode(Operand3)=ValueAddr) ;
               (*
                  Consistancy check to see whether the SIZE(Operand1)=SIZE(ADDRESS)
                  If this should ever fail then we know the compiler would produce
                  incorrect code.
                  The reason I changed PushAPU -> ADR was that PushAPU fails when
                  a record structure is passed as a VAR parameter and is altered.
                  ADR will only push the address of the record thus implementing
                  the VAR parameter correctly, however, we must make sure that
                  Operand1 is the same size as ADDRESS otherwise on a stack machine
                  all hell breaks out!
               *)
               PushSize(Operand1) ;
               Assert(PopCard()=AddressSize) ;

               MoveFromTo(Operand3, Operand1, MemoryDirect, MemoryDirect,
                          GetOffset(Operand3), GetOffset(Operand1),
                          GetSize(Operand1),
                          GetScope(Operand3), GetScope(Operand1))
            END
         END
      END
   END
***************************************************************************)
END CodeAddr ;


(*
   MakeCodeAddMethod - Briefly:
                                Operand1 := Operand2 + Operand3
                                Checks for Constant folding - otherwise a method is generated.
*)

PROCEDURE MakeCodeAddMethod (op: QuadOperator ; op1, op2, op3: CARDINAL) ;
BEGIN
   IF IsConst(op1)
   THEN
      PushValue(op3) ;
      PushValue(op2) ;
      Add ;
      PopValue(op1)
   ELSIF IsCoprocessorUsed(op1, op2, op3)
   THEN
      InternalError('not yet implemented', __FILE__, __LINE__) ;
      (* CoprocessorAdd(Operand1, Operand2, Operand3) *)
   ELSE
      PushMethod(codeadd,
                 op1, op2, op3, 0, 0, 0, 0, 0)
   END
END MakeCodeAddMethod ;


(*
   WriteCodeAdd - 
*)

PROCEDURE WriteCodeAdd (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[1]), OffsetAddr) ;
      WriteString(' + ') ; WriteSymbol(ToCard(w[2]), OffsetAddr) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), OffsetAddr) ;
      WriteString(' )')
   END
END WriteCodeAdd ;


(*
   IsAliasedToZero - 
*)

PROCEDURE IsAliasedToZero (a: AliasList; Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ;
VAR
   al: ARRAY [0..MaxAlias] OF AliasType ;
   Zero,
   n,
   i   : CARDINAL ;
BEGIN   
   Zero := MakeConstLit(MakeKey('0')) ;  (* MakeConstLit(MakeKey('0.0')) *)
   n := CollectAliases(a, al, Sym, Mode) ;
   i := 0 ;
   WHILE i<n DO
      IF al[i].Sym=Zero
      THEN
         IF al[i].Mode#ImmediateAddr
         THEN
            InternalError('constant type is not immediate ??', __FILE__, __LINE__)
         END ;
         RETURN( TRUE )
      END ;
      INC(i)
   END ;
   RETURN( FALSE )
END IsAliasedToZero ;


(*
   IsAliasedToOne - 
*)

PROCEDURE IsAliasedToOne (a: AliasList; Sym: CARDINAL; Mode: ModeOfAddr) : BOOLEAN ;
VAR
   al: ARRAY [0..MaxAlias] OF AliasType ;
   One,
   n,
   i   : CARDINAL ;
BEGIN   
   One := MakeConstLit(MakeKey('1')) ;  (* MakeConstLit(MakeKey('1.0')) *)
   n := CollectAliases(a, al, Sym, Mode) ;
   i := 0 ;
   WHILE i<n DO
      IF al[i].Sym=One
      THEN
         IF al[i].Mode#ImmediateAddr
         THEN
            InternalError('constant type is not immediate ??', __FILE__, __LINE__)
         END ;
         RETURN( TRUE )
      END ;
      INC(i)
   END ;
   RETURN( FALSE )
END IsAliasedToOne ;


(*
   OffsetUnlessConst - returns OffsetAddr unless, sym, is a constant in
                       which case ImmediateAddr is returned.
*)

PROCEDURE OffsetUnlessConst (sym: CARDINAL) : ModeOfAddr ;
BEGIN
   IF IsConst(sym)
   THEN
      RETURN( ImmediateAddr )
   ELSE
      RETURN( OffsetAddr )
   END
END OffsetUnlessConst ;



(*
   CodeAdd - symbols are referenced by Mem[Sym]
*)

PROCEDURE CodeAdd (a: AliasList; h: CARDINAL;
                   BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   op1, op2, op3: CARDINAL ;
   b1, b2, b3   : ARRAY [0..MaxAlias] OF AliasType ;
   i1, i2, i3,
   n1, n2, n3   : CARDINAL ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      op2 := ToCard(w[1]) ;
      op3 := ToCard(w[2]) ;
      IF IsAliasedToZero(a, op2, OffsetUnlessConst(op2))
      THEN
         MakeMoveFromToMethod(op3, OffsetUnlessConst(op3), op1, OffsetUnlessConst(op1)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsAliasedToZero(a, op3, OffsetUnlessConst(op3))
      THEN
         MakeMoveFromToMethod(op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsAliased(a, op2, OffsetUnlessConst(op2), op1, OffsetUnlessConst(op1))
      THEN
         MakeIncMethod(op1, OffsetUnlessConst(op1), op3, OffsetUnlessConst(op3)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsAliased(a, op3, OffsetUnlessConst(op3), op1, OffsetUnlessConst(op1))
      THEN
         MakeIncMethod(op1, OffsetUnlessConst(op1), op2, OffsetUnlessConst(op2)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         n1 := CollectAliases(a, b1, op1, OffsetUnlessConst(op1)) ;
         n2 := CollectAliases(a, b2, op2, OffsetUnlessConst(op2)) ;
         n3 := CollectAliases(a, b3, op3, OffsetUnlessConst(op3)) ;
         i1 := 0 ;
         WHILE i1<n1 DO
            i2 := 0 ;
            WHILE i2<n2 DO
               i3 := 0 ;
               WHILE i3<n3 DO
                  IF NOT IsConstLit(b1[i1].Sym)
                  THEN
                     MakePeepAddMethod(b1[i1].Sym, b1[i1].Mode, b2[i2].Sym, b2[i2].Mode, b3[i3].Sym, b3[i3].Mode) ;
                     BestCost := DoNextProcedure(a, BestCost, CurrentCost)
                  END ;
                  INC(i3)
               END ;
               INC(i2)
            END ;
            INC(i1)
         END ;
         MakePeepAddMethod(op1, OffsetUnlessConst(op1), op2, OffsetUnlessConst(op2), op3, OffsetUnlessConst(op3)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeAdd ;


(*
   SwapSymbols - 
*)

PROCEDURE SwapSymbols (VAR s1: CARDINAL; VAR m1: ModeOfAddr; VAR s2: CARDINAL; VAR m2: ModeOfAddr) ;
VAR
   t: CARDINAL ;
   n: ModeOfAddr ;
BEGIN
   t := s1 ; s1 := s2 ; s2 := t ;
   n := m1 ; m1 := m2 ; m2 := n
END SwapSymbols ;


(*
   MakePeepAddMethod - 
*)

PROCEDURE MakePeepAddMethod (op1: CARDINAL; m1: ModeOfAddr;
                             op2: CARDINAL; m2: ModeOfAddr;
                             op3: CARDINAL; m3: ModeOfAddr) ;
BEGIN
   PushMethod(peepadd,
              op1, m1, op2, m2, op3, m3, 0, 0)
END MakePeepAddMethod ;


(*
   CodePeepAdd - 
*)

PROCEDURE CodePeepAdd (a: AliasList; h: CARDINAL;
                       BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   num          : ARRAY [0..MaxNumChars] OF CHAR ;
   v            : INTEGER ;
   i, j, n,
   size,
   op1, op2, op3: CARDINAL ;
   m1, m2, m3   : ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      op3 := ToCard(w[4]) ;
      m3 := ToMode(w[5]) ;
      IF IsConst(op2) AND IsConst(op3)
      THEN
         IF IsConst(op1)
         THEN
            PushValue(op2) ;
            PushValue(op3) ;
            Add ;
            PopValue(op1) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         ELSE
            PushValue(op2) ;
            PushValue(op3) ;
            Add ;
            v := PopInt() ;
            IntToStr(v, 0, num) ;
            MakeMoveFromToMethod(MakeConstLit(MakeKey(num)), ImmediateAddr, op1, m1) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         END
      ELSE
         IF IsConst(op2)
         THEN
            size := GetSize(op3) ;
            SwapSymbols(op2, m2, op3, m3)
         ELSE
            size := GetSize(op2)
         END ;
         (*
            A number of choices:

            (i)    op2 -> op1 ; inc op1, op3
            (ii)   op3 -> op1 ; inc op1, op2
         *)

         (* (i)   op2 -> op1 ; op1 += op3 *)
         MakeIncMethod(op1, m1, op3, m3) ;
         MakeMoveFromToMethod(op2, m2, op1, m1) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;

         (* (ii)  op3 -> op1 ; op1 += op2 *)
         MakeIncMethod(op1, m1, op2, m2) ;
         MakeMoveFromToMethod(op3, m3, op1, m1) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;

         (* (iii) op3 -> r ; r += op2 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            MakeMoveFromToMethod(GetRegister(i), OffsetAddr, op1, m1) ;
            MakeIncMethod(GetRegister(i), OffsetAddr, op2, m2) ;
            MakeMoveFromToMethod(op3, m3, GetRegister(i), OffsetAddr) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
            INC(i)
         END ;
         (* (iv)  op2 -> r ; r += op3 ; r -> op1 *)
         i := 0 ;
         n := NoOfRegisters() ;
         WHILE i<n DO
            MakeMoveFromToMethod(GetRegister(i), OffsetAddr, op1, m1) ;
            MakeIncMethod(GetRegister(i), OffsetAddr, op3, m3) ;
            MakeMoveFromToMethod(op2, m2, GetRegister(i), OffsetAddr) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
            INC(i)
         END
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodePeepAdd ;


(*
   WritePeepAdd - 
*)

PROCEDURE WritePeepAdd (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' peep+ ') ; WriteSymbol(ToCard(w[4]), ToMode(w[5])) ;
      WriteString(' -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WritePeepAdd ;


(*
   WriteCodeInc - 
*)

PROCEDURE WriteCodeInc (p: ProcedureList; m: CARDINAL) ;
BEGIN
   WITH p.ListOfProcedures[m].WhichParameters DO
      WriteString('( ') ; WriteSymbol(ToCard(w[2]), ToMode(w[3])) ;
      WriteString(' += -> ') ; WriteSymbol(ToCard(w[0]), ToMode(w[1])) ;
      WriteString(' )')
   END
END WriteCodeInc ;


(*
   MakeIncMethod - 
*)

PROCEDURE MakeIncMethod (op1: CARDINAL; m1: ModeOfAddr; op2: CARDINAL; m2: ModeOfAddr) ;
BEGIN
   PushMethod(codeinc,
              op1, m1, op2, m2, 0, 0, 0, 0)
END MakeIncMethod ;


(*
   CodeInc -
*)

PROCEDURE CodeInc (a: AliasList; h: CARDINAL;
                   BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   al      : ARRAY [0..MaxAlias] OF AliasType ;
   size,
   i, n, r,
   op1, op2: CARDINAL ;
   m1, m2  : ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      op1 := ToCard(w[0]) ;
      m1 := ToMode(w[1]) ;
      op2 := ToCard(w[2]) ;
      m2 := ToMode(w[3]) ;
      IF IsConstLit(op1)
      THEN
         InternalError('should not be incrementing a constant', __FILE__, __LINE__)
      END ;
      size := GetSize(op1) ;
      IF IsAliasedToOne(a, op2, m2)
      THEN
         n := CollectAliases(a, al, op1, m1) ;
         i := 0 ;
         WHILE i<n DO
            MakeChanged(op1, m1) ;
            MakeInstBinMethod(inc, size, op1, m1) ;
            MakeCleanSymbolMethod(op1, m1) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
            INC(i)
         END
      ELSIF IsConst(op2)
      THEN
         (* add op1, #op2 *)
         MakeChanged(op1, m1) ;
         IF IsRegister(op1)
         THEN
            MakeInstRegConstantMethod(add, size, op1, op2)
         ELSE
            MakeInstMDConstantMethod(add, size, GetOffset(op1), 0, GetScope(op1), op1, op2)
         END ;
         MakeCleanSymbolMethod(op1, m1) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END ;
      i := 0 ;
      n := NoOfRegisters() ;
      WHILE i<n DO
         r := GetRegister(i) ;
         IF IsConst(op2)
         THEN
            (* (i)       op1 -> r ; add r,#op2 ; r -> op1 *)
            MakeMoveFromToMethod(r, OffsetAddr, op1, m1) ;
            MakeChanged(r, OffsetAddr) ;
            MakeInstRegConstantMethod(add, size, r, op2) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            MakeMoveFromToMethod(op1, m1, r, OffsetAddr) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         ELSE
            (* (i)       op1 -> r ; add r, op2 ; r -> op1 *)
            MakeMoveFromToMethod(r, OffsetAddr, op1, m1) ;
            MakeChanged(r, OffsetAddr) ;
            MakeInstRegMDMethod(add, size, r, GetOffset(op2), 0, GetScope(op2), op2) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            MakeMoveFromToMethod(op1, m1, r, OffsetAddr) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;

            (* (ii)      op2 -> r ; add r, op1 ; r -> op1 *)
            MakeMoveFromToMethod(r, OffsetAddr, op1, m1) ;
            MakeChanged(r, OffsetAddr) ;
            MakeInstRegMDMethod(add, size, r, GetOffset(op1), 0, GetScope(op1), op1) ;
            MakeCleanSymbolMethod(op1, m1) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            MakeMoveFromToMethod(op2, m2, r, OffsetAddr) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;

            (* (iii)     op2 -> r ; add op1, r *)
            MakeChanged(op1, m1) ;
            MakeInstMDRegMethod(add, size, GetOffset(op1), 0, GetScope(op1), op1, r) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            MakeCleanSymbolMethod(op1, m1) ;
            MakeMoveFromToMethod(op2, m2, r, OffsetAddr) ;
            MakeCleanSymbolMethod(r, OffsetAddr) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         END ;
         INC(i)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeInc ;



(*
   
         IF IsConst(op3)
         THEN
            IF IsZero(op3)
            THEN
               (* op1 := op2 *)
               MakeMoveFromToMethod(op2, m2, op1, m1) ;
               BestCost := DoNextProcedure(a, BestCost, CurrentCost)
            ELSIF IsOne(op3)
            THEN

            (* Reg := Sym1 *)
            Reg := SymToReg(Sym1) ;
            InstRegSym(Inst, Unit, Reg, Sym2) ;
            (* Reg now contains the result value *)
      ELSIF IsConst(Sym2)
      THEN
         (* Reg := Mem[Sym1] *)
         Reg := SymToReg(Sym1) ;
         (* Inst  Reg, Const *)
         InstRegConstant(Inst, Unit, Reg, Sym2) ;
         (* Reg now contains the result value *)
      ELSE
         (* Neither Sym1 or Sym2 are constants *)
         (*
            format:

                    mov   reg, mem[sym1]
                    inst  reg, mem[sym2]
         *)
         (* Reg := Mem[Sym1] *)
         Reg := SymToReg(Sym1) ;
         (* Inst  Reg, Const *)
         InstRegSym(Inst, Unit, Reg, Sym2) ;
         (* Reg now contains the result value *)
      END ;
      RETURN( Reg )
   END ;
*)


(*
   MakeChanged - indicates that the symbol, [sym:mode] has changed and
                 therefore we must remove all previous aliases to this symbol.
                 This procedure should be called when we attempto to evauluate
                 binary operations (eg. inc, dec).
*)

PROCEDURE MakeChanged (sym: CARDINAL; mode: ModeOfAddr) ;
BEGIN
   PushMethod(makechanged,
              sym, mode, 0, 0, 0, 0, 0, 0)
END MakeChanged ;


(*
   CodeMakeClean - makes a dirty symbol clean by moving one alias into it.
*)

PROCEDURE CodeMakeClean (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   al   : ARRAY [0..MaxAlias] OF AliasType ;
   i, n : CARDINAL ;
   sym  : CARDINAL ;
   mode : ModeOfAddr ;
   found: BOOLEAN ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      sym  := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      IF IsDirty(a, sym, mode)
      THEN
         n := CollectAliases(a, al, sym, mode) ;
         i := 0 ;
         found := FALSE ;
         WHILE i<n DO
            (* try all aliases and take the cheapest *)
            IF (al[i].Sym#sym) OR (al[i].Mode#mode)
            THEN
               (* ok not moving (a -> a) .. *)
               IF IsClean(a, al[i].Sym, al[i].Mode)
               THEN
                  MakeWillBeCleanMethod(sym, mode) ;
                  MakeSymToSymMethod(al[i].Sym, al[i].Mode, sym, mode) ;
                  BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
                  found := TRUE
               END
            END ;
            INC(i)
         END ;
         IF NOT found
         THEN
            IF DebugOn
            THEN
               WriteLn ;
               WriteString('aborting this line - as cannot clean symbol: ') ; WriteSymbol(sym, mode)
            END ;
            (* DebugMethods(BestCost, CurrentCost) ; *)
            (* InternalError('cannot find make a symbol clean', __FILE__, __LINE__) *)
            (* at a pinch you could just return MaxCost at this point but I believe this should never occur - no *)
         END
      ELSE
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeMakeClean ;


(*
   CodeMakeChanged - notes that a symbol has changed and therefore removes any aliases to it.
*)

PROCEDURE CodeMakeChanged (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   sym : CARDINAL ;
   mode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      sym  := ToCard(w[0]) ;
      mode := ToMode(w[1]) ;
      RemoveAliasTo(a, sym, mode) ;
      MakeWillBeCleanMethod(sym, mode) ;
      BestCost := DoNextProcedure(a, BestCost, CurrentCost)
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeMakeChanged ;


(*
   CodeSymToSym - should only be called via DoNextProcedure
*)

PROCEDURE CodeSymToSym (a: AliasList; h: CARDINAL;
                        BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   from, to        : CARDINAL ;
   frommode, tomode: ModeOfAddr ;
BEGIN
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to   := ToCard(w[2]) ;
      tomode := ToMode(w[3]) ;
      MakeAssign(from, frommode, to, tomode) ;
      IF FitsIntoRegister(from) AND FitsIntoRegister(to)
      THEN
         RETURN( SymToSymMachineWord(a, h, BestCost, CurrentCost) )
      ELSE
         InternalError('SymToSymBlock not written yet', __FILE__, __LINE__) ;
         (* RETURN( SymToSymBlock(a, h, BestCost, CurrentCost) ) *)
      END
   END
END CodeSymToSym ;


(*
   SymToSymMachineWord - 
*)

PROCEDURE SymToSymMachineWord (a: AliasList; h: CARDINAL;
                               BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   to, from        : CARDINAL ;
   tomode, frommode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   (* really does do the move, providing not a clean alias *)
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      IF IsAliased(a, from, frommode, to, tomode) AND IsClean(a, to, tomode)
      THEN
         (* ignore as it is clean aliased *)
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         IF IsConst(from)
         THEN
            (* just see whether it is cheaper to move constant directly into memory with using a register *)
            MakeInstMDConstantMethod(mov, GetSize(to), GetOffset(to), 0, GetScope(to), to, from) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         END ;
         (* now do the move *)
         MakeForeachRegisterDoSymToSym(from, frommode, to, tomode) ;
         IF DebugOn
         THEN
            WriteString('AFTER FOREACH REGISTER') ; WriteLn ;
            DebugMethods(BestCost, CurrentCost) ;
         END ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
         IF DebugOn
         THEN
            WriteString('FINISHED FOREACH REGISTER') ; WriteLn ;
            DebugMethods(BestCost, CurrentCost) ;
         END
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END SymToSymMachineWord ;


(*
   MakeForeachRegisterDoSymToSym - 
*)

PROCEDURE MakeForeachRegisterDoSymToSym (from: CARDINAL; frommode: ModeOfAddr;
                                         to: CARDINAL; tomode: ModeOfAddr) ;
BEGIN
   PushMethod(forregsym,
              from, frommode, to, tomode, 0, 0, 0, 0)
END MakeForeachRegisterDoSymToSym ;


(*
   CodeForeachRegisterDoSymToSym - 
*)

PROCEDURE CodeForeachRegisterDoSymToSym (a: AliasList; h: CARDINAL;
                                         BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   n, i            : CARDINAL ;
   to, from        : CARDINAL ;
   tomode, frommode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   n := NoOfRegisters() ;
   i := 0 ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      WHILE i<n DO
         MakeMoveThroughRegisterMethod(from, frommode, to, tomode, GetRegister(i)) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
         INC(i)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeForeachRegisterDoSymToSym ;


(*
   CodeMoveThroughRegister - 
*)

PROCEDURE CodeMoveThroughRegister (a: AliasList; h: CARDINAL;
                                   BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   to,  from: CARDINAL;
   tomode,
   frommode : ModeOfAddr;
   reg      : CARDINAL ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      tomode   := ToMode(w[3]) ;
      reg      := ToCard(w[4]) ;
      MakeRegToSymMethod(reg, to, tomode) ;      (* (ii) *)
      MakeSymToRegMethod(from, frommode, reg) ;  (* (i)  *)
      BestCost := DoNextProcedure(a, BestCost, CurrentCost) ;
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeMoveThroughRegister ;


(*
   MakeMoveThroughRegisterMethod - 
*)

PROCEDURE MakeMoveThroughRegisterMethod (from: CARDINAL; frommode: ModeOfAddr;
                                         to  : CARDINAL; tomode  : ModeOfAddr;
                                         reg: CARDINAL) ;
BEGIN
   PushMethod(moveviareg,
              from, frommode, to, tomode, reg, 0, 0, 0)
END MakeMoveThroughRegisterMethod ;


(*
   CodeSymToReg - 
*)

PROCEDURE CodeSymToReg (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   to,  from: CARDINAL ;
   frommode : ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      to       := ToCard(w[2]) ;
      IF IsAliased(a, from, frommode, to, OffsetAddr) AND IsClean(a, to, OffsetAddr)
      THEN
         (* ignore as it is clean aliased *)
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsRegister(from)
      THEN
         (* maybe this should be size of symbol? *)
         MakeAssign(from, frommode, to, OffsetAddr) ;
         MakeRegToRegMethod(from, to, WordSize) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         MakeAssign(from, frommode, to, OffsetAddr) ;
         MakeExtToRegMethod(from, frommode, to) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeSymToReg ;


(*
   CodeExtToReg - places a symbol Sym into a spare register.
                  If the Sym is Value or Offset then reg contains
                  Mem[Sym]
                  CASE Mode(Sym) OF

                  Const:  Reg := Constant(Sym) |
                  Value,
                  Offset: Reg := Mem[Sym]

                  END
*)

PROCEDURE CodeExtToReg (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   Size       : CARDINAL ;
   toreg, from: CARDINAL;
   frommode   : ModeOfAddr;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      frommode := ToMode(w[1]) ;
      toreg    := ToCard(w[2]) ;
      IF IsConstString(from)
      THEN
         Size := GetStringLength(from) ;
         IF Size=1
         THEN
            MakeAssign(from, frommode, toreg, OffsetAddr) ;
            MakeInstRegConstantMethod(mov, WordSize, toreg, from) ;
            BestCost := DoNextProcedure(a, BestCost, CurrentCost)
         ELSE
            InternalError('cannot place a string with more than 1 character in it into a register', __FILE__, __LINE__)
         END
      ELSIF IsConst(from)
      THEN
         MakeAssign(from, frommode, toreg, OffsetAddr) ;
         MakeInstRegConstantMethod(mov, WordSize, toreg, from) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         Size := GetSize(from) ;
         Assert(Size<=WordSize) ;
         MakeAssign(from, frommode, toreg, OffsetAddr) ;
         MakeInstRegMDMethod(mov, Size, toreg, GetOffset(from), 0, GetScope(from), from) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeExtToReg ;


(*
   CodeRegToSym - 
*)

PROCEDURE CodeRegToSym (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   from, to: CARDINAL ;
   tomode  : ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      to       := ToCard(w[1]) ;
      tomode   := ToMode(w[2]) ;
      IF IsAliased(a, from, OffsetAddr, to, tomode) AND IsClean(a, to, tomode)
      THEN
         (* ignore as it is clean aliased *)
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSIF IsRegister(to)
      THEN
         MakeAssign(from, OffsetAddr, to, tomode) ;
         MakeRegToRegMethod(from, to, WordSize) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         MakeAssign(from, OffsetAddr, to, tomode) ;
         MakeRegToExtMethod(from, to, tomode) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeRegToSym ;


(*
   CodeRegToExt - places register contents into a symbol Sym.
                  If the Sym is Value or Offset then reg is placed into Mem[Sym]
                  CASE Mode(Sym) OF

                  Const:  Error |
                  Value,
                  Offset: Mem[Sym] := Reg

                  END
*)

PROCEDURE CodeRegToExt (a: AliasList; h: CARDINAL;
                        BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   Size     : CARDINAL ;
   to,  from: CARDINAL;
   tomode,
   frommode : ModeOfAddr;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      to       := ToCard(w[1]) ;
      tomode   := ToMode(w[2]) ;
      IF IsConst(to)
      THEN
         InternalError('Cannot place a register into a constant', __FILE__, __LINE__)
      ELSE
         Size := GetSize(to) ;
         Assert(Size<=WordSize) ;
         (* Correct since we ignore ValueAddr mode and treat as OffsetAddr *)
         MakeAssign(from, OffsetAddr, to, tomode) ;
         MakeInstMDRegMethod(mov, Size, GetOffset(to), 0, GetScope(to), to, from) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeRegToExt ;


(*
   CodeRegToReg - 
*)

PROCEDURE CodeRegToReg (a: AliasList; h: CARDINAL; BestCost, CurrentCost: CARDINAL) : CARDINAL ;
VAR
   size            : CARDINAL ;
   from, to        : CARDINAL ;
   tomode, frommode: ModeOfAddr ;
BEGIN
   IncrementLevel ;
   WITH DoneProcedures.ListOfProcedures[h].WhichParameters DO
      from     := ToCard(w[0]) ;
      to       := ToCard(w[1]) ;
      size     := ToCard(w[2]) ;
      IF IsAliased(a, from, OffsetAddr, to, OffsetAddr) AND IsClean(a, to, OffsetAddr)
      THEN
         (* ignore as it is clean aliased *)
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      ELSE
         MakeAssign(from, OffsetAddr, to, OffsetAddr) ;
         MakeInstRegRegMethod(mov, size, from, to) ;
         BestCost := DoNextProcedure(a, BestCost, CurrentCost)
      END
   END ;
   DecrementLevel ;
   RETURN( BestCost )
END CodeRegToReg ;


(*
   NoOfRegisters - returns the number of registers for this processor.
*)

PROCEDURE NoOfRegisters () : CARDINAL ;
BEGIN
   RETURN( 2 )  (* should be 5 - (2 for testing) *)
END NoOfRegisters ;


(*
   GetRegister - returns the symbol representing register, i.
*)

PROCEDURE GetRegister (i: CARDINAL) : CARDINAL ;
BEGIN
   CASE i OF

   0:  RETURN( EAX ) |
   1:  RETURN( EBX ) |
   2:  RETURN( ECX ) |
   3:  RETURN( EDX ) |
   4:  RETURN( ESI ) |
   5:  RETURN( EDI )

   ELSE
      InternalError('this register does not exist', __FILE__, __LINE__)
   END
END GetRegister ;


(*
   IsRegister - 
*)

PROCEDURE IsRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (Sym=EAX) OR (Sym=EBX) OR (Sym=ECX) OR (Sym=EDX) OR
          (Sym=ESP) OR (Sym=EBP) OR (Sym=ESI) OR (Sym=EDI)
         )
END IsRegister ;


(*
   Init - 
*)

PROCEDURE Init ;
VAR
   System: CARDINAL ;
BEGIN
   (* use SYSTEM module to create our registers *)
   System := MakeDefinitionSource(MakeKey('SYSTEM')) ;
   StartScope(System) ;

   EAX := MakeTemporary(OffsetAddr) ; PutVar(EAX, Word) ;
   EBX := MakeTemporary(OffsetAddr) ; PutVar(EBX, Word) ;
   ECX := MakeTemporary(OffsetAddr) ; PutVar(ECX, Word) ;
   EDX := MakeTemporary(OffsetAddr) ; PutVar(EDX, Word) ;
   ESP := MakeTemporary(OffsetAddr) ; PutVar(ESP, Word) ;
   EBP := MakeTemporary(OffsetAddr) ; PutVar(EBP, Word) ;
   ESI := MakeTemporary(OffsetAddr) ; PutVar(ESI, Word) ;
   EDI := MakeTemporary(OffsetAddr) ; PutVar(EDI, Word) ;

   EndScope

END Init ;


BEGIN
   Init
END M2GenDyn486.
