IMPLEMENTATION MODULE M2Options ;


IMPORT CmdArgs ;
FROM FIO IMPORT Exists ;
FROM Args IMPORT GetArg, Narg ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StrLib IMPORT StrEqual, StrLen, StrCopy ;
FROM M2Defaults IMPORT GetOptions ;
FROM M2Search IMPORT SetSearchPath ;
FROM M2Version IMPORT WriteVersion ;
FROM libc IMPORT exit ;
FROM NumberIO IMPORT StrToCard ;

FROM M2Quads IMPORT BuildCodeOn, BuildCodeOff,
                    BuildOptimizeOn, BuildOptimizeOff,
                    BuildProfileOn, BuildProfileOff ;



(* %%%FORWARD%%%
PROCEDURE WriteName (Name: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE DisplayVersion ; FORWARD ;
PROCEDURE ScanEnvironmentArgs ; FORWARD ;
   %%%FORWARD%%% *)


CONST
   MaxStringLen = 1000 ;


VAR
   AssemblerFile: ARRAY [0..MaxStringLen] OF CHAR ;


(*
   DisplayVersion - displays the version of the compiler.
*)

PROCEDURE DisplayVersion ;
BEGIN
   WriteString('gm2  [') ;
   WriteVersion ;
   WriteString('] [') ;
   WriteString(__DATE__) ;
   WriteString(']') ; WriteLn
END DisplayVersion ;


(*
   SetAssemblerFile - sets the global variable AssemblerFile to a.
*)

PROCEDURE SetAssemblerFile (a: ARRAY OF CHAR) ;
BEGIN
   StrCopy(a, AssemblerFile) ;
   AssemblerFileSpecified := TRUE ;
   IF Exists(a)
   THEN
      WriteString('destination file: ') ; WriteString(a) ;
      WriteString(' already exists') ; WriteLn ;
      exit(1)
   END
END SetAssemblerFile ;


(*
   GetAssemblerFileName - assigns, a, to the assembler file
                          specified.
*)

PROCEDURE GetAssemblerFileName (VAR a: ARRAY OF CHAR) ;
BEGIN
   StrCopy(AssemblerFile, a) ;
END GetAssemblerFileName ;


(*
   ParseOptions - parses the options and sets the option flags
                  accordingly.
*)

PROCEDURE ParseOptions ;
VAR
   a   : ARRAY [0..MaxStringLen] OF CHAR ;
   i, n: CARDINAL ;
BEGIN
   n := Narg() ;
   IF n=1
   THEN
      DisplayOptions
   END ;
   i := 1 ;
   REPEAT
      IF GetArg(a, i)
      THEN
         (* WriteString('Argument found: ') ; WriteString(a) ; WriteLn ; *)
         IF StrEqual(a, '-M')
         THEN
            INC(i) ;
            IF GetArg(a, i)
            THEN
               SetSearchPath(a)
            ELSE
               WriteString('Expecting path after -M flag') ; WriteLn
            END
         ELSIF StrEqual(a, '-version')
         THEN
            DisplayVersion
         ELSIF StrEqual(a, '-o')
         THEN
            INC(i) ;
            IF GetArg(a, i)
            THEN
               SetAssemblerFile(a)
            ELSE
               WriteString('Expecting assembler file after -o') ; WriteLn
            END
         ELSIF NOT IsAnOption(a)
         THEN
            (*
               do nothing - note that we may have set some flags by
               calling IsAnOption.
            *)
         END
      END ;
      INC(i)
   UNTIL i>n
END ParseOptions ;


(*
   ScanForInitialOptions - parses the options and sets any initial
                           options immediately.
*)

PROCEDURE ScanForInitialOptions ;
VAR
   a   : ARRAY [0..MaxStringLen] OF CHAR ;
   i, n: CARDINAL ;
BEGIN
   n := Narg() ;
   i := 1 ;
   REPEAT
      IF GetArg(a, i)
      THEN
         (* WriteString('Argument found: ') ; WriteString(a) ; WriteLn ; *)
         IF StrEqual(a, '-M')
         THEN
            INC(i) ;
         ELSIF NOT IsAnOption(a)
         THEN
            (*
               do nothing - note that we may have set some flags by
               calling IsAnOption.
            *)
         END
      END ;
      INC(i)
   UNTIL i>n
END ScanForInitialOptions ;


(*
   IsAnOption - Returns true if a legal option was specified in a.
                If legal then the appropriate flag was set.
*)

PROCEDURE IsAnOption (a: ARRAY OF CHAR) : BOOLEAN ;
VAR
   Legal: BOOLEAN ;
BEGIN
   IF StrEqual(a, '-d')
   THEN
      CompilerDebugging := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-g')
   THEN
      GenerateDebugging := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-ng')
   THEN
      GenerateDebugging := FALSE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-gline')
   THEN
      GenerateLineDebug := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-O')
   THEN
      DefaultOptimizing := TRUE ;
      Optimizing := TRUE ;
      OptimizeUncalledProcedures := TRUE ;
      OptimizeBasicBlock := TRUE ;
      OptimizeCommonSubExpressions := TRUE ;
      OptimizeDynamic := FALSE ;  (* TRUE when it works.. *)
      Legal := TRUE
   ELSIF StrEqual(a, '-Obb')
   THEN
      OptimizeBasicBlock := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-Ouncalled')
   THEN
      OptimizeUncalledProcedures := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-Omemcopy')
   THEN
      APUUseMemCopy := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-Odynamic')
   THEN
      WriteString('warning dynamic code generation is currently under development (it is very likely broken)') ; WriteLn ;
      OptimizeDynamic := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-Ocse')
   THEN
      OptimizeCommonSubExpressions := FALSE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-bounds')
   THEN
      BoundsChecking := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-nomath')
   THEN
      MathCoprocessor := FALSE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-p')
   THEN
      DefaultProfiling := FALSE ;
      Profiling := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-c')
   THEN
      DefaultCoding := FALSE ;
      Coding := FALSE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-32')
   THEN
      WordLength := 32 ;
      Legal := TRUE
   ELSIF StrEqual(a, '-16')
   THEN
      WordLength := 16 ;
      Legal := TRUE
   ELSIF StrEqual(a, '-statistics')
   THEN
      Statistics := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-pedantic')
   THEN
      Pedantic := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-verbose')
   THEN
      Verbose := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-quiet')
   THEN
      Quiet := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-q')
   THEN
      DisplayQuadruples := TRUE ;
      Legal := TRUE
   ELSIF StrEqual(a, '-apu')
   THEN
      APUCodeGenerator := TRUE ;
      Legal := TRUE
   ELSIF (a[0]='-') AND (a[1]='O')
   THEN
      a[0] := ' ' ;
      a[1] := ' ' ;
      StrToCard(a, OptimizeLength) ;
      Legal := TRUE
   ELSE
      Legal := FALSE
   END ;
   RETURN( Legal )
END IsAnOption ;


(*
   DisplayOptions - displays a string of legal options and a brief
                    explanation about each option.
*)

PROCEDURE DisplayOptions ;
BEGIN
   WriteString('       -g                  generate runtime debugging information') ; WriteLn ;
   WriteString('       -O                  switch on all optimizing during code generation') ; WriteLn ;
   WriteString('       -Ouncalled          remove all uncalled procedures') ; WriteLn ;
(* WriteString('       -Odynamic           the dynamic code generator (do not use)') ; WriteLn ; *)
   WriteString('       -Omemcopy           use the APU memcopy instruction (only used with -apu)') ; WriteLn ;
   WriteString('       -O<n>               number of quadruples to dynamically optimize') ; WriteLn ;
   WriteString('       -Ocse               turn off common subexpression elimination') ; WriteLn ;
   WriteString('       -bounds             do generate runtime array subscript and subrange checks') ; WriteLn ;
   WriteString('       -nomath             do not generate any math coprocessor code') ; WriteLn ;
   WriteString('       -32                 generate code using 32 bit data') ; WriteLn ;
   WriteString('       -16                 generate code using 16 bit data') ; WriteLn ;
   WriteString('       -q                  display internal quadruples') ; WriteLn ;
   WriteString('       -verbose            generate verbose error messages') ; WriteLn ;
   WriteString('       -quiet              do not display any pass information') ; WriteLn ;
   WriteString('       -statistics         generate statistics about code generation') ; WriteLn ;
   WriteString('       -pedantic           check for any pedantic errors') ; WriteLn ;
   WriteString('       -M                  the next argument is the search path') ; WriteLn ;
   WriteString('                           multiple paths are specified by ". ../this ../that"') ; WriteLn
END DisplayOptions ;


(*
   QuietString - writes a string if the boolean, Quiet, is FALSE.
*)

PROCEDURE QuietString (a: ARRAY OF CHAR) ;
BEGIN
   IF NOT Quiet
   THEN
      WriteString(a)
   END
END QuietString ;


(*
   QuietLn - writes a string if the boolean Quiet is FALSE.
*)

PROCEDURE QuietLn ;
BEGIN
   IF NOT Quiet
   THEN
      WriteLn
   END
END QuietLn ;


(*
   QuietName - displays a name, Name, with formatted spaces
               after the string, providing that the -quiet
      	       option was not specified.
*)

PROCEDURE QuietName (Name: ARRAY OF CHAR) ;
BEGIN
   IF NOT Quiet
   THEN
      WriteName(Name)
   END
END QuietName ;


(*
   WriteName - displays a name, Name, with formatted spaces
               after the string.
*)

PROCEDURE WriteName (Name: ARRAY OF CHAR) ;
CONST
   MaxSpaces = 20 ;
VAR
   i, High: CARDINAL ;
BEGIN
   WriteString(Name) ;
   High := StrLen(Name) ;
   IF High<MaxSpaces
   THEN
      High := MaxSpaces - High ;
      i := 0 ;
      WHILE i<High DO
         Write(' ') ;
         INC(i)
      END
   END
END WriteName ;


BEGIN
   WordLength                   := 32 ;
   APUCodeGenerator             := FALSE ;
   Statistics                   := FALSE ;
   CompilerDebugging            := FALSE ;
   GenerateDebugging            := FALSE ;
   Optimizing                   := FALSE ;
   Coding                       := TRUE ;
   Pedantic                     := FALSE ;
   Verbose                      := FALSE ;
   Quiet                        := FALSE ;
   Profiling                    := FALSE ;
   DisplayQuadruples            := FALSE ;
   DefaultCoding                := TRUE ;
   DefaultProfiling             := TRUE ;
   DefaultOptimizing            := FALSE ;
   OptimizeBasicBlock           := FALSE ;
   OptimizeUncalledProcedures   := FALSE ;
   OptimizeCommonSubExpressions := FALSE ;
   BoundsChecking               := FALSE ;
   MathCoprocessor              := TRUE ;
   AssemblerFileSpecified       := FALSE ;
   APUUseMemCopy                := FALSE ;
   OptimizeLength               := 1 ;
   StrCopy('', AssemblerFile) ;
   ScanForInitialOptions
END M2Options.
