IMPLEMENTATION MODULE i486 ;


FROM SymbolTable IMPORT IsConst, IsProcedure, IsVar, GetType,
                        PushSize, PushVarSize, PushOffset,
                        GetFirstUsed,
                        GetVarFather ;

FROM M2Options IMPORT WordLength ;
FROM M2ALU IMPORT PopInt, PopCard ;
FROM M2System IMPORT WordSize ;
FROM M2Lexical IMPORT WriteError, NearToken ;


VAR
   MaxDatumValue: DatumPrefix ;


(*
   MaxDatum - returns the maximum size of data that can be contained in
              a register.
*)

PROCEDURE MaxDatum () : DatumPrefix ;
BEGIN
   RETURN( MaxDatumValue )
END MaxDatum ;


(*
   GetScope - returns the scope of Sym.
              The scope is either Local or Global.
*)

PROCEDURE GetScope (Sym: CARDINAL) : KindOfScope ;
BEGIN
   IF IsConst(Sym) OR IsProcedure(Sym)
   THEN
      RETURN( NoScope )
   ELSIF IsProcedure(GetVarFather(Sym))
   THEN
      RETURN( Local )
   ELSE
      RETURN( Global )
   END
END GetScope ;


(*
   GetOffset - returns the offset for a symbol, Sym.
*)

PROCEDURE GetOffset (Sym: CARDINAL) : INTEGER ;
BEGIN
   IF IsConst(Sym) OR IsProcedure(Sym)
   THEN
      RETURN( 0 )
   ELSE
      PushOffset(Sym) ;
      RETURN( PopInt() )
   END
END GetOffset ;


(*
   GetSize - returns the size of a symbol, Sym.
             returns the amount of space variable, Sym, requires.
             NOTE this is not the same as GetVarTypeSize since
                  if Sym is ValueAddr then the size will be
                  TSIZE(ADDRESS) since Sym is in fact a "pointer".
                  ie VAR parameter...
*)

PROCEDURE GetSize (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF IsVar(Sym)
   THEN
      PushVarSize(Sym)
   ELSE
      PushSize(Sym)
   END ;
   RETURN( PopCard() )
END GetSize ;


(*
   GetUnitPrefix - returns the prefix mode which corresponds to size.
*)

PROCEDURE GetUnitPrefix (Size: CARDINAL) : DatumPrefix ;
BEGIN
   IF WordLength=32
   THEN
      CASE Size OF

      BYTESIZE : RETURN( PByte ) |
      WORDSIZE : RETURN( PWord ) |
      DWORDSIZE: RETURN( PDWord )

      ELSE
         WriteError('No prefix for this size')
      END
   ELSE
      CASE Size OF

      BYTESIZE : RETURN( PByte ) |
      WORDSIZE : RETURN( PWord )

      ELSE
         WriteError('No prefix for this size using 16 bit code generation')
      END
   END
END GetUnitPrefix ;


(*
   GetVarTypeSize - returns the size of a variable symbol, Sym, type.
*)

PROCEDURE GetVarTypeSize (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF IsVar(Sym)
   THEN
      RETURN( GetSize(GetType(Sym)) )
   ELSE
      NearToken('Expecting Sym to be a variable', GetFirstUsed(Sym))
   END
END GetVarTypeSize ;


(*
   EqualSizes - returns true if the two symbols have the same size.
*)

PROCEDURE EqualSizes (Sym1, Sym2: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym1) OR IsConst(Sym2)
   THEN
      RETURN( TRUE )   (* Constants can be compared *)
   ELSE
      RETURN( GetVarTypeSize(Sym1)=GetVarTypeSize(Sym2) )
   END
END EqualSizes ;


(*
   FitsInRegister - returns true if Sym will fit within a register.
*)

PROCEDURE FitsInRegister (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF IsConst(Sym)
   THEN
      RETURN( TRUE )   (* Constants will match any register size *)
   ELSE
      RETURN( GetSize(Sym) <= WordSize )
   END
END FitsInRegister ;


(*
   Init - initializes global module data.
*)

PROCEDURE Init ;
BEGIN
   IF WordLength=32
   THEN
      MaxDatumValue := PDWord
   ELSE
      MaxDatumValue := PWord
   END
END Init ;


BEGIN
   Init
END i486.
