IMPLEMENTATION MODULE FIO ;

(*
 *	FIO Module.
 *
 *	WARNING: Only make changes to the files inside the Src/ directory.
 *
 ****** History:
 *
 *	1.0	3rd August 1987   	Wrote implementation part
 *	1.1	27th August 1987   	Added IOError, IOStatus() function
 *	1.2	28th August 1987   	Sue saw it :-) back to drawing board
 *	1.3	4th September 1987	Wrote CIsupport... called it from here
 *	1.4	23rd Sept 1987		Added UnReadChar() for symmetry
 *	1.5	dcw	1st Mar 1988   	Replaced IllegalOp with immediate errors
 *	1.5.1	dcw	3rd Mar 1988   	Altered close so not fatal if not open
 *	1.5.2	dcw	15th July 1988	CI.unixgetchar() gives INTEGER now
 *	1.6	dcw	ditto		Added NewLineCh for input testing
 *					[augments EOLN() function]
 *					altered ReadLine to use this..
 *	1.7	dcw	24th Aug 1988	all errors to StdErr. Added
 *					'checkopenfor', removed checkopen
 *					close aborts [again] if file not open
 *					changed some access-modes in defn.
 *	1.8	dcw	30th Aug 1988	char->int in CIsupport.unixputc
 *					caused me to have to change the two
 *					calls to it. In particular, changed
 *					writeb() a bit.. elsif -> else...
 *	1.8.1	dcw	12th Sep 1988	chr(i) -> chr(ord(i)) cos arg=cardinal
 *	1.8.2	dcw	10th Oct 1988	Removed LONGREAL crap..
 *)


FROM SYSTEM IMPORT ADR, TSIZE, BYTE ;
FROM ASCII IMPORT nl, nul ;
FROM Math IMPORT ABS ;
FROM StrLib IMPORT StrLen ;

FROM CIsupport IMPORT
	unixexit, unixfseek, unixftell,
	unixfopen, unixfclose,
	unixfread, unixfwrite,
	unixgetc, unixungetc, unixputc,
(*
	unixfreadcardinal, unixfwritecardinal,
	unixfreadinteger, unixfwriteinteger,
	unixfreadreal, unixfwritereal,
*)
	unixfreadstr, unixfwritestr;


CONST
   Maxfiles      = 20;
   NewLineCh     = 12C;
   MaxBufferSize = 8192 ;
   MaxNameSize   = 80 ;

TYPE
   (* File is the index into files array *)

   FileType = ( none, reading, writing, random );
   FCB       = RECORD
                  type     : FileType;
                  name     : ARRAY [0..MaxNameSize] OF CHAR;
                  filedes  : INTEGER;		(* file descriptor *)
                  ioresult : Status;
               END ;
   FileArray = ARRAY [-1..Maxfiles] OF FCB;


(* NOTE: element
 *	 -1 is for open's which fail
 *       0 is stdin
 *       1 is stdout
 *       2 is stderr
 *)

VAR
	files    : FileArray;
	modename : ARRAY FileType OF ARRAY [0..30] OF CHAR;

(* %%%FORWARD%%%
PROCEDURE IOStatus( Id : File ) : Status; FORWARD ;
PROCEDURE ReportError( Id : File ); FORWARD ;
PROCEDURE IsNoError (Id: File) : BOOLEAN ; FORWARD ;
PROCEDURE Exists (fname: ARRAY OF CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE OpenToRead( fname : ARRAY OF CHAR ) : File ; FORWARD ;
PROCEDURE OpenToWrite( fname : ARRAY OF CHAR ) : File ; FORWARD ;
PROCEDURE OpenForRandom( fname : ARRAY OF CHAR ; Create : BOOLEAN ) : File ; FORWARD ;
PROCEDURE Close( Id : File ) ; FORWARD ;
PROCEDURE EOF( Id : File ) : BOOLEAN ; FORWARD ;
PROCEDURE EOLN( Id : File ) : BOOLEAN ; FORWARD ;
PROCEDURE SetPosition( Id : File; Pos : CARDINAL ; End : EndType ) ; FORWARD ;
PROCEDURE FindPosition( Id : File ) : CARDINAL; FORWARD ;
PROCEDURE Rewind( Id : File ) ; FORWARD ;
PROCEDURE ReadNBytes( Id : File; n : CARDINAL; buffer : ADDRESS ) ; FORWARD ;
PROCEDURE WriteNBytes( Id : File; n : CARDINAL; buffer : ADDRESS ) ; FORWARD ;
PROCEDURE WriteBase( Id      : File;
		     base, num : CARDINAL;
		     width     : INTEGER ); FORWARD ;
PROCEDURE ReadCardinal( Id : File ) : CARDINAL ; FORWARD ;
PROCEDURE WriteCardinal( Id : File; object : CARDINAL ) ; FORWARD ;
PROCEDURE ReadInteger( Id : File ) : INTEGER ; FORWARD ;
PROCEDURE WriteInteger( Id : File; object : INTEGER ) ; FORWARD ;
PROCEDURE ReadReal( Id : File ) : REAL ; FORWARD ;
PROCEDURE WriteReal( Id : File; object : REAL ) ; FORWARD ;
PROCEDURE ReadChar( Id : File ) : CHAR ; FORWARD ;
PROCEDURE WriteChar( Id : File; Object : CHAR ) ; FORWARD ;
PROCEDURE ReadString( Id : File; VAR Object: ARRAY OF CHAR ) ; FORWARD ;
PROCEDURE WriteString( Id : File; Object : ARRAY OF CHAR ) ; FORWARD ;
PROCEDURE ReadLine( Id : File ) ; FORWARD ;
PROCEDURE WriteLine( Id : File ) ; FORWARD ;
PROCEDURE UnReadChar( Id : File ; c : CHAR ) ; FORWARD ;
PROCEDURE len( s : ARRAY OF CHAR ) : INTEGER; FORWARD ;
PROCEDURE writeb(     f      : INTEGER;
		      b      : CARDINAL;
		      n      : CARDINAL;
		  VAR spaces : BOOLEAN;
		      digits : INTEGER ) : BOOLEAN; FORWARD ;
PROCEDURE errstr( s : ARRAY OF CHAR ); FORWARD ;
PROCEDURE errline; FORWARD ;
PROCEDURE openfile(	fname : ARRAY OF CHAR;
			fmode : ARRAY OF CHAR;
			ftype : FileType;
			fstat : Status ) : File ; FORWARD ;
PROCEDURE findfree() : File; FORWARD ;
PROCEDURE checkopenfor( procname : ARRAY OF CHAR;
			file : File;
			mustbe : BOOLEAN;
			mode : FileType ); FORWARD ;
   %%%FORWARD%%% *)




(*
   GetUnixFileDescriptor - returns the UNIX file descriptor of a file.
*)

PROCEDURE GetUnixFileDescriptor (f: File) : INTEGER ;
BEGIN
   RETURN( files[f].filedes )
END GetUnixFileDescriptor ;


(*
   Max - returns the maximum of two values.
*)

PROCEDURE Max (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a>b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Max ;


(*
   Min - returns the minimum of two values.
*)

PROCEDURE Min (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a<b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Min ;
      

PROCEDURE IOStatus( Id : File ) : Status;
BEGIN
	IF ( Id < -1 ) OR ( Id > Maxfiles ) THEN
		ReportError( Id );
	ELSE
		RETURN files[ Id ].ioresult;
	END
END IOStatus;


PROCEDURE ReportError( Id : File );
BEGIN
	IF ( Id < -1 ) OR (Id > Maxfiles ) THEN Id := -1 END;

	CASE files[ Id ].ioresult OF

	NoError:
		RETURN				(* only way of returning *)

	| UnInitialized :
		errstr( "File being used when not open - unitialized" );

	| CantOpenFile :
		errstr( "File '" );
		errstr( files[Id].name ) ;
		errstr( "' does not exist, or cannot be opened" );

	| CantCreateFile :
		errstr( "File '" );
		errstr( files[Id].name ) ;
		errstr( "' cannot be created" );

	| TooManyOpen :
		errstr( "Can't open file '" );
		errstr( files[Id].name ) ;
		errstr( "' too many files open already" );

	| OpFailed :
		errstr( "Operation on file '" );
		errstr( files[ Id ].name );
		errstr( "' failed" ) ;
	END;
	errline;
	unixexit(1);
END ReportError;


PROCEDURE IsNoError (Id: File) : BOOLEAN ;
BEGIN
   IF Id=-1
   THEN
      RETURN( FALSE )
   ELSE
      RETURN( files[Id].ioresult = NoError )
   END
END IsNoError ;


PROCEDURE Exists (fname: ARRAY OF CHAR) : BOOLEAN ;
VAR
   f: File ;
BEGIN
   f := OpenToRead(fname) ;
   IF IsNoError(f)
   THEN
      Close(f) ;
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END Exists ;


PROCEDURE OpenToRead( fname : ARRAY OF CHAR ) : File ;

BEGIN
	RETURN openfile( fname, "r ", reading, CantOpenFile );
END OpenToRead;


PROCEDURE OpenToWrite( fname : ARRAY OF CHAR ) : File ;

BEGIN
	RETURN openfile( fname, "w ", writing, CantCreateFile );
END OpenToWrite;


PROCEDURE OpenForRandom( fname : ARRAY OF CHAR ; Create : BOOLEAN ) : File ;
(* dcw: 24th Aug 1988: both returned exit status OpFailed before now *)
BEGIN
	IF Create THEN
		RETURN openfile( fname, "w ", random, CantCreateFile );
	ELSE
		RETURN openfile( fname, "r ", random, CantOpenFile );
	END;
END OpenForRandom;


PROCEDURE Close (Id: File) ;
BEGIN
	(* check that the file is open, first *)
	IF ( Id<-1 ) OR ( Id>Maxfiles ) OR ( files[Id].type = none ) THEN
		ReportError( Id );
	END;

	(* now close it *)
	unixfclose( files[ Id ].filedes );
	files[ Id ].type     := none;
	files[ Id ].ioresult := UnInitialized;
END Close ;


PROCEDURE EOF( Id : File ) : BOOLEAN ;
VAR
	i : INTEGER;
	f : INTEGER;
BEGIN
	checkopenfor( "EOF", Id, TRUE, reading );
	f := files[ Id ].filedes;
	i := unixgetc( f );
	unixungetc( f, i );
	files[ Id ].ioresult := NoError;
	RETURN i = -1;	(* dcw: was c # CHR(255) *)
END EOF;


PROCEDURE EOLN( Id : File ) : BOOLEAN ;
VAR
	i : INTEGER;
	f : INTEGER;
BEGIN
	checkopenfor( "EOLN", Id, TRUE, reading );
	f := files[ Id ].filedes;
	i := unixgetc( f );
	unixungetc( f, i );
	files[ Id ].ioresult := NoError;
	RETURN i = INTEGER(ORD(NewLineCh));
END EOLN;


PROCEDURE SetPosition( Id : File; Pos : CARDINAL ; End : EndType ) ;

VAR	val : INTEGER;

BEGIN
	checkopenfor( "SetPosition", Id, TRUE, random );
	CASE End
	OF	FromStart:	val := 0;
	|	FromCurrent:	val := 1;
	|	AddToEnd:	val := 2;
	END;
	IF unixfseek( files[ Id ].filedes, Pos, val ) = 0 THEN
		files[ Id ].ioresult := NoError;
	ELSE
		files[ Id ].ioresult := OpFailed;
	END ;
END SetPosition;


PROCEDURE FindPosition( Id : File ) : CARDINAL;
BEGIN
	checkopenfor( "FindPosition", Id, TRUE, random );
	files[ Id ].ioresult := NoError;
	RETURN unixftell( files[ Id ].filedes );
END FindPosition;


PROCEDURE Rewind( Id : File ) ;
BEGIN
	SetPosition( Id, 0, FromStart );
END Rewind;


PROCEDURE ReadNBytes (Id :File; n :CARDINAL; buffer :ADDRESS) ;
VAR
   k: INTEGER;
BEGIN
   checkopenfor("ReadNBytes", Id, FALSE, writing) ;
   IF unixfread( files[ Id ].filedes, n, buffer ) = INTEGER(n)
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
END ReadNBytes;


PROCEDURE WriteNBytes (Id: File; n: CARDINAL; buffer: ADDRESS) ;
BEGIN
   checkopenfor("WriteNBytes", Id, FALSE, reading) ;
   IF unixfwrite( files[ Id ].filedes, n, buffer ) = INTEGER(n)
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
END WriteNBytes ;


PROCEDURE WriteBase (Id        : File;
                     base, num : CARDINAL;
		     width     : INTEGER);

VAR
   spaces : BOOLEAN;
BEGIN
   checkopenfor("WriteBase", Id, FALSE, reading) ;
   IF (base < 2) OR (base > 36)
   THEN
      errstr( "WriteBase: base out of range [2..36]" );
      errline;
      unixexit(1)
   END ;
   spaces := width>0 ;
   IF writeb(files[ Id ].filedes, base, num, spaces, ABS(width))
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
END WriteBase;


PROCEDURE WriteCardinal (Id : File; Object: CARDINAL) ;
BEGIN
   checkopenfor( "WriteCardinal", Id, FALSE, reading );
   IF unixfwrite( files[ Id ].filedes, TSIZE(CARDINAL), ADR(Object) ) = TSIZE(CARDINAL)
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
END WriteCardinal;


PROCEDURE WriteShort (Id : File; Object: CARDINAL) ;
VAR
   short: ARRAY [0..1] OF BYTE ;
BEGIN
   checkopenfor( "WriteShort", Id, FALSE, reading );
   short[0] := VAL(BYTE, Object MOD 0100H) ;
   short[1] := VAL(BYTE, (Object DIV 0100H) MOD 0100H) ;
   IF unixfwrite( files[ Id ].filedes, 2, ADR(short) ) = 2
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
END WriteShort ;


(*
PROCEDURE ReadCardinal( Id : File ) : CARDINAL ;

VAR k : CARDINAL ;

BEGIN
	checkopenfor( "ReadCardinal", Id, FALSE, writing );
	IF unixfreadcardinal( files[ Id ].filedes, k ) = 0 THEN
		files[ Id ].ioresult := NoError;
	ELSE
		files[ Id ].ioresult := OpFailed;
	END;
	RETURN k;
END ReadCardinal;


PROCEDURE ReadInteger( Id : File ) : INTEGER ;

VAR k : INTEGER ;

BEGIN
	checkopenfor( "ReadInteger", file, FALSE, writing );
	IF unixfreadinteger( files[ Id ].filedes, k ) = 0 THEN
		files[ Id ].ioresult := NoError;
	ELSE
		files[ Id ].ioresult := OpFailed;
	END ;
	RETURN k;
END ReadInteger;


PROCEDURE WriteInteger( Id : File; object : INTEGER ) ;

BEGIN
	checkopenfor( "WriteInteger", Id, FALSE, reading );
	IF unixfwriteinteger( files[ Id ].filedes, object ) = 0 THEN
		files[ Id ].ioresult := NoError;
	ELSE
		files[ Id ].ioresult := OpFailed;
	END;
END WriteInteger;


PROCEDURE ReadReal( Id : File ) : REAL ;

VAR k : REAL ;

BEGIN
	checkopenfor( "ReadReal", file, FALSE, writing );
	IF unixfreadreal( files[ Id ].filedes, k ) = 0 THEN
		files[ Id ].ioresult := NoError;
	ELSE
		files[ Id ].ioresult := OpFailed;
	END;
	RETURN k;
END ReadReal;


PROCEDURE WriteReal( Id : File; object : REAL ) ;

BEGIN
	checkopenfor( "WriteReal", file, FALSE, reading );
	IF unixfwritereal( files[Id].filedes, object ) = 0 THEN
		files[Id].ioresult := NoError;
	ELSE
		files[Id].ioresult := OpFailed;
	END;
END WriteReal;
*)


PROCEDURE ReadChar (Id: File) : CHAR ;
VAR
   i: INTEGER;
   k: CHAR ;
BEGIN
   checkopenfor( "ReadChar", Id, FALSE, writing );
   i := unixgetc( files[ Id ].filedes );
   IF i = -1
   THEN
      (* shouldn't reading beyond eof abort with an error mesg?*)
      files[ Id ].ioresult := OpFailed;
      k := 0C;
   ELSE
      files[ Id ].ioresult := NoError;
      k := CHR(ORD(i));
   END;
   RETURN k;
END ReadChar;


PROCEDURE WriteChar (Id: File; Object: CHAR) ;
BEGIN
   checkopenfor("WriteChar", Id, FALSE, reading) ;
   IF unixputc( files[Id].filedes, INTEGER(ORD(Object)) ) = 0
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
END WriteChar ;


PROCEDURE ReadString (Id: File; VAR Object: ARRAY OF CHAR) ;
VAR
   high,
   i   : CARDINAL ;
   ch  : CHAR ;
BEGIN
   checkopenfor( "ReadString", Id, FALSE, writing );
   high := HIGH(Object) ;
   i := 0 ;
   REPEAT
      ch := ReadChar(Id) ;
      IF i<=high
      THEN
         IF (ch=nl) OR (NOT IsNoError(Id))
         THEN
            Object[i] := nul ;
            INC( i )
         ELSE
            Object[i] := ch ;
            INC( i )
         END
      END
   UNTIL (ch=nl) OR (i>high) OR (NOT IsNoError(Id))

(* was
   IF unixfreadstr( files[ Id ].filedes, ADR(Object), HIGH(Object) ) = 0
   THEN
      files[ Id ].ioresult := NoError
   ELSE
      files[ Id ].ioresult := OpFailed
   END
*)
END ReadString;


PROCEDURE WriteString (Id: File; Object: ARRAY OF CHAR) ;
VAR
   Start,
   i     : CARDINAL ;
   l     : INTEGER ;
   Buffer: ARRAY [0..MaxBufferSize] OF CHAR ;
BEGIN
   checkopenfor( "WriteString", Id, FALSE, reading );
   l := len(Object) ;
   Start := 0 ;
   REPEAT
      i := 0 ;
      WHILE (Object[i]#0C) AND (i<MaxBufferSize) DO
         Buffer[i] := Object[Start+i] ;
         INC(i)
      END ;
      IF unixfwrite( files[ Id ].filedes, i, ADR(Buffer) ) = i
      THEN
         files[ Id ].ioresult := NoError
      ELSE
         files[ Id ].ioresult := OpFailed ;
         RETURN
      END ;
      INC(Start, i)
   UNTIL Start=l
END WriteString ;


PROCEDURE ReadLine (Id :File) ;
BEGIN
   checkopenfor( "ReadLine", Id, FALSE, writing );
   REPEAT
      IF EOF(Id)
      THEN
         errstr( "ReadLine: Attempt to read beyond EOF on '" ) ;
         errstr( files[Id].name ) ;
         errstr( "' " ) ;
         errline ;
         unixexit(1)
      END ;
   UNTIL unixgetc(files[Id].filedes) = INTEGER(ORD(NewLineCh));
   files[Id].ioresult := NoError
END ReadLine;


PROCEDURE WriteLine (Id :File) ;
BEGIN
   WriteChar(Id, NewLineCh)
END WriteLine ;


PROCEDURE UnReadChar( Id : File ; c : CHAR ) ;

(* currently no check to avoid two successive unreadchar calls *)
(* probably worth having one.... *)

BEGIN
   checkopenfor( "UnReadChar", Id, FALSE, writing );
   unixungetc( INTEGER(Id), INTEGER(ORD(c)) );
   files[ Id ].ioresult := NoError;
END UnReadChar;


(* -------------------- Private procedures section ------------------- *)



PROCEDURE len( s : ARRAY OF CHAR ) : INTEGER;
VAR i : INTEGER;

BEGIN
	FOR i := 0 TO INTEGER(HIGH(s)) DO
		IF s[i] = 0C THEN RETURN i END;
	END;
	RETURN 1+HIGH(s);
END len;


PROCEDURE writeb(     f      : INTEGER;
		      b      : CARDINAL;
		      n      : CARDINAL;
		  VAR spaces : BOOLEAN;
		      digits : INTEGER ) : BOOLEAN;
CONST
	ASpace = ' ';

VAR
	i : INTEGER;

BEGIN
	IF ( n >= b ) OR ( digits > 1 ) THEN
		IF NOT writeb( f, b, n DIV b, spaces, digits - 1 ) THEN
			RETURN FALSE;
		END;
	END;
	n := n MOD b;

	IF ( n = 0 ) AND ( digits > 1 ) AND spaces THEN
		i := ORD(ASpace);
	ELSE
		spaces := FALSE;
		IF n > 9 THEN
			i := n-10 + ORD('A');
		ELSE
			i := n + ORD('0');
		END;
	END;
	RETURN unixputc( f, i ) = 0;
END writeb;



PROCEDURE errstr( s : ARRAY OF CHAR );
BEGIN
	WriteString( StdErr, s );
END errstr;



PROCEDURE errline;
BEGIN
	WriteLine( StdErr );
END errline;


PROCEDURE openfile(fname: ARRAY OF CHAR;
                   fmode: ARRAY OF CHAR;
                   ftype: FileType;
                   fstat: Status) : File ;

(* open a file: called by all three OpenTo procedures *)

VAR
   filepos : File;
   i       : INTEGER;
   High    ,
   c       : CARDINAL;
   fmodecp : ARRAY [0..20] OF CHAR ;
BEGIN
   filepos := findfree();
   (* copy name into array *)
   High := Max(StrLen(fname), (MaxNameSize)-1) ;
   FOR c := 0 TO High DO
      files[filepos].name[c] := fname[c];
   END ;
   FOR c := High+1 TO MaxNameSize DO
      files[filepos].name[c] := 0C;
   END ;
   (* copy mode info our local copy *)
   High := Min(StrLen(fmode), HIGH(fmodecp)-1) ;
   c := 0 ;
   WHILE c<High DO
      fmodecp[c] := fmode[c] ;
      INC(c)
   END ;
   FOR c := High TO HIGH(fmodecp) DO
      fmodecp[c] := 0C
   END ;
   IF filepos = -1
   THEN
      files[ filepos ].ioresult := TooManyOpen;
   ELSE
      i := unixfopen( ADR(files[filepos].name), ADR(fmodecp) );
      IF i=-1
      THEN
         files[filepos].ioresult := fstat;
      ELSE
         files[filepos].type    := ftype;
         files[filepos].filedes := i;
         files[filepos].ioresult:= NoError;
      END;
   END;
   RETURN filepos;
END openfile;



PROCEDURE findfree() : File;
(* finds the first free position in the file array *)
VAR i : File;
BEGIN
	FOR i := 3 TO Maxfiles DO
		IF files[ i ].type = none THEN RETURN i END;
	END;
	RETURN -1;
END findfree;


PROCEDURE checkopenfor( procname : ARRAY OF CHAR;
			file : File;
			mustbe : BOOLEAN;
			mode : FileType );

VAR	wasit : ARRAY [ 0..30 ] OF CHAR;

BEGIN
	(* check that the file is open, first *)

	IF ( file<-1 ) OR ( file>Maxfiles ) OR ( files[file].type = none ) THEN
		errstr( procname );
		errstr( " : File has not been opened !!" );
		errline;
		unixexit(1);
	END;

 	IF ( files[ file ].type # mode ) AND mustbe THEN
		wasit := "was not";
 	ELSIF ( files[ file ].type = mode ) AND NOT mustbe THEN
		wasit := "shouldn't have been";
	ELSE
		RETURN			(* all ok *)
	END;
	errstr( procname );
	errstr( " : file '" );
	errstr( files[file].name ) ;
	errstr( "' " );
	errstr( wasit );
	errstr( " opened for " );
	errstr( modename[ mode ] );
	errline;
	unixexit(1);
END checkopenfor;



(* -------------------------- initialization code -------------------------- *)

VAR i : INTEGER ;

BEGIN
	modename[ none ] := "none...ahem..";
	modename[ reading ] := "reading";
	modename[ writing ] := "writing";
	modename[ random ] := "random";

	FOR i := -1 TO Maxfiles DO
		files[ i ].type     := none;
		files[ i ].filedes  := -1;
		files[ i ].ioresult := UnInitialized;
		files[ i ].name     := "no name";
	END;

	StdIn  := 0;
	StdOut := 1;
	StdErr := 2;

	files[ StdIn ].type     := reading;
	files[ StdIn ].filedes  := 0;
	files[ StdIn ].ioresult := NoError;
	files[ StdIn ].name     := "stdin";

	files[ StdOut ].type     := writing;
	files[ StdOut ].filedes  := 1;
	files[ StdOut ].ioresult := NoError;
	files[ StdOut ].name     := "stdout";

	files[ StdErr ].type     := writing;
	files[ StdErr ].filedes  := 2;
	files[ StdErr ].ioresult := NoError;
	files[ StdErr ].name     := "stderr";
END FIO.
