IMPLEMENTATION MODULE Scan ;


IMPORT StdIO ;

FROM ASCII IMPORT nul, lf, cr, bs, del, bel ;
FROM StdIO IMPORT Write ;
FROM StrLib IMPORT StrEqual, StrLen, StrCopy ;
FROM NumberIO IMPORT WriteCard ;
FROM FIO IMPORT OpenToRead, IsNoError, Close, File, ReadChar ;
FROM StrIO IMPORT WriteLn, WriteString ;
FROM libc IMPORT exit ;


(* %%%FORWARD%%%
PROCEDURE OpenSource (a: ARRAY OF CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE CloseSource ; FORWARD ;
PROCEDURE GetNextSymbol (VAR a: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE ChuckUpToSymbol ; FORWARD ;
PROCEDURE SymbolChar () : CHAR ; FORWARD ;
PROCEDURE NextChar ; FORWARD ;
PROCEDURE NonSymbolChar () : BOOLEAN ; FORWARD ;
PROCEDURE WriteError (a: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE ReadString (VAR a: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE Read (VAR ch: CHAR) ; FORWARD ;
   %%%FORWARD%%% *)


CONST
   MaxLength = 255 ;  (* Max Length of Source Line *)

VAR
   CurrentString   : ARRAY [0..MaxLength] OF CHAR ;
   CurrentLineNo   : CARDINAL ;
   CurrentCursorPos: CARDINAL ;
   EOF             : BOOLEAN ;
   LengthOfCurSym  : CARDINAL ;
   f               : File ;
   Opened          : BOOLEAN ;
   HaltOnError     : BOOLEAN ;


PROCEDURE OpenSource (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   f := OpenToRead(a) ;
   IF IsNoError(f)
   THEN
      StrCopy( '', CurrentString ) ;
      LengthOfCurSym := 0 ;
      CurrentCursorPos := 0 ;
      EOF := FALSE ;
      CurrentLineNo := 1 ;
      Opened := TRUE
   ELSE
      Opened := FALSE
   END ;
   RETURN( Opened )
END OpenSource ;


PROCEDURE CloseSource ;
BEGIN
   IF Opened
   THEN
      Close( f ) ;
      Opened := FALSE
   END
END CloseSource ;


PROCEDURE GetNextSymbol (VAR a: ARRAY OF CHAR) ;
VAR
   index,
   High : CARDINAL ;
BEGIN
   ChuckUpToSymbol ;
   index := 0 ;
   High := HIGH( a ) ;
   WHILE (index<High) AND (NOT NonSymbolChar()) AND (NOT EOF)
   DO
      a[index] := SymbolChar() ;
      NextChar ;
      INC( index )
   END ;
   IF index<High
   THEN
      a[index] := nul
   END ;
   LengthOfCurSym := index
END GetNextSymbol ;


PROCEDURE ChuckUpToSymbol ;
BEGIN
   WHILE (NOT EOF) AND NonSymbolChar() DO
      NextChar
   END
END ChuckUpToSymbol ;


(* SymbolChar returns a character from the CurrentString, if the end  *)
(* of CurrentString is found then SymbolChar returns nul.             *)

PROCEDURE SymbolChar () : CHAR ;
BEGIN
   IF EOF
   THEN
      RETURN( nul )
   ELSE
      IF CurrentCursorPos<StrLen(CurrentString)
      THEN
         RETURN( CurrentString[CurrentCursorPos] )
      ELSE
         RETURN( nul )
      END
   END
END SymbolChar ;


(* NextChar advances the CurrentCursorPos along a line of the source, *)
(* resetting the CurrentCursorPos every time a newline is read.       *)

PROCEDURE NextChar ;
VAR
   ch: CHAR ;
BEGIN
   IF NOT EOF
   THEN
      IF CurrentCursorPos<StrLen(CurrentString)
      THEN
         INC( CurrentCursorPos )
      ELSE
         ReadString( CurrentString ) ;
         (* WriteString( CurrentString ) ; WriteLn ; *)
         INC( CurrentLineNo ) ;
         CurrentCursorPos := 0 ;
         LengthOfCurSym := 0 ;
         (* IF StrEqual( CurrentString, 'END' )
         THEN
            EOF := TRUE
         END  *)
      END
   END
END NextChar ;


PROCEDURE NonSymbolChar () : BOOLEAN ;
BEGIN
   RETURN( CurrentString[CurrentCursorPos]<=' ' )
END NonSymbolChar ;


PROCEDURE WriteError (a: ARRAY OF CHAR) ;
VAR
   i: CARDINAL ;
BEGIN
   WriteString('Error in line') ;
   WriteCard( CurrentLineNo, 6 ) ; WriteLn ;
   WriteString( CurrentString ) ; WriteLn ;
   i := 0 ;
   WHILE i<CurrentCursorPos-LengthOfCurSym DO
      Write(' ') ;
      INC( i )
   END ;
   FOR i := 1 TO LengthOfCurSym DO
      Write('^')
   END ;
   WriteLn ;
   WriteString( a ) ; WriteLn ;
   IF HaltOnError
   THEN
      exit(1)
   END
END WriteError ;


PROCEDURE ReadString (VAR a: ARRAY OF CHAR) ;
VAR
   n    ,
   high : CARDINAL ;
   ch   : CHAR ;
BEGIN
   high := HIGH( a ) ;
   n := 0 ;
   REPEAT
      Read( ch ) ;
      IF (ch=del) OR (ch=bs)
      THEN
         IF n=0
         THEN
            Write( bel )
         ELSE
            Write( bs ) ;
            Write(' ') ;
            Write( bs ) ;
            DEC( n )
         END
      ELSIF n <= high
      THEN
         IF (ch = cr) OR (cr = lf)
         THEN
            a[n] := nul
         ELSE
(*            Write( ch ) ;
 *)           a[n] := ch
         END ;
         INC( n )
      ELSE
         ch := cr  (* exit gracefully *)
      END
   UNTIL ch = cr
END ReadString ;


PROCEDURE Read (VAR ch: CHAR) ;
BEGIN
   IF Opened
   THEN
      ch := ReadChar(f) ;
      EOF := NOT IsNoError(f)
   ELSE
      StdIO.Read( ch )
   END ;
   IF ch=lf THEN ch := cr END
END Read ;


(*
   TerminateOnError - exits with status 1 if we call WriteError.
*)

PROCEDURE TerminateOnError ;
BEGIN
   HaltOnError := TRUE
END TerminateOnError ;


BEGIN
   StrCopy( '', CurrentString ) ;
   LengthOfCurSym := 0 ;
   CurrentCursorPos := 0 ;
   EOF := FALSE ;
   CurrentLineNo := 1 ;
   Opened := FALSE ;
   HaltOnError := FALSE
END Scan.
