IMPLEMENTATION MODULE M2Quads ;


FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StrLib IMPORT StrEqual, StrConCat ;
FROM NumberIO IMPORT CardToStr ;
FROM M2Debug IMPORT Assert, WriteDebug ;
FROM NumberIO IMPORT WriteCard ;
FROM NameKey IMPORT NulName, WriteKey, MakeKey, GetKey ;

FROM SymbolTable IMPORT ModeOfAddr, GetMode, GetSymName, IsUnknown,
                        MakeTemporary, MakeConstLit, MakeConstLitString,
                        RequestSym,
                        GetType, GetLowestType,
                        GetSubrange,
                        GetModule,
                        GetCurrentModule, GetLocalSym,
                        GetStringLength, GetString,
                        GetParam, GetProcedureQuads,
                        GetNth, GetNthParam,
                        GetFirstUsed, GetDeclared,
                        PutConstString,
                        PutModuleStartQuad, PutModuleEndQuad,
                        PutProcedureStartQuad, PutProcedureEndQuad,
                        PutVar, PutVarQuad, RemoveVarQuad,
                        IsVarParam, IsProcedure, IsPointer,
                        IsUnboundedParam, IsEnumeration,
                        NoOfElements,
                        NoOfParam,
                        StartScope, EndScope,
                        Father, IsRecordField, IsFieldVarient,
                        IsVar, IsProcType, IsType, IsSubrange,
                        IsConst, IsConstString,
                        IsArray, IsUnbounded,
                        IsSet,
                        IsSubscript,
                        IsTemporary,
                        IsAModula2Type,
                        PutVarTypeAndSize,
                        PushSize,
                        NulSym ;

FROM M2Configure IMPORT PushParametersLeftToRight ;

FROM M2Comp IMPORT CompilingImplementationModule,
                   CompilingProgramModule ;

FROM M2Lexical IMPORT WriteError, CurrentToken, GetToken, PutToken,
                      TypeOfToken, TokenType, NearToken, GetFileName,
                      GetTokenNo, GetLineNo, GetPreviousTokenLineNo,
                      NearTokens, InternalError, WriteErrorFormat1, WriteErrorFormat2 ;

FROM M2Error IMPORT BeginError, EndError ;

FROM M2Reserved IMPORT PlusTok, MinusTok, TimesTok, DivTok, ModTok,
                       DivideTok,
                       OrTok, AndTok, AmbersandTok,
                       EqualTok, LessEqualTok, GreaterEqualTok,
                       LessTok, GreaterTok, HashTok, LessGreaterTok,
                       InTok,
                       UpArrowTok, RParaTok, LParaTok, CommaTok,
                       SemiColonTok ;

FROM M2Base IMPORT True, False, Boolean, Cardinal, Integer, Char,
                   Real, LongReal,
                   MixTypes, CheckCompatibleTypes, IsCompatible,
                   Unbounded, ArrayAddress, ArrayHigh,
                   High, New, Dispose, Inc, Dec, Incl, Excl,
                   Ord, Chr, Convert, Val, Float, Trunc,
                   IsPseudoBaseProcedure, IsPseudoBaseFunction,
                   IsBaseType ;

FROM M2System IMPORT IsPseudoSystemFunction, IsSystemType,
                     Adr, Size, TSize,
                     Address, Bitset, Byte, Word ;

FROM M2Math IMPORT IsPseudoMathFunction ;

FROM M2ALU IMPORT PopCard ;

FROM M2Options IMPORT BoundsChecking,
                      DefaultCoding, DefaultProfiling, DefaultOptimizing,
                      Pedantic, CompilerDebugging, GenerateDebugging,
                      GenerateLineDebug,
                      Profiling, Coding, Optimizing ;

FROM M2Gen487 IMPORT IsCoprocessorType ;


CONST
   MaxStack     =  1000 ;
   MaxExitStack =    10 ;
   MaxWithStack =    10 ;
   MaxQuad      = 50000 ;
   MaxString    =  4096 ;

TYPE
   Frame = RECORD
              TrueExit : CARDINAL ;
              FalseExit: CARDINAL ;
              BooleanOp: BOOLEAN ;
           END ;


   QuadFrame = RECORD
                  Operator           : QuadOperator ;
                  Operand1           : CARDINAL ;
                  Operand2           : CARDINAL ;
                  Operand3           : CARDINAL ;
                  Next               : CARDINAL ;     (* Next quadruple                 *)
                  LineNo             : CARDINAL ;     (* Line No of source text         *)
                  TokenNo            : CARDINAL ;     (* Token No of source text        *)
                  NoOfTimesReferenced: CARDINAL ;     (* No of times quad is referenced *)
               END ;


   WithFrame = RECORD
                  PtrSym   : CARDINAL ;
                  RecordSym: CARDINAL ;
               END ;

VAR
   Stack                : ARRAY [0..MaxStack] OF Frame ;
   ExitStack            : ARRAY [0..MaxExitStack] OF CARDINAL ;
   WithStack            : ARRAY [0..MaxWithStack] OF WithFrame ;
   Ptr                  : CARDINAL ;
   ExitPtr              : CARDINAL ;
   WithPtr              : CARDINAL ;

   Quads                : ARRAY [1..MaxQuad] OF QuadFrame ;
   NextQuad             : CARDINAL ;  (* Next quadruple number to be created.    *)
   FreeList             : CARDINAL ;  (* FreeList of quadruples.                 *)
   ReturnQuad           : CARDINAL ;  (* ReturnQuad of the procedure.            *)
   CurrentProc          : CARDINAL ;  (* Current procedure being compiled, used  *)
                                      (* to determine which procedure a RETURN   *)
                                      (* ReturnValueOp must have as its 3rd op.  *)
   InitQuad             : CARDINAL ;  (* Initial Quad BackPatch that starts the  *)
                                      (* suit of Modules.                        *)
   LastQuadNo           : CARDINAL ;  (* Last Quadruple accessed by GetQuad.     *)
   LogicalOrTok         : CARDINAL ;  (* Internal _LOR token.                    *)
   LogicalAndTok        : CARDINAL ;  (* Internal _LAND token.                   *)
   LogicalDifferenceTok : CARDINAL ;  (* Internal _LDIFF token.                  *)
   LogicalXorTok        : CARDINAL ;  (* Internal _LXOR token.                   *)
   NoOfDynamic          : CARDINAL ;
   MustNotCheckBounds   : BOOLEAN ;
   GrowInitialization   : CARDINAL ;  (* upper limit of where the initialized    *)
                                      (* quadruples.                             *)


(* %%%FORWARD%%%
PROCEDURE CheckRemoveVariable (Sym: CARDINAL; Quad: CARDINAL) ; FORWARD ;
PROCEDURE CheckFunctionReturn (ProcSym: CARDINAL) ; FORWARD ;
PROCEDURE CheckAddVariable (Sym: CARDINAL; Quad: CARDINAL) ; FORWARD ;
PROCEDURE ConvertBooleanToVariable (i: CARDINAL) ; FORWARD ;
PROCEDURE BuildPseudoMathFunction ; FORWARD ;
PROCEDURE BuildFloatFunction ; FORWARD ;
PROCEDURE BuildTruncFunction ; FORWARD ;
PROCEDURE CheckAssignmentCompatible (Des, Exp: CARDINAL) ; FORWARD ;
PROCEDURE CheckForLogicalOperator (Tok: CARDINAL; e1, t1, e2, t2: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE DisplayType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE CheckProcedureParameters ; FORWARD ;
PROCEDURE CheckParameter (Call, Param, ProcSym: CARDINAL; i: CARDINAL) ; FORWARD ;
PROCEDURE FailParameter (CurrentState : ARRAY OF CHAR;
                         Given        : CARDINAL;
                         Expecting    : CARDINAL;
                         ProcedureSym : CARDINAL;
                         ParameterNo  : CARDINAL) ; FORWARD ;
PROCEDURE DisplayType (Sym: CARDINAL) ; FORWARD ;
PROCEDURE AlterReference (Head, OldQuad, NewQuad: CARDINAL) ; FORWARD ;
PROCEDURE RemoveReference (q: CARDINAL) ; FORWARD ;
PROCEDURE ManipulateReference (q: CARDINAL; to: CARDINAL) ; FORWARD ;
PROCEDURE AreConstant (b: BOOLEAN) : ModeOfAddr ; FORWARD ;
PROCEDURE AssignUnboundedNonVar (Sym, UnboundedSym, ParamType: CARDINAL) ; FORWARD ;
PROCEDURE AssignUnboundedVar (Sym, UnboundedSym, ParamType: CARDINAL) ; FORWARD ;
PROCEDURE BackPatch (QuadNo, Value: CARDINAL) ; FORWARD ;
PROCEDURE BuildAccessWithField ; FORWARD ;
PROCEDURE BuildAdrFunction ; FORWARD ;
PROCEDURE BuildChrFunction ; FORWARD ;
PROCEDURE BuildConvertFunction ; FORWARD ;
PROCEDURE BuildDecProcedure ; FORWARD ;
PROCEDURE BuildDisposeProcedure ; FORWARD ;
PROCEDURE BuildDynamicArray ; FORWARD ;
PROCEDURE BuildHighFromChar ; FORWARD ;
PROCEDURE BuildHighFromArray ; FORWARD ;
PROCEDURE BuildHighFromString ; FORWARD ;
PROCEDURE BuildHighFromUnbounded ; FORWARD ;
PROCEDURE BuildHighFunction ; FORWARD ;
PROCEDURE BuildIncProcedure ; FORWARD ;
PROCEDURE BuildNewProcedure ; FORWARD ;
PROCEDURE BuildInclProcedure ; FORWARD ;
PROCEDURE BuildExclProcedure ; FORWARD ;
PROCEDURE BuildOrdFunction ; FORWARD ;
PROCEDURE BuildPseudoFunctionCall ; FORWARD ;
PROCEDURE BuildPseudoProcedureCall ; FORWARD ;
PROCEDURE BuildRealFunctionCall ; FORWARD ;
PROCEDURE BuildRealProcedureCall ; FORWARD ;
PROCEDURE BuildSizeFunction ; FORWARD ;
PROCEDURE BuildStaticArray ; FORWARD ;
PROCEDURE BuildTSizeFunction ; FORWARD ;
PROCEDURE BuildTypeCoercion ; FORWARD ;
PROCEDURE BuildValFunction ; FORWARD ;
PROCEDURE CheckBooleanId ; FORWARD ;
PROCEDURE DisplayQuad (QuadNo: CARDINAL) ; FORWARD ;
PROCEDURE DisplayStack ; FORWARD ;
PROCEDURE DisposeQuad (QuadNo: CARDINAL) ; FORWARD ;
PROCEDURE GenQuad (Operation: QuadOperator;
                   Op1, Op2, Op3: CARDINAL) ; FORWARD ;
PROCEDURE GetItemPointedTo (Sym: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE Init ; FORWARD ;
PROCEDURE InitQuads ; FORWARD ;
PROCEDURE IsBoolean (Pos: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE IsReallyAPointer (Sym: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE MakeOp (t: CARDINAL) : QuadOperator ; FORWARD ;
PROCEDURE ManipulateParameters ; FORWARD ;
PROCEDURE Merge (QuadList1, QuadList2: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE NewQuad (VAR QuadNo: CARDINAL) ; FORWARD ;
PROCEDURE PopBool (VAR True, False: CARDINAL) ; FORWARD ;
PROCEDURE PopExit() : CARDINAL ; FORWARD ;
PROCEDURE PopInit (VAR q: CARDINAL) ; FORWARD ;
PROCEDURE PopWith ; FORWARD ;
PROCEDURE PushBool (True, False: CARDINAL) ; FORWARD ;
PROCEDURE PushExit (Exit: CARDINAL) ; FORWARD ;
PROCEDURE PushInit (q: CARDINAL) ; FORWARD ;
PROCEDURE PushWith (Sym, Type: CARDINAL) ; FORWARD ;
PROCEDURE UnboundedNonVarLinkToArray (ArraySym, UnboundedSym, ParamType: CARDINAL) ; FORWARD ;
PROCEDURE UnboundedVarLinkToArray (ArraySym, UnboundedSym, ParamType: CARDINAL) ; FORWARD ;
PROCEDURE WriteMode (Mode: ModeOfAddr) ; FORWARD ;
PROCEDURE WriteOperand (Sym: CARDINAL) ; FORWARD ;
PROCEDURE WriteOperator (Operator: QuadOperator) ; FORWARD ;
PROCEDURE WriteQuad (BufferQuad: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   SetOptionProfiling - builds a profile quadruple if the profiling
                        option was given to the compiler.
*)

PROCEDURE SetOptionProfiling (b: BOOLEAN) ;
BEGIN
   IF b#Profiling
   THEN
      IF b
      THEN
         BuildProfileOn
      ELSE
         BuildProfileOff
      END ;
      Profiling := b
   END
END SetOptionProfiling ;


(*
   SetOptionCoding - builds a code quadruple if the profiling
                        option was given to the compiler.
*)

PROCEDURE SetOptionCoding (b: BOOLEAN) ;
BEGIN
   IF b#Coding
   THEN
      IF b
      THEN
         BuildCodeOn
      ELSE
         BuildCodeOff
      END ;
      Coding := b
   END
END SetOptionCoding ;


(*
   SetOptionOptimizing - builds a quadruple to say that the optimization option
                         has been found in a comment.
*)

PROCEDURE SetOptionOptimizing (b: BOOLEAN) ;
BEGIN
   IF b
   THEN
      BuildOptimizeOn
   ELSE
      BuildOptimizeOff
   END
END SetOptionOptimizing ;


(*
   Opposite - returns the opposite comparison operator.
*)

PROCEDURE Opposite (Operator: QuadOperator) : QuadOperator ;
VAR
   Op: QuadOperator ;
BEGIN
   CASE Operator OF

   IfNotEquOp : Op := IfEquOp |
   IfEquOp    : Op := IfNotEquOp |
   IfLessEquOp: Op := IfGreOp |
   IfGreOp    : Op := IfLessEquOp |
   IfGreEquOp : Op := IfLessOp |
   IfLessOp   : Op := IfGreEquOp |
   IfInOp     : Op := IfNotInOp |
   IfNotInOp  : Op := IfInOp

   ELSE
      InternalError('unexpected operator', __FILE__, __LINE__)
   END ;
   RETURN( Op )
END Opposite ;


(*
   IsReferenced - returns true if QuadNo is referenced inside the Quad list
                  determined by Head.
*)

PROCEDURE IsReferenced (Head: CARDINAL; QuadNo: CARDINAL) : BOOLEAN ;
VAR
   i: CARDINAL ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( (Operator=NewLocalVarOp) OR (NoOfTimesReferenced>0) )
   END
END IsReferenced ;


(*
   IsUnConditional - returns true if QuadNo is an unconditional jump.
*)

PROCEDURE IsUnConditional (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      CASE Operator OF

      CallOp,
      ReturnOp,
      GotoOp     : RETURN( TRUE )

      ELSE
         RETURN( FALSE )
      END ;
   END
END IsUnConditional ;


(*
   IsConditional - returns true if QuadNo is a conditional jump.
*)

PROCEDURE IsConditional (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      CASE Operator OF

      IfInOp,
      IfNotInOp,
      IfEquOp,
      IfNotEquOp,
      IfLessOp,
      IfLessEquOp,
      IfGreOp,
      IfGreEquOp : RETURN( TRUE )

      ELSE
         RETURN( FALSE )
      END ;
   END
END IsConditional ;


(*
   IsCall - returns true if QuadNo is a call operation.
*)

PROCEDURE IsCall (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( Operator=CallOp )
   END
END IsCall ;


(*
   IsReturn - returns true if QuadNo is a return operation.
*)

PROCEDURE IsReturn (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( Operator=ReturnOp )
   END
END IsReturn ;


(* 
   IsNewLocalVar - returns true if QuadNo is a NewLocalVar operation.
*) 
 
PROCEDURE IsNewLocalVar (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( Operator=NewLocalVarOp )
   END
END IsNewLocalVar ;


(* 
   IsKillLocalVar - returns true if QuadNo is a KillLocalVar operation.
*) 
 
PROCEDURE IsKillLocalVar (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( Operator=KillLocalVarOp )
   END
END IsKillLocalVar ;


(*
   IsOptimizeOn - returns true if the Optimize flag was true at QuadNo.
*)

PROCEDURE IsOptimizeOn (QuadNo: CARDINAL) : BOOLEAN ;
VAR
   n,
   q : CARDINAL ;
   On: BOOLEAN ;
BEGIN
   On := DefaultOptimizing ;
   q := Head ;
   WHILE (q#0) AND (q#QuadNo) DO
      WITH Quads[q] DO
         IF Operator=OptimizeOnOp
         THEN
            On := TRUE
         ELSIF Operator=OptimizeOffOp
         THEN
            On := FALSE
         END ;
         n := Next
      END ;
      q := n
   END ;
   RETURN( On )
END IsOptimizeOn ;


(*
   IsProfileOn - returns true if the Profile flag was true at QuadNo.
*)

PROCEDURE IsProfileOn (QuadNo: CARDINAL) : BOOLEAN ;
VAR
   n,
   q : CARDINAL ;
   On: BOOLEAN ;
BEGIN
   On := DefaultProfiling ;
   q := Head ;
   WHILE (q#0) AND (q#QuadNo) DO
      WITH Quads[q] DO
         IF Operator=ProfileOnOp
         THEN
            On := TRUE
         ELSIF Operator=ProfileOffOp
         THEN
            On := FALSE
         END ;
         n := Next
      END ;
      q := n
   END ;
   RETURN( On )
END IsProfileOn ;


(*
   IsCodeOn - returns true if the Code flag was true at QuadNo.
*)

PROCEDURE IsCodeOn (QuadNo: CARDINAL) : BOOLEAN ;
VAR
   n,
   q : CARDINAL ;
   On: BOOLEAN ;
BEGIN
   On := DefaultCoding ;
   q := Head ;
   WHILE (q#0) AND (q#QuadNo) DO
      WITH Quads[q] DO
         IF Operator=CodeOnOp
         THEN
            On := TRUE
         ELSIF Operator=CodeOffOp
         THEN
            On := FALSE
         END ;
         n := Next
      END ;
      q := n
   END ;
   RETURN( On )
END IsCodeOn ;


(*
   IsModFile - returns TRUE if QuadNo is a start of Module file
               directive.
*)

PROCEDURE IsModFile (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( Operator=StartModFileOp )
   END
END IsModFile ;


(*
   IsPseudoQuad - returns true if QuadNo is a compiler directive.
                  ie code, profile and optimize.
                     StartFile, EndFile,
*)

PROCEDURE IsPseudoQuad (QuadNo: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Quads[QuadNo] DO
      RETURN( (Operator=CodeOnOp) OR (Operator=CodeOffOp) OR
              (Operator=ProfileOnOp) OR (Operator=ProfileOffOp) OR
              (Operator=OptimizeOnOp) OR (Operator=OptimizeOffOp) OR
              (Operator=EndFileOp) OR (Operator=EndOp) OR
              (Operator=StartDefFileOp) OR (Operator=StartModFileOp)
            )
   END
END IsPseudoQuad ;


(*
   GetLastFileQuad - returns the Quadruple number of the last StartDefFile or
                     StartModFile quadruple.
*)

PROCEDURE GetLastFileQuad (QuadNo: CARDINAL) : CARDINAL ;
VAR
   Head,
   i,
   FileQuad: CARDINAL ;
BEGIN
   Head := 1 ;  (* Warning Head assummed to start at quadruple 1 *)
   FileQuad := 0 ;
   REPEAT
      WITH Quads[Head] DO
         IF (Operator=StartModFileOp) OR (Operator=StartDefFileOp)
         THEN
            FileQuad := Head
         END ;
         i := Next
      END ;
      Head := i
   UNTIL (i=QuadNo) OR (i=0) ;
   Assert(i#0) ;  (* Should never occur - if it does - check warning above *)
   Assert(FileQuad#0) ;  (* Should never occur *)
   RETURN( FileQuad )
END GetLastFileQuad ;


(*
   GetLastQuadNo - returns the last quadruple number referenced
                   by a GetQuad.
*)

PROCEDURE GetLastQuadNo () : CARDINAL ;
BEGIN
   RETURN( LastQuadNo )
END GetLastQuadNo ;


(*
   QuadToLineNo - Converts a QuadNo into the approprate line number of the
                  source file, the line number is returned.

                  This may be used to yield an idea where abouts in the
                  source file the code generetion is
                  processing.
*)

PROCEDURE QuadToLineNo (QuadNo: CARDINAL) : CARDINAL ;
BEGIN
   IF LastQuadNo=0
   THEN
      RETURN( 0 )
   ELSE
      RETURN( Quads[QuadNo].LineNo )
   END
END QuadToLineNo ;


(*
   QuadToTokenNo - Converts a QuadNo into the approprate token number of the
                   source file, the line number is returned.

                   This may be used to yield an idea where abouts in the
                   source file the code generetion is
                   processing.
*)

PROCEDURE QuadToTokenNo (QuadNo: CARDINAL) : CARDINAL ;
BEGIN
   IF LastQuadNo=0
   THEN
      RETURN( 0 )
   ELSE
      RETURN( Quads[QuadNo].TokenNo )
   END
END QuadToTokenNo ;


(*
   GetQuad - returns the Quadruple QuadNo.
*)

PROCEDURE GetQuad (QuadNo: CARDINAL;
                   VAR Op: QuadOperator;
                   VAR Oper1, Oper2, Oper3: CARDINAL) ;
BEGIN
   LastQuadNo := QuadNo ;
   WITH Quads[QuadNo] DO
      Op := Operator ;
      Oper1 := Operand1 ;
      Oper2 := Operand2 ;
      Oper3 := Operand3
   END
END GetQuad ;


(*
   AddQuadInformation - adds variable analysis and jump analysis to the new quadruple.
*)

PROCEDURE AddQuadInformation (QuadNo: CARDINAL;
                              Op: QuadOperator;
                              Oper1, Oper2, Oper3: CARDINAL) ;
BEGIN
   CASE Op OF

   IfInOp,
   IfNotInOp,
   IfEquOp,
   IfNotEquOp,
   IfLessOp,
   IfLessEquOp,
   IfGreOp,
   IfGreEquOp : ManipulateReference(QuadNo, Oper3) ;
                CheckAddVariable(Oper1, QuadNo) ;
                CheckAddVariable(Oper2, QuadNo) |

   GotoOp     : ManipulateReference(QuadNo, Oper3) |

   (* variable references *)

   BitOp,
   UnboundedVarOp,
   UnboundedOp,
   HighOp,
   ReturnValueOp,
   FunctValueOp,
   OffsetOp,
   NegateOp,
   BecomesOp,
   CopyOp,
   SizeOp,
   AddrOp            : CheckAddVariable(Oper1, QuadNo) ;
                       CheckAddVariable(Oper3, QuadNo) |
   ReturnOp,
   CallOp,
   NewLocalVarOp,
   KillLocalVarOp    : CheckAddVariable(Oper3, QuadNo) |

   ParamOp           : CheckAddVariable(Oper2, QuadNo) ;
                       CheckAddVariable(Oper3, QuadNo) |

   BaseOp,
   MathOp,
   LogicalOrOp,
   LogicalAndOp,
   LogicalXorOp,
   LogicalDifferenceOp,
   BitRangeOp,
   CoerceOp,
   ConvertOp,
   AddOp,
   SubOp,
   MultOp,
   ModOp,
   DivOp             : CheckAddVariable(Oper1, QuadNo) ;
                       CheckAddVariable(Oper2, QuadNo) ;
                       CheckAddVariable(Oper3, QuadNo)


   ELSE
   END
END AddQuadInformation ;


(*
   PutQuad - overwrites a quadruple QuadNo with Op, Oper1, Oper2, Oper3
*)

PROCEDURE PutQuad (QuadNo: CARDINAL;
                   Op: QuadOperator;
                   Oper1, Oper2, Oper3: CARDINAL) ;
BEGIN
   EraseQuad(QuadNo) ;
   AddQuadInformation(QuadNo, Op, Oper1, Oper2, Oper3) ;
   WITH Quads[QuadNo] DO
      Operator := Op ;
      Operand1 := Oper1 ;
      Operand2 := Oper2 ;
      Operand3 := Oper3 ;
   END
END PutQuad ;


(*
   EraseQuad - erases a quadruple QuadNo, the quaduple is still in the list
               but wiped clean.
*)

PROCEDURE EraseQuad (QuadNo: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   WITH Quads[QuadNo] DO
      CASE Operator OF

      (* jumps, calls and branches *)
      IfInOp,
      IfNotInOp,
      IfEquOp,
      IfNotEquOp,
      IfLessOp,
      IfLessEquOp,
      IfGreOp,
      IfGreEquOp        : RemoveReference(QuadNo) ;
                          CheckRemoveVariable(Operand1, QuadNo) ;
                          CheckRemoveVariable(Operand2, QuadNo) |

      GotoOp            : RemoveReference(QuadNo) |

      (* variable references *)

      BitOp,
      UnboundedVarOp,
      UnboundedOp,
      HighOp,
      ReturnValueOp,
      FunctValueOp,
      OffsetOp,
      NegateOp,
      BecomesOp,
      CopyOp,
      SizeOp,
      AddrOp            : CheckRemoveVariable(Operand1, QuadNo) ;
                          CheckRemoveVariable(Operand3, QuadNo) |
      ReturnOp,
      CallOp,
      NewLocalVarOp,
      KillLocalVarOp    : CheckRemoveVariable(Operand3, QuadNo) |
      ParamOp           : CheckRemoveVariable(Operand2, QuadNo) ;
                          CheckRemoveVariable(Operand3, QuadNo) |

      BaseOp,
      MathOp,
      LogicalOrOp,
      LogicalAndOp,
      LogicalXorOp,
      LogicalDifferenceOp,
      BitRangeOp,
      CoerceOp,
      ConvertOp,
      AddOp,
      SubOp,
      MultOp,
      ModOp,
      DivOp             : CheckRemoveVariable(Operand1, QuadNo) ;
                          CheckRemoveVariable(Operand2, QuadNo) ;
                          CheckRemoveVariable(Operand3, QuadNo)

      ELSE
      END ;
      Operator := DummyOp ;   (* finally blank it out *)
      Operand1 := 0 ;
      Operand2 := 0 ;
      Operand3 := 0
   END
END EraseQuad ;


(*
   CheckAddVariable - checks to see whether symbol, Sym, is a variable and
                      if so it then adds this quadruple to the variable list.
*)

PROCEDURE CheckAddVariable (Sym: CARDINAL; Quad: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      PutVarQuad(Sym, Quad)
   END
END CheckAddVariable ;


(*
   CheckRemoveVariable - checks to see whether, Sym, is a variable and
                         if so then it removes the quadruple from the
                         variable list.
*)

PROCEDURE CheckRemoveVariable (Sym: CARDINAL; Quad: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      RemoveVarQuad(Sym, Quad)
   END
END CheckRemoveVariable ;


(*
   GetNextQuad - returns the Quadruple number following QuadNo.
*)

PROCEDURE GetNextQuad (QuadNo: CARDINAL) : CARDINAL ;
BEGIN
   QuadNo := Quads[QuadNo].Next ;
   WHILE (QuadNo#0) AND (Quads[QuadNo].Operator=DummyOp) DO
      QuadNo := Quads[QuadNo].Next
   END ;
   RETURN( QuadNo )
END GetNextQuad ;


(*
   AddQuad - adds a quadruple in between left and right then
             returns the new quad number.
*)

PROCEDURE AddQuad (VAR Left, Right: CARDINAL;
                   Op: QuadOperator;
                   Oper1, Oper2, Oper3: CARDINAL) : CARDINAL ;
VAR
   q: CARDINAL ;
BEGIN
   NewQuad(q) ;
   Assert(Quads[Left].Next=Right) ;
   Quads[Left].Next := q ;
   WITH Quads[q] DO
      Operator := Op ;
      Operand1 := Oper1 ;
      Operand2 := Oper2 ;
      Operand3 := Oper3 ;
      Next := Right
   END ;
   RETURN( q )
END AddQuad ;


(*
   SubQuad - subtracts a quadruple QuadNo from a list Head.
*)

PROCEDURE SubQuad (VAR Head: CARDINAL; QuadNo: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   AlterReference(Head, QuadNo, Quads[QuadNo].Next) ;
   WITH Quads[QuadNo] DO
      CASE Operator OF

      (* jumps, calls and branches *)
      IfInOp,
      IfNotInOp,
      IfEquOp,
      IfNotEquOp,
      IfLessOp,
      IfLessEquOp,
      IfGreOp,
      IfGreEquOp        : RemoveReference(QuadNo) ;
                          CheckRemoveVariable(Operand1, QuadNo) ;
                          CheckRemoveVariable(Operand2, QuadNo) |

      GotoOp            : RemoveReference(QuadNo) |

      (* variable references *)

      BitOp,
      UnboundedVarOp,
      UnboundedOp,
      HighOp,
      ReturnValueOp,
      FunctValueOp,
      OffsetOp,
      NegateOp,
      BecomesOp,
      CopyOp,
      SizeOp,
      AddrOp            : CheckRemoveVariable(Operand1, QuadNo) ;
                          CheckRemoveVariable(Operand3, QuadNo) |
      ReturnOp,
      CallOp,
      NewLocalVarOp,
      KillLocalVarOp    : CheckRemoveVariable(Operand3, QuadNo) |
      ParamOp           : CheckRemoveVariable(Operand2, QuadNo) ;
                          CheckRemoveVariable(Operand3, QuadNo) |

      BaseOp,
      MathOp,
      LogicalOrOp,
      LogicalAndOp,
      LogicalXorOp,
      LogicalDifferenceOp,
      BitRangeOp,
      CoerceOp,
      ConvertOp,
      AddOp,
      SubOp,
      MultOp,
      ModOp,
      DivOp             : CheckRemoveVariable(Operand1, QuadNo) ;
                          CheckRemoveVariable(Operand2, QuadNo) ;
                          CheckRemoveVariable(Operand3, QuadNo)

      ELSE
      END
   END ;
   IF Head=QuadNo
   THEN
      Head := Quads[QuadNo].Next
   ELSE
      i := Head ;
      WHILE Quads[i].Next#QuadNo DO
         i := Quads[i].Next
      END ;
      Quads[i].Next := Quads[QuadNo].Next
   END ;
   Quads[QuadNo].Operator := DummyOp
END SubQuad ;


(*
   GetRealQuad - returns the Quadruple number of the real quadruple
                 at QuadNo or beyond.
*)

PROCEDURE GetRealQuad (QuadNo: CARDINAL) : CARDINAL ;
BEGIN
   WHILE QuadNo#0 DO
      WITH Quads[QuadNo] DO
         IF Operator=EndOp
         THEN
            (*
               EndOp is not a real quadruple - but it marks the end
               of the quadruple list. So we stop here.
            *)
            RETURN( QuadNo )
         ELSIF (NOT IsPseudoQuad(QuadNo)) AND
            (Operator#DummyOp) AND (Operator#LineNumberOp)
         THEN
            RETURN( QuadNo )
         END
      END ;
      INC(QuadNo)
   END
END GetRealQuad ;


(*
   AlterReference - alters all references from OldQuad, to NewQuad in a
                    quadruple list Head.
*)

PROCEDURE AlterReference (Head, OldQuad, NewQuad: CARDINAL) ;
VAR
   OldOperand3,
   i          : CARDINAL ;
BEGIN
   WHILE (Quads[OldQuad].NoOfTimesReferenced>0) AND (Head#0) DO
      WITH Quads[Head] DO
         CASE Operator OF

         IfInOp,
         IfNotInOp,
         IfEquOp,
         IfNotEquOp,
         IfLessOp,
         IfLessEquOp,
         IfGreOp,
         IfGreEquOp,
         GotoOp     : IF Operand3=OldQuad
                      THEN
                         ManipulateReference(Head, NewQuad)
                      END

         ELSE
         END ;
         i := Next
      END ;
      Head := i
   END
END AlterReference ;


(*
   GrowQuads - grows the list of quadruples to the quadruple, to.
*)

PROCEDURE GrowQuads (to: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   IF (to#0) AND (to>GrowInitialization)
   THEN
      i := GrowInitialization+1 ;
      WHILE i<=to DO
         Quads[i].NoOfTimesReferenced := 0 ;
         INC(i)
      END ;
      GrowInitialization := to
   END
END GrowQuads ;


(*
   ManipulateReference - manipulates the quadruple, q, so that it now points to quad, to.
*)

PROCEDURE ManipulateReference (q: CARDINAL; to: CARDINAL) ;
BEGIN
   GrowQuads(to) ;
   RemoveReference(q) ;
   WITH Quads[q] DO
      Operand3 := to ;
      IF to#0
      THEN
         INC(Quads[to].NoOfTimesReferenced)
      END
   END
END ManipulateReference ;
   

(*
   RemoveReference - remove the reference by quadruple, q, to wherever
                     it was pointing to.
*)

PROCEDURE RemoveReference (q: CARDINAL) ;
BEGIN
   WITH Quads[q] DO
      IF (Operand3#0) AND (Operand3<NextQuad)
      THEN
         Assert(Quads[Operand3].NoOfTimesReferenced#0) ;
         DEC(Quads[Operand3].NoOfTimesReferenced)
      END
   END
END RemoveReference ;
   
   
(*
   CountQuads - returns the number of quadruple contained within the list Head.
*)

PROCEDURE CountQuads (Head: CARDINAL) : CARDINAL ;
VAR
   n: CARDINAL ;
BEGIN
   n := 0 ;
   WHILE Head#0 DO
      IF Quads[Head].Operator#DummyOp
      THEN
         INC(n)
      END ;
      Head := Quads[Head].Next
   END ;
   RETURN( n )
END CountQuads ;


(*
   KillQuad - destroys a quadruple.
*)

PROCEDURE KillQuad (QuadNo: CARDINAL) ;
BEGIN
   DisposeQuad(QuadNo)
END KillQuad ;


(*
   NewQuad - sets QuadNo to a new quadruple.
*)

PROCEDURE NewQuad (VAR QuadNo: CARDINAL) ;
BEGIN
   IF FreeList=MaxQuad
   THEN
      InternalError('no more quads available, increase MaxQuad', __FILE__, __LINE__)
   ELSE
      QuadNo := FreeList ;
      WITH Quads[QuadNo] DO
         (* must not set NoOfTimesReferenced := 0 ; as we might overwrite it *)
         Operator := DummyOp ;
         Operand3 := 0 ;
         Next := 0
      END ;
      INC(FreeList)
   END
END NewQuad ;


(*
   DisposeQuad - returns QuadNo to the FreeList of quadruples.
*)

PROCEDURE DisposeQuad (QuadNo: CARDINAL) ;
BEGIN
   InternalError('not expecting DisposeQuad to be called', __FILE__, __LINE__) ;
   Quads[QuadNo].Next := FreeList ;
   FreeList := QuadNo
END DisposeQuad ;


(*
   StartBuildDefFile - generates a StartFileDefOp quadruple indicating the file
                       that has produced the subsequent quadruples.
                       The code generator uses the StartDefFileOp quadruples
                       to relate any error to the appropriate file.


                       Entry                   Exit
                       =====                   ====


                Ptr ->                                        <- Ptr
                       +------------+          +------------+
                       | ModuleName |          | ModuleName | 
                       |------------|          |------------|


                       Quadruples Produced

                       q     StartDefFileOp  _  _  ModuleSym
*)

PROCEDURE StartBuildDefFile ;
VAR
   ModuleName: CARDINAL ;
BEGIN
   PopT(ModuleName) ;
   PushT(ModuleName) ;
   GenQuad(StartDefFileOp, GetPreviousTokenLineNo(), NulSym, GetModule(ModuleName))
END StartBuildDefFile ;


(*
   StartBuildModFile - generates a StartModFileOp quadruple indicating the file
                       that has produced the subsequent quadruples.
                       The code generator uses the StartModFileOp quadruples
                       to relate any error to the appropriate file.


                       Entry                   Exit
                       =====                   ====


                Ptr ->                                        <- Ptr
                       +------------+          +------------+
                       | ModuleName |          | ModuleName | 
                       |------------|          |------------|


                       Quadruples Produced

                       q     StartModFileOp  _  _  ModuleSym
*)

PROCEDURE StartBuildModFile ;
VAR
   ModuleName: CARDINAL ;
BEGIN
   PopT(ModuleName) ;
   PushT(ModuleName) ;
   GenQuad(StartModFileOp, GetPreviousTokenLineNo(), NulSym, GetModule(ModuleName))
END StartBuildModFile ;


(*
   EndBuildFile - generates an EndFileOp quadruple indicating the file
                  that has produced the previous quadruples has ended.

                  Entry                   Exit
                  =====                   ====


           Ptr ->                                        <- Ptr
                  +------------+          +------------+
                  | ModuleName |          | ModuleName | 
                  |------------|          |------------|


                  Quadruples Produced

                  q     EndFileOp  _  _  ModuleSym
*)

PROCEDURE EndBuildFile ;
VAR
   ModuleName: CARDINAL ;
BEGIN
   PopT(ModuleName) ;
   PushT(ModuleName) ;
   GenQuad(EndFileOp, NulSym, NulSym, GetModule(ModuleName))
END EndBuildFile ;


(*
   StartBuildModuleSuit - Starts the initialization order of modules.

                          Entry                   Exit
                          =====                   ====


                                                                  <- Ptr
                          Empty                   +------------+
                                                  | q          | 
                                                  |------------|


                          Quadruples Produced

                          q     GotoOp  _  _  0
*)

PROCEDURE StartBuildModuleSuit ;
BEGIN
   WriteString('StartBuildModuleSuit') ; WriteLn ;
   PushInit(NextQuad) ;
   GenQuad(GotoOp, NulSym, NulSym, 0)
END StartBuildModuleSuit ;

 
(*
   EndBuildModuleSuit - End the initialization order of modules.
 
                        Entry                   Exit
                        =====                   ====
 
 
                 Ptr ->                         Empty
                        +------------+
                        | q1         |   
                        |------------|
 
 
                        BackPatch q1 to point to q2
 
                        Quadruples Produced
 
                        q2    EndOp
*)
 
PROCEDURE EndBuildModuleSuit ;
VAR
   q1: CARDINAL ;
BEGIN
   PopInit(q1) ;
   BackPatch(q1, NextQuad) ;
   GenQuad(EndOp, GetPreviousTokenLineNo(), NulSym, NulSym) ;
   DisplayQuadList(Head)
END EndBuildModuleSuit ;


(*
   StartBuildInit - Sets the start of initialization code of the
                    current module to the next quadruple.
*)
 
PROCEDURE StartBuildInit ;
BEGIN
   PutModuleStartQuad(GetCurrentModule(), NextQuad) ;
   GenQuad(StartOp, GetPreviousTokenLineNo(), NulSym, GetCurrentModule())
END StartBuildInit ;
 
 
(*
   EndBuildInit - Sets the end initialization code of a module.
*)
 
PROCEDURE EndBuildInit ;
BEGIN
   PutModuleEndQuad(GetCurrentModule(), NextQuad) ;
   GenQuad(EndOp, GetPreviousTokenLineNo(), NulSym, GetCurrentModule())
END EndBuildInit ;


(*
   CheckSubrange - providing that the user has requested -bounds and GetType(Des) is a subrange
                   then this function emits quadruples to check that Exp lies in this subrange.
*)

PROCEDURE CheckSubrange (Des, Exp: CARDINAL) ;
VAR
   FileName : ARRAY [0..MaxString] OF CHAR ;
   num      : ARRAY [0..10] OF CHAR ;
   t, f,
   type     : CARDINAL ;
   old      : BOOLEAN ;
BEGIN
   IF BoundsChecking AND (NOT IsConst(Des)) AND (NOT MustNotCheckBounds)
   THEN
      old := MustNotCheckBounds ;
      MustNotCheckBounds := TRUE ;  (* stop recursive checking *)
      type := GetType(Des) ;
      IF IsSubrange(type)
      THEN
         (*
            we cannot yet GetSubrange for low and high as these entities
            may not yet have been evaluated. P3SymBuild evaluates them,
            possibly after this procedure.

            GetSubrange(type, high, low) ;  (* high, low may not be defined yet *)

            To work around this problem we simply leave the comparison
            to compare against the subrange type, which we will fill in
            or (backpatch) at a later stage - once the subranges are known.
         *)
         (* q+0  if exp < low        q+4  *)
         (* q+1  goto                q+2  *)
         PushT(Exp) ;        (* BuildRelOp   1st parameter *)
         PushT(LessTok) ;    (*              2nd parameter *)
         PushT(type) ;       (*              3rd parameter *)
         BuildRelOp ;
         PopBool(t, f) ;     (*              return value  *)
         BackPatch(f, NextQuad) ;
         (* q+2  if exp > high       q+4  *)
         (* q+3  goto                q+   *)
         PushT(Exp) ;        (* BuildRelOp   1st parameter *)
         PushT(GreaterTok) ; (*              2nd parameter *)
         PushT(type) ;       (*              3rd parameter *)
         BuildRelOp ;
         BackPatch(t, NextQuad) ;
         PopBool(t, f) ;
         BackPatch(t, NextQuad) ;
         (* now call the error routine *)
         IF IsSubscript(Des)
         THEN
            (* array index is out of bounds *)
            (* assignment to a subrange variable is out of bounds *)
            PushTF(RequestSym(MakeKey('ArraySubscriptError')),
                   GetType(RequestSym(MakeKey('ArraySubscriptError'))))
         ELSE
            (* assignment to a subrange variable is out of bounds *)
            PushTF(RequestSym(MakeKey('SubrangeAssignmentError')),
                   GetType(RequestSym(MakeKey('SubrangeAssignmentError'))))
         END ;
         GetFileName(FileName) ;
         PushT(MakeConstLitString(MakeKey(FileName))) ;
         CardToStr(GetLineNo(), 0, num) ;
         PushT(MakeConstLit(MakeKey(num))) ;
         PushT(2) ;
         BuildProcedureCall ;
         (* call error function           *)
         BackPatch(f, NextQuad) ;   (* no range error *)
      END ;
      MustNotCheckBounds := old   (* stop recursive checking *)
   END
END CheckSubrange ;


(*
   CheckDynamicArray - providing that the user has requested -bounds then this function
                       emits quadruples to check that Exp is legal: 0..HIGH(Sym).
                       We assume that Sym is an unbounded array.
*)

PROCEDURE CheckDynamicArray (Sym, Exp: CARDINAL) ;
VAR
   FileName : ARRAY [0..MaxString] OF CHAR ;
   num      : ARRAY [0..10] OF CHAR ;
   t, f,
   high     : CARDINAL ;
   old      : BOOLEAN ;
BEGIN
   IF BoundsChecking AND (NOT MustNotCheckBounds)
   THEN
      old := MustNotCheckBounds ;
      MustNotCheckBounds := TRUE ;  (* stop recursive checking *)
      (* q+0  if exp < low        q+4  *)
      (* q+1  goto                q+2  *)
      PushT(Exp) ;            (* BuildRelOp   1st parameter *)
      PushT(LessTok) ;        (*              2nd parameter *)
      PushT(MakeConstLit(MakeKey('0'))) ;  (* 3rd parameter *)
      BuildRelOp ;
      PopBool(t, f) ;     (*              return value  *)
      BackPatch(f, NextQuad) ;

      PushTF(High, Cardinal) ;             (* High function *)
      PushT(Sym) ;                       (* unbounded array *)
      PushT(1) ;                     (* 1 parameter to HIGH *)
      BuildFunctionCall ;
      PopT(high) ;

      (* q+2  if exp > high       q+4  *)
      (* q+3  goto                q+   *)
      PushT(Exp) ;           (* BuildRelOp   1st parameter *)
      PushT(GreaterTok) ;    (*              2nd parameter *)
      PushT(high) ;          (*              3rd parameter *)
      BuildRelOp ;
      BackPatch(t, NextQuad) ;
      PopBool(t, f) ;
      BackPatch(t, NextQuad) ;

      (* now call the error routine as the array index is out of bounds *)
      PushTF(RequestSym(MakeKey('ArraySubscriptError')),
             GetType(RequestSym(MakeKey('ArraySubscriptError')))) ;
      GetFileName(FileName) ;
      PushT(MakeConstLitString(MakeKey(FileName))) ;
      CardToStr(GetLineNo(), 0, num) ;
      PushT(MakeConstLit(MakeKey(num))) ;
      PushT(2) ;
      BuildProcedureCall ;
      (* call error function           *)
      BackPatch(f, NextQuad) ;   (* no range error *)
      MustNotCheckBounds := old   (* stop recursive checking *)
   END
END CheckDynamicArray ;


(*
   CollectLowIfSubrange - returns the low of the subrange value, providing, sym,
                          is a subrange. Otherwise return sym.
*)

PROCEDURE CollectLowIfSubrange (sym: CARDINAL) : CARDINAL ;
VAR
   low, high: CARDINAL ;
BEGIN
   IF IsSubrange(sym)
   THEN
      GetSubrange(sym, high, low) ;
      RETURN( low )
   ELSE
      RETURN( sym )
   END
END CollectLowIfSubrange ;


(*
   CollectHighIfSubrange - returns the high of the subrange value, providing, sym,
                           is a subrange. Otherwise return sym.
*)

PROCEDURE CollectHighIfSubrange (sym: CARDINAL) : CARDINAL ;
VAR
   low, high: CARDINAL ;
BEGIN
   IF IsSubrange(sym)
   THEN
      GetSubrange(sym, high, low) ;
      RETURN( high )
   ELSE
      RETURN( sym )
   END
END CollectHighIfSubrange ;


(*
   BackPatchSubranges - runs through all the quadruples and finds any conditional
                        quadruple which has as its second operand a subrange.
                        It alters:

                        Before:
                        If >   op1   op2    _              (* op2 is a subrange *)

                        After:
                        If >   op1   high   _

                        Before:
                        If <   op1   op2    _              (* op2 is a subrange *)

                        After:
                        If <   op1   low    _
*)

PROCEDURE BackPatchSubranges (Head: CARDINAL) ;
BEGIN
   IF BoundsChecking AND (Head#0)
   THEN
      REPEAT
         WITH Quads[Head] DO
            CASE Operator OF

            IfGreOp :  Operand2 := CollectHighIfSubrange(Operand2) |
            IfLessOp:  Operand2 := CollectLowIfSubrange(Operand2)

            ELSE
            END ;
            Head := Next
         END
      UNTIL Head=0
   END
END BackPatchSubranges ;


(*
   BuildAssignmentWithoutBounds - calls BuildAssignment but makes sure we do not
                                  check bounds.
*)

PROCEDURE BuildAssignmentWithoutBounds ;
VAR
   old: BOOLEAN ;
BEGIN
   old := MustNotCheckBounds ;
   MustNotCheckBounds := TRUE ;
   BuildAssignment ;
   MustNotCheckBounds := old
END BuildAssignmentWithoutBounds ;


(*
   BuildAssignment - Builds an assignment from the values given on the
                     quad stack. Either an assignment to an
                     arithmetic expression or an assignment to a
                     boolean expression.
                     The Stack is expected to contain:


       Either

                     Entry                   Exit
                     =====                   ====

              Ptr ->
                     +------------+
                     | Expression |
                     |------------|
                     | Designator |
                     |------------|          +------------+
                     |            |          |            |  <- Ptr
                     |------------|          |------------|


                     Quadruples Produced

                     q     BecomesOp  Designator  _  Expression

       OR

                     Entry                   Exit
                     =====                   ====

              Ptr ->
                     +------------+
                     | True |False|
                     |------------|
                     | Designator |
                     |------------|          +------------+
                     |            |          |            |  <- Ptr
                     |------------|          |------------|


                     Quadruples Produced

                     q     BecomesOp  Designator  _  TRUE
                     q+1   GotoOp                    q+3
                     q+2   BecomesOp  Designator  _  FALSE

*)

PROCEDURE BuildAssignment ;
VAR
   t, f,
   Des, Exp: CARDINAL ;
BEGIN
   IF IsBoolean(Ptr-1)
   THEN
      PopBool(t, f) ;
      PopT(Des) ;
      (* Conditional Boolean Assignment *)
      BackPatch(t, NextQuad) ;
      GenQuad(BecomesOp, Des, NulSym, True) ;
      GenQuad(GotoOp, NulSym, NulSym, NextQuad+2) ;
      BackPatch(f, NextQuad) ;
      GenQuad(BecomesOp, Des, NulSym, False)
   ELSE
      PopT(Exp) ;
      PopT(Des) ;
      CheckSubrange(Des, Exp) ;
      (* Traditional Assignment *)
      GenQuad(BecomesOp, Des, NulSym, Exp) ;

      (*
         Ugly hack - to determine whether an assignment is a CONST
                     string := 'asdkjhasd' ;
                     we test Exp here and set Des to ConstStringSym.
                     This test should be done during Pass 2 and it
                     should allow CONSTs to be declared after usage.
      *)
      IF IsConstString(Exp) AND IsConst(Des)
      THEN
         PutConstString(Des, GetString(Exp))
      END ;
      CheckAssignmentCompatible(Des, Exp)
   END
END BuildAssignment ;


(*
   CheckAssignmentCompatible - checks to see that an assignment is compatible.
*)

PROCEDURE CheckAssignmentCompatible (Des, Exp: CARDINAL) ;
VAR
   DesT, ExpT, DesL: CARDINAL ;
BEGIN
   DesT := GetType(Des) ;
   ExpT := GetType(Exp) ;
   DesL := GetLowestType(Des) ;
   IF IsProcedure(Exp) AND
      ((DesT#NulSym) AND (NOT IsProcType(DesT))) AND
      ((DesL#NulSym) AND (NOT IsProcType(DesL)))
   THEN
      WriteErrorFormat1('incorrectly assigning a procedure to a variable %s (variable is not a procedure type)', GetSymName(Des))
   ELSIF IsConstString(Exp)
   THEN
   ELSIF (DesT#NulSym) AND (IsUnbounded(DesT))
   THEN
   ELSIF (ExpT#NulSym) AND (IsUnbounded(ExpT))
   THEN
   ELSIF (DesT#NulSym) AND (IsEnumeration(DesT))
   THEN
      IF DesT#ExpT
      THEN
         WriteErrorFormat1('variable %s with enumeration type can only be assigned with the same type', GetSymName(Des))
      END
   ELSIF (DesL#NulSym) AND IsArray(DesL)
   THEN
   ELSIF IsConst(Exp) AND (ExpT#Address) AND (NOT IsConst(Des)) AND
         (DesL#NulSym) AND ((DesL=Cardinal) OR (NOT IsSubrange(DesL))) AND
         (NOT IsEnumeration(DesL))
   THEN
      IF NOT (IsBaseType(DesL) OR IsSystemType(DesL))
      THEN
         WriteErrorFormat1('assignment of a constant (%s) can only be made to a variable which has a Modula-2 reserved type as its base',
                           GetSymName(Des))
      END
   ELSIF (DesT#NulSym) AND IsSet(DesT) AND IsConst(Exp)
   THEN
      (*
         unfortunatly we cannot check types of these items yet as the Exp will be
         of type CARDINAL and DesT is a set. Exp may be generated by M2Quads.
         If we give it a set type (PutVar(Exp, DesT) I believe it will cause
         code gen problems as it looks at the type to decide whether OR or + (eg).

         So for the moment we ignore checking.
      *)
   ELSIF ((DesT#NulSym) AND ((DesT=Word) OR (DesT=Byte))) OR
         ((ExpT#NulSym) AND ((ExpT=Word) OR (ExpT=Byte)))
   THEN
      (*
         do not check if one side of the assignment has a WORD or BYTE type.
         The sizes of both sides is checked in the code generation phase.
         (Remember that we don't know the sizes of non base types at present)
      *)
   ELSE
      IF (DesT#NulSym) AND IsProcType(DesT) AND IsProcedure(Exp)
      THEN
         DesT := GetType(DesT)  (* we can at least check RETURN values of procedure variables *)
      END ;
      (* subranges must have TSIZE of between TSIZE(CHAR)..TSIZE(CARDINAL) *)

      (* Integer and Cardinal constants have special case as long as they are assigned to subranges
         of Integer and Cardinal, the problem is that we cannot tell yet what the type of the subrange is...
         This is an oversight which should be sorted out asap.. for now any constant integer or cardinal
         assignment to a subrange is legal..

         fixme
      *)

      CheckCompatibleTypes(ExpT, DesT)
   END
END CheckAssignmentCompatible ;


(*
   CheckBooleanId - Checks to see if the top operand is a boolean.
                    If the operand is not a boolean then it is tested
                    with true and a boolean is generated.
                    The Stack:


                    Entry                     Exit
             Ptr ->                                          <- Ptr
                    +------------+            +------------+
                    | Sym        |            | t   | f    |
                    |------------|            |------------|

                        Quadruples

                        q   If=      Sym   True   _
                        q+1 GotoOp   _     _      _
*)

PROCEDURE CheckBooleanId ;
BEGIN
   IF NOT IsBoolean(Ptr-1)
   THEN
      IF IsVar(Operand(Ptr-1))
      THEN
         IF GetType(Operand(Ptr-1))#Boolean
         THEN
            WriteErrorFormat1('symbol %s is not a boolean and boolean expression is expected', GetSymName(Operand(Ptr-1)))
         END
      END ;
      PushT(EqualTok) ;
      PushT(True) ;
      BuildRelOp
   END
END CheckBooleanId ;


(*
   BuildRepeat - Builds the repeat statement from the quad stack.
                 The Stack is expected to contain:


                 Entry                   Exit
                 =====                   ====


                 Empty
                                                        <- Ptr
                                         +------------+
                                         | RepeatQuad |
                                         |------------|

*)

PROCEDURE BuildRepeat ;
BEGIN
   PushT(NextQuad) ;
   PushExit(0)
END BuildRepeat ;


(*
   BuildUntil - Builds the until part of the repeat statement
                from the quad stack.
                The Stack is expected to contain:


                Entry                   Exit
                =====                   ====

        Ptr ->
                +------------+
                | t   | f    |
                |------------|
                | RepeatQuad |          Empty
                |------------|
*)

PROCEDURE BuildUntil ;
VAR
   t, f,
   Repeat: CARDINAL ;
BEGIN
   CheckBooleanId ;
   PopBool(t, f) ;
   PopT(Repeat) ;
   BackPatch(f, Repeat) ;          (* If False then keep on repeating *)
   BackPatch(t, NextQuad) ;        (* If True then exit repeat        *)
   BackPatch(PopExit(), NextQuad)  (* BackPatch any EXIT statements   *)
END BuildUntil ;


(*
   BuildWhile - Builds the While part of the While statement
                from the quad stack.
                The Stack is expected to contain:


                Entry                   Exit
                =====                   ====

                                                       <- Ptr
                                        |------------|
                Empty                   | WhileQuad  |
                                        |------------|
*)

PROCEDURE BuildWhile ;
BEGIN
   PushT(NextQuad) ;
   PushExit(0)           (* EXIT statement link *)
END BuildWhile ;


(*
   BuildDoWhile - Builds the Do part of the while statement
                  from the quad stack.
                  The Stack is expected to contain:


                  Entry                   Exit
                  =====                   ====

          Ptr ->
                  +------------+          +------------+
                  | t   | f    |          | 0    | f   |
                  |------------|          |------------|
                  | WhileQuad  |          | WhileQuad  |
                  |------------|          |------------|

                  Quadruples

                  BackPatch t exit to the NextQuad
*)

PROCEDURE BuildDoWhile ;
VAR
   t, f: CARDINAL ;
BEGIN
   CheckBooleanId ;
   PopBool(t, f) ;
   BackPatch(t, NextQuad) ;
   PushBool(0, f)
END BuildDoWhile ;


(*
   BuildEndWhile - Builds the end part of the while statement
                   from the quad stack.
                   The Stack is expected to contain:


                   Entry                   Exit
                   =====                   ====

           Ptr ->
                   +------------+
                   | t   | f    |
                   |------------|
                   | WhileQuad  |          Empty
                   |------------|

                   Quadruples

                   q    GotoOp  WhileQuad
                   False exit is backpatched with q+1
*)

PROCEDURE BuildEndWhile ;
VAR
   While,
   t, f : CARDINAL ;
BEGIN
   PopBool(t, f) ;
   Assert(t=0) ;
   PopT(While) ;
   GenQuad(GotoOp, NulSym, NulSym, While) ;
   BackPatch(f, NextQuad) ;
   BackPatch(PopExit(), NextQuad)   (* EXIT statements linked outside while *)
END BuildEndWhile ;


(*
   BuildLoop - Builds the Loop part of the Loop statement
               from the quad stack.
               The Stack is expected to contain:


               Entry                   Exit
               =====                   ====

                                                      <- Ptr
               Empty                   +------------+
                                       | LoopQuad   |
                                       |------------|
*)

PROCEDURE BuildLoop ;
BEGIN
   PushT(NextQuad) ;
   PushExit(0)         (* Seperate Exit Stack for loop end *)
END BuildLoop ;


(*
   BuildExit - Builds the Exit part of the Loop statement.
*)

PROCEDURE BuildExit ;
BEGIN
   IF ExitPtr>0
   THEN
      GenQuad(GotoOp, NulSym, NulSym, 0) ;
      PushExit(Merge(PopExit(), NextQuad-1))
   ELSE
      WriteError('EXIT - only allowed in LOOP, WHILE, REPEAT, FOR statements')
   END
END BuildExit ;


(*
   BuildEndLoop - Builds the End part of the Loop statement
                  from the quad stack.
                  The Stack is expected to contain:


                  Entry                   Exit
                  =====                   ====

          Ptr ->
                  +------------+
                  | LoopQuad   |          Empty
                  |------------|

                  Quadruples

                  Goto  _  _  LoopQuad
*)

PROCEDURE BuildEndLoop ;
VAR
   Loop: CARDINAL ;
BEGIN
   PopT(Loop) ;
   GenQuad(GotoOp, NulSym, NulSym, Loop) ;
   BackPatch(PopExit(), NextQuad)
END BuildEndLoop ;


(*
   BuildThenIf - Builds the Then part of the If statement
                 from the quad stack.
                 The Stack is expected to contain:


                 Entry                   Exit
                 =====                   ====

         Ptr ->                                          <- Ptr
                 +------------+          +------------+
                 | t   | f    |          | 0    | f   |
                 |------------|          |------------|

                 Quadruples

                 The true exit is BackPatched to point to
                 the NextQuad.
*)

PROCEDURE BuildThenIf ;
VAR
   t, f: CARDINAL ;
BEGIN
   CheckBooleanId ;
   PopBool(t, f) ;
   BackPatch(t, NextQuad) ;
   PushBool(0, f)
END BuildThenIf ;


(*
   BuildElse - Builds the Else part of the If statement
               from the quad stack.
               The Stack is expected to contain:


               Entry                   Exit
               =====                   ====

       Ptr ->
               +------------+          +------------+
               | t   | f    |          | t+q  | 0   |
               |------------|          |------------|

               Quadruples

               q    GotoOp  _  _  0
               q+1  <- BackPatched from f
*)

PROCEDURE BuildElse ;
VAR
   t, f: CARDINAL ;
BEGIN
   GenQuad(GotoOp, NulSym, NulSym, 0) ;
   PopBool(t, f) ;
   BackPatch(f, NextQuad) ;
   PushBool(Merge(t, NextQuad-1), 0)   (* NextQuad-1 = Goto Quad *)
END BuildElse ;


(*
   BuildEndIf - Builds the End part of the If statement
                from the quad stack.
                The Stack is expected to contain:


                Entry                   Exit
                =====                   ====

        Ptr ->
                +------------+
                | t   | f    |          Empty
                |------------|

                Quadruples

                Both t and f are backpatched to point to the NextQuad
*)

PROCEDURE BuildEndIf ;
VAR
   t, f: CARDINAL ;
BEGIN
   PopBool(t, f) ;
   BackPatch(t, NextQuad) ;
   BackPatch(f, NextQuad)
END BuildEndIf ;


(*
   BuildElsif1 - Builds the Elsif part of the If statement
                 from the quad stack.
                 The Stack is expected to contain:


                 Entry                   Exit
                 =====                   ====

         Ptr ->
                 +------------+          +------------+
                 | t   | f    |          | t+q  | 0   |
                 |------------|          |------------|

                 Quadruples

                 q    GotoOp  _  _  0
                 q+1  <- BackPatched from f
*)

PROCEDURE BuildElsif1 ;
VAR
   t, f: CARDINAL ;
BEGIN
   GenQuad(GotoOp, NulSym, NulSym, 0) ;
   PopBool(t, f) ;
   BackPatch(f, NextQuad) ;
   PushBool(Merge(t, NextQuad-1), 0)   (* NextQuad-1 = Goto Quad *)
END BuildElsif1 ;


(*
   BuildElsif2 - Builds the Elsif until part of the If statement
                 from the quad stack.
                 The Stack is expected to contain:


                 Entry                   Exit
                 =====                   ====

          Ptr ->
                 +--------------+
                 | 0    | f1    |                            <- Ptr
                 |--------------|          +---------------+
                 | t2   | f2    |          | t2    | f1+f2 |
                 |--------------|          |---------------|
*)

PROCEDURE BuildElsif2 ;
VAR
   t1, f1,
   t2, f2: CARDINAL ;
BEGIN
   PopBool(t1, f1) ;
   Assert(t1=0) ;
   PopBool(t2, f2) ;
   PushBool(t2, Merge(f1, f2))
END BuildElsif2 ;


(*
   BuildPseudoBy - Builds the Non existant part of the By
                   clause of the For statement
                   from the quad stack.
                   The Stack is expected to contain:


                   Entry                   Exit
                   =====                   ====

                                                           <- Ptr
                                           +------------+
                   Empty                   | BySym      |
                                           |------------|
*)

PROCEDURE BuildPseudoBy ;
BEGIN
   PushT(MakeConstLit(MakeKey('1')))
END BuildPseudoBy ;


(*
   BuildForToByDo - Builds the For To By Do part of the For statement
                    from the quad stack.
                    The Stack is expected to contain:


                    Entry                   Exit
                    =====                   ====


             Ptr ->                                        <- Ptr
                    +------------+          +------------+
                    | BySym      |          | t   | f    |
                    |------------|          |------------|
                    | e2         |          | ForQuad    |
                    |------------|          |------------|
                    | e1         |          | BySym      |
                    |------------|          |------------|
                    | Ident      |          | IdentSym   |
                    |------------|          |------------|

                    Quadruple

                    q   BecomesOp  IdentSym  _  e1
                    q+1 if <       by        0  q+5
                    q+2 GotoOp                  q+3
                    q+3 If >       IdentSym  e2 _
                    q+4 GotoOp                  q+7
                    q+5 If <=      IdentSym  e2 _
                    q+6 GotoOp                  q+7


                    The For Loop is regarded:

                    For ident := e1 To e2 By by Do

                    End
*)

PROCEDURE BuildForToByDo ;
VAR
   e1, e2,
   Id,
   IdSym,
   BySym,
   t, f  : CARDINAL ;
   t1, f1: CARDINAL ;
BEGIN
   PushExit(0) ;  (* Allows EXIT to be used to exit the for loop *)
   PopT(BySym) ;
   PopT(e2) ;
   PopT(e1) ;
   PopT(Id) ;
   IdSym := RequestSym(Id) ;
   CheckSubrange(IdSym, e1) ;
   GenQuad(BecomesOp, IdSym, NulSym, e1) ;
   PushT(IdSym) ;     (* Push information for future FOR reference *)
   PushT(BySym) ;
   PushT(NextQuad) ;
   (* q+1 if <       by        0  q+5 *)
   (* q+2 GotoOp                  q+3 *)
   PushT(BySym) ;           (* BuildRelOp  1st parameter *)
   PushT(LessTok) ;         (*             2nd parameter *)
   PushT(MakeConstLit(MakeKey('0'))) ;  (* 3rd parameter *)
   BuildRelOp ;
   PopBool(t, f) ;          (*             return value  *)
   BackPatch(f, NextQuad) ;
   (* q+3 If >       IdentSym  e2 _   *)
   (* q+4 GotoOp                  q+7 *)
   PushT(IdSym) ;           (* BuildRelOp  1st parameter *)
   PushT(GreaterTok) ;      (*             2nd parameter *)
   PushT(e2) ;              (*             3rd parameter *)
   BuildRelOp ;
   BackPatch(t, NextQuad) ; (* fixes q+1 Destimation to q+5 *)
   PopBool(t, f) ;          (* true = exits, false = in  *)
   PushT(IdSym) ;           (* BuildRelOp  1st parameter *)
   PushT(LessTok) ;         (*             2nd parameter *)
   PushT(e2) ;              (*             3nd parameter *)
   BuildRelOp ;
   BackPatch(f, NextQuad) ; (* Fixes q+4 GotoOp          *)
   PopBool(t1, f1) ;
   BackPatch(f1, NextQuad) ; (* Fixes q+6 GotoOp         *)
   PushBool(Merge(t1, t), 0) ;
         (* q+3 and q+5 provide the exit to the for loop *)
   CheckSubrange(IdSym, IdSym)
END BuildForToByDo ;


(*
   BuildEndFor - Builds the End part of the For statement
                 from the quad stack.
                 The Stack is expected to contain:


                 Entry                   Exit
                 =====                   ====

         Ptr ->
                 +------------+
                 | t   | f    |
                 |------------|
                 | ForQuad    |
                 |------------|
                 | BySym      |
                 |------------|
                 | IdSym      |          Empty
                 |------------|
*)

PROCEDURE BuildEndFor ;
VAR
   t, f,
   ForQuad,
   BySym,
   IdSym  : CARDINAL ;
BEGIN
   PopBool(t, f) ;
   Assert(f=0) ;
   PopT(ForQuad) ;
   PopT(BySym) ;
   PopT(IdSym) ;
   GenQuad(AddOp, IdSym, IdSym, BySym) ;
   GenQuad(GotoOp, NulSym, NulSym, ForQuad) ;
   BackPatch(t, NextQuad) ;
   BackPatch(PopExit(), NextQuad)
END BuildEndFor ;


(*
   BuildCaseStart - starts the case statement.
                    It initializes a backpatch list on the compile
                    time stack, the list is used to contain all
                    case break points. The list is later backpatched
                    and contains all positions of the case statement
                    which jump to the end of the case statement.
                    The stack also contains room for a boolean
                    expression, this is needed to allow , operator
                    in the CaseField alternatives.

                    The Stack is expected to contain:


                    Entry                   Exit
                    =====                   ====

                                                           <- Ptr
                                            +------------+
                    Empty                   | 0    | 0   |
                                            |------------|
                                            | 0    | 0   |
                                            |------------|
*)

PROCEDURE BuildCaseStart ;
BEGIN
   PushBool(0, 0) ;  (* BackPatch list initialized *)
   PushBool(0, 0)    (* Room for a boolean expression *)
END BuildCaseStart ;


(*
   BuildCaseStartStatementSequence - starts the statement sequence
                                     inside a case clause.
                                     BackPatches the true exit to the
                                     NextQuad.
                                     The Stack:

                                     Entry             Exit

                              Ptr ->                                  <- Ptr
                                     +-----------+     +------------+
                                     | t   | f   |     | 0   | f    |
                                     |-----------|     |------------|
*)

PROCEDURE BuildCaseStartStatementSequence ;
VAR
   t, f: CARDINAL ;
BEGIN
   PopBool(t, f) ;
   BackPatch(t, NextQuad) ;
   PushBool(0, f)
END BuildCaseStartStatementSequence ;


(*
   BuildCaseEndStatementSequence - ends the statement sequence
                                   inside a case clause.
                                   BackPatches the false exit f1 to the
                                   NextQuad.
                                   Asserts that t1 and f2 is 0
                                   Pushes t2+q and 0

                                   Quadruples:

                                   q  GotoOp  _  _  0

                                   The Stack:

                                   Entry             Exit

                            Ptr ->                                  <- Ptr
                                   +-----------+     +------------+
                                   | t1  | f1  |     | 0    | 0   |
                                   |-----------|     |------------|
                                   | t2  | f2  |     | t2+q | 0   |
                                   |-----------|     |------------|
*)

PROCEDURE BuildCaseEndStatementSequence ;
VAR
   t1, f1,
   t2, f2: CARDINAL ;
BEGIN
   GenQuad(GotoOp, NulSym, NulSym, 0) ;
   PopBool(t1, f1) ;
   PopBool(t2, f2) ;          (* t2 contains the break list for the case *)
   BackPatch(f1, NextQuad) ;  (* f1 no longer needed *)
   Assert(t1=0) ;
   Assert(f2=0) ;
   PushBool(Merge(t2, NextQuad-1), 0) ;  (* NextQuad-1 = Goto Quad *)
   PushBool(0, 0)             (* Room for boolean expression *)
END BuildCaseEndStatementSequence ;


(*
   BuildCaseRange - builds the range testing quaruples for
                    a case clause.

                    IF (e1>=ce1) AND (e1<=ce2)
                    THEN

                    ELS..

                    The Stack:

                    Entry             Exit

             Ptr ->
                    +-----------+
                    | ce2       |                   <- Ptr
                    |-----------|     +-----------+
                    | ce1       |     | t   | f   |
                    |-----------|     |-----------|
                    | t1  | f1  |     | t1  | f1  |
                    |-----------|     |-----------|
                    | t2  | f2  |     | t2  | f2  |
                    |-----------|     |-----------|
                    | e1        |     | e1        |
                    |-----------|     |-----------|
*)

PROCEDURE BuildCaseRange ;
VAR
   ce1, ce2,
   e1,
   t2, f2,
   t1, f1  : CARDINAL ;
BEGIN
   PopT(ce2) ;
   PopT(ce1) ;
   PopBool(t1, f1) ;
   PopBool(t2, f2) ;
   PopT(e1) ;
   PushT(e1) ;         (* leave e1 on bottom of stack when exit procedure *)
   PushBool(t2, f2) ;
   PushBool(t1, f1) ;  (* also leave t1 and f1 on the bottom of the stack *)
   PushT(e1) ;
   PushT(GreaterEqualTok) ;
   PushT(ce1) ;
   BuildRelOp ;
   PushT(AndTok) ;
   RecordOp ;
   PushT(e1) ;
   PushT(LessEqualTok) ;
   PushT(ce2) ;
   BuildRelOp ;
   BuildBinaryOp
END BuildCaseRange ;


(*
   BuildCaseEquality - builds the range testing quadruples for
                       a case clause.

                       IF e1=ce1
                       THEN

                       ELS..

                       The Stack:

                       Entry             Exit

                Ptr ->
                       +-----------+     +-----------+
                       | ce1       |     | t   | f   |
                       |-----------|     |-----------|
                       | t1  | f1  |     | t1  | f1  |
                       |-----------|     |-----------|
                       | t2  | f2  |     | t2  | f2  |
                       |-----------|     |-----------|
                       | e1        |     | e1        |
                       |-----------|     |-----------|
*)
 
PROCEDURE BuildCaseEquality ;
VAR
   ce1, e1,
   t2, f2,
   t1, f1 : CARDINAL ;
BEGIN
   PopT(ce1) ;
   PopBool(t1, f1) ;
   PopBool(t2, f2) ;
   PopT(e1) ;
   PushT(e1) ;        (* leave e1 on bottom of stack when exit procedure *)
   PushBool(t2, f2) ; (* also leave t2 and f2 on the bottom of the stack *)
   PushBool(t1, f1) ;
   PushT(e1) ;
   PushT(EqualTok) ;
   PushT(ce1) ;
   BuildRelOp
END BuildCaseEquality ;


(*
   BuildCaseList - merges two case tests into one

                   The Stack:
 
                   Entry             Exit
 
            Ptr ->
                   +-----------+
                   | t2  | f2  |
                   |-----------|     +-------------+
                   | t1  | f1  |     | t1+t2| f1+f2|
                   |-----------|     |-------------|
*)

PROCEDURE BuildCaseList ; 
VAR
   t2, f2,
   t1, f1: CARDINAL ;
BEGIN
   PopBool(t2, f2) ;
   PopBool(t1, f1) ;
   PushBool(Merge(t1, t2), Merge(f1, f2))
END BuildCaseList ;


(*
   BuildCaseOr - builds the , in the case clause.
 
                 The Stack:
 
                 Entry             Exit
 
          Ptr ->                                  <- Ptr
                 +-----------+     +------------+
                 | t   | f   |     | t    | 0   |
                 |-----------|     |------------|
*)
 
PROCEDURE BuildCaseOr ;
VAR 
   t, f: CARDINAL ;
BEGIN 
   PopBool(t, f) ; 
   BackPatch(f, NextQuad) ; 
   PushBool(t, 0)
END BuildCaseOr ; 


(*
   BuildCaseElse - builds the else of case clause.
 
                  The Stack:
 
                  Entry             Exit
 
           Ptr ->                                  <- Ptr
                  +-----------+     +------------+
                  | t   | f   |     | t    | 0   |
                  |-----------|     |------------|
*)
 
PROCEDURE BuildCaseElse ;
VAR 
   t, f: CARDINAL ;
BEGIN 
   PopBool(t, f) ; 
   BackPatch(f, NextQuad) ; 
   PushBool(t, 0)
END BuildCaseElse ; 


(* 
   BuildCaseEnd - builds the end of case clause.

                  The Stack:

                  Entry             Exit

           Ptr ->
                  +-----------+
                  | t1  | f1  |
                  |-----------|
                  | t2  | f2  |
                  |-----------|
                  | e1        |
                  |-----------|     Empty
*)

PROCEDURE BuildCaseEnd ;
VAR
   e1,
   t, f: CARDINAL ;
BEGIN
   PopBool(t, f) ;
   BackPatch(f, NextQuad) ;
   BackPatch(t, NextQuad) ;
   PopBool(t, f) ;
   BackPatch(f, NextQuad) ;
   BackPatch(t, NextQuad) ;
   PopT(e1)
END BuildCaseEnd ;


(*
   BuildNulParam - Builds a nul parameter on the stack.
                   The Stack:

                   Entry             Exit

                                                    <- Ptr
                   Empty             +------------+
                                     | 0          |
                                     |------------|
*)

PROCEDURE BuildNulParam ;
BEGIN
   PushT(0)
END BuildNulParam ;


(*
   BuildProcedureCall - builds a procedure call.
                        Although this procedure does not directly
                        destroy the procedure parameters, it calls
                        routine which will manipulate the stack and
                        so the entry and exit states of the stack are shown.

                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |
                        |----------------|
                        | Param 1        |
                        |----------------|
                        | Param 2        |
                        |----------------|
                        .                .
                        .                .
                        .                .
                        |----------------|
                        | Param #        |
                        |----------------|
                        | ProcSym | Type |         Empty
                        |----------------|
*)

PROCEDURE BuildProcedureCall ;
VAR
   NoOfParam,
   ProcSym,
   Ptr        : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   PushT(NoOfParam) ;  (* Compile time stack restored to entry state *)
   IF IsPseudoBaseProcedure(ProcSym)
   THEN
      BuildPseudoProcedureCall
   ELSE
      BuildRealProcedureCall
   END
END BuildProcedureCall ;


(*
   BuildRealProcedureCall - builds a real procedure call.
                            The Stack:


                            Entry                      Exit

                     Ptr ->
                            +----------------+
                            | NoOfParam      |
                            |----------------|
                            | Param 1        |
                            |----------------|
                            | Param 2        |
                            |----------------|
                            .                .
                            .                .
                            .                .
                            |----------------|
                            | Param #        |
                            |----------------|
                            | ProcSym | Type |         Empty
                            |----------------|
*)

PROCEDURE BuildRealProcedureCall ;
VAR
   NoOfParam,
   i,
   ProcSym,
   Proc,
   t, Ptr,
   true, false,
   Des        : CARDINAL ;
BEGIN
   CheckProcedureParameters ;
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   PushT(NoOfParam) ;  (* Restore stack to origional state *)
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF IsVar(ProcSym)
   THEN
      (* Procedure Variable ? *)
      Proc := Stack[Ptr].FalseExit
   ELSE
      Proc := ProcSym
   END ;
   IF GetType(Proc)#NulSym
   THEN
      NearTokens('trying to call a function and ignoring its return value',
                 GetTokenNo(), GetDeclared(Proc))
   END ;
   ManipulateParameters ;
   IF PushParametersLeftToRight
   THEN
      i := NoOfParam ;
      WHILE i>0 DO
         GenQuad(ParamOp, i, Proc, Operand(i+Ptr)) ;
         DEC(i)
      END
   ELSE
      i := 1 ;
      WHILE i<=NoOfParam DO
         GenQuad(ParamOp, i, Proc, Operand(i+Ptr)) ;
         INC(i)
      END
   END ;
   GenQuad(CallOp, NulSym, NulSym, ProcSym) ;
   PutPtr(Ptr)   (* Destroy arguments *)
END BuildRealProcedureCall ;


(*
   CheckProcedureParameters - Checks the parameters which are being passed to
                              procedure ProcSym.

                              The Stack:
   
   
                              Entry                      Exit
   
                       Ptr ->                                               <- Ptr
                              +----------------+         +----------------+
                              | NoOfParam      |         | NoOfParam      |
                              |----------------|         |----------------|
                              | Param 1        |         | Param 1        |
                              |----------------|         |----------------|
                              | Param 2        |         | Param 2        |
                              |----------------|         |----------------|
                              .                .         .                .
                              .                .         .                .
                              .                .         .                .
                              |----------------|         |----------------|
                              | Param #        |         | Param #        |
                              |----------------|         |----------------|
                              | ProcSym | Type |         | ProcSym | Type |
                              |----------------|         |----------------|

*)

PROCEDURE CheckProcedureParameters ;
VAR
   Unbounded   : BOOLEAN ;
   CallParam,
   ParamI,
   ParamIType,
   ParamTotal,
   TypeSym,
   Ptr,
   Proc,
   ProcSym,
   i, Var      : CARDINAL ;
BEGIN
   PopT(ParamTotal) ;
   GetPtr(Ptr) ;
   PushT(ParamTotal) ;  (* Restore stack to origional state *)
   DEC(Ptr, ParamTotal+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF IsVar(ProcSym)
   THEN
      (* Procedure Variable ? *)
      Proc := Stack[Ptr].FalseExit
   ELSE
      Proc := ProcSym
   END ;
   IF NOT (IsProcedure(Proc) OR IsProcType(Proc))
   THEN
      BeginError ;
      WriteString('undeclared symbol is: ') ; WriteKey(GetSymName(Proc)) ;
      WriteLn ;
      NearToken('undeclared procedure or procedure variable', GetFirstUsed(Proc)) ;
      EndError
   END ;
   IF CompilerDebugging
   THEN
      WriteString('  ') ; WriteKey(GetSymName(Proc)) ; WriteString(' ( ')
   END ;
   i := 1 ;
   WHILE i<=ParamTotal DO
      IF i<=NoOfParam(Proc)
      THEN
         ParamI := GetParam(Proc, i) ;
         IF CompilerDebugging
         THEN
            WriteKey(GetSymName(ParamI)) ; WriteString(': ') ; WriteKey(GetSymName(GetType(ParamI)))
         END ;
         IF IsBoolean(Ptr+i)
         THEN
            CheckParameter(Boolean, ParamI, Proc, i)
         ELSE
            CallParam := Operand(Ptr+i) ;
            IF IsConst(CallParam)
            THEN
               IF IsVarParam(Proc, i)
               THEN
                  FailParameter('trying to pass a constant to a VAR parameter',
                                CallParam, ParamI, Proc, i)
               ELSIF IsConstString(CallParam)
               THEN
                  IF (GetStringLength(CallParam) = 1)   (* if = 1 then it maybe treated as a char *)
                  THEN
                     CheckParameter(CallParam, ParamI, Proc, i)
                  ELSIF NOT IsUnboundedParam(Proc, i)
                  THEN
                     (* we possibly need to allow passing strings of exact size to an ARRAY [0..n] OF CHAR *)
                     FailParameter('cannot pass a string constant to a non unbounded array parameter',
                                   CallParam, ParamI, Proc, i)
                  END
               END
            ELSE
               CheckParameter(CallParam, ParamI, Proc, i)
            END
         END
      ELSE
         FailParameter('too many parameters', Operand(Ptr+i), NulSym, Proc, i)
      END ;
      INC(i) ;
      IF CompilerDebugging
      THEN
         IF i<=ParamTotal THEN WriteString('; ') ELSE WriteString(' ) ; ') ; WriteLn END
      END
   END
END CheckProcedureParameters ;


(*
   CheckParameter - checks that types CallType and ParamType are compatible for parameter
                    passing. ProcSym is the procedure and i is the parameter number.

                    We obey the following rules:

                    (1)  we allow WORD and BYTE to be compitable with any like sized
                         type.
                    (2)  we allow ADDRESS to be compatible with any pointer type.
                    (3)  we relax INTEGER and CARDINAL checking for Temporary variables.

                    Note that type sizes are checked during the code generation pass.
*)

PROCEDURE CheckParameter (Call, Param, ProcSym: CARDINAL; i: CARDINAL) ;
VAR
   CallType, ParamType: CARDINAL ;
BEGIN
   ParamType := GetType(Param) ;
   IF IsConstString(Call) AND (GetStringLength(Call) = 1)   (* if = 1 then it maybe treated as a char *)
   THEN
      CallType := Char
   ELSIF Call=Boolean
   THEN
      CallType := Call
   ELSE
      CallType := GetType(Call)
   END ;
   IF IsProcType(ParamType)
   THEN
      IF (NOT IsProcedure(Call)) AND ((CallType=NulSym) OR (NOT IsProcType(CallType)))
      THEN
         FailParameter('expecting a procedure or procedure variable as a parameter',
                       Call, Param, ProcSym, i)
      END ;
      (* we can check the return type of both proc types *)
      IF (CallType#NulSym) AND IsProcType(CallType)
      THEN
         IF ((GetType(CallType)#NulSym) AND (GetType(ParamType)=NulSym))
         THEN
            FailParameter('the item being passed is a function whereas the formal procedure parameter is a procedure',
                          Call, Param, ProcSym, i)
         ELSIF ((GetType(CallType)=NulSym) AND (GetType(ParamType)#NulSym))
         THEN
            FailParameter('the item being passed is a procedure whereas the formal procedure parameter is a function',
                          Call, Param, ProcSym, i)
         ELSIF NOT IsCompatible(GetType(CallType), GetType(ParamType))
         THEN
            FailParameter('the return result of the procedure variable parameter is not compatible with the return result of the item being passed',
                          Call, Param, ProcSym, i)
         END
      END
   ELSIF (CallType#ParamType) AND (CallType#NulSym)
   THEN
      IF IsUnknown(ParamType)
      THEN
         FailParameter('procedure parameter type is undeclared',
                       Call, Param, ProcSym, i)
      END ;
      IF IsUnbounded(CallType) AND (NOT IsUnboundedParam(ProcSym, i))
      THEN
         FailParameter('attempting to pass an unbounded array to an NON unbounded parameter',
                       Call, Param, ProcSym, i)
      ELSIF IsUnboundedParam(ProcSym, i)
      THEN
         IF IsUnbounded(CallType) OR IsArray(CallType)
         THEN
            CallType := GetType(CallType)
         END ;
         ParamType := GetType(ParamType)
      END ;
      IF CallType#ParamType
      THEN
         IF ((ParamType=Word)     OR  (ParamType=Byte))    OR
            ((CallType=Word)      OR  (CallType=Byte))     OR
            (IsPointer(ParamType) AND (CallType=Address))  OR
            (IsPointer(CallType)  AND (ParamType=Address)) OR
            IsTemporary(Call)
         THEN
            (* it is legal *)
(*
   was
            (
             ((CallType#Integer) AND (CallType#Cardinal)) OR
             (NOT IsTemporary(Call)) OR
             ((ParamType#Integer) AND (ParamType#Cardinal))
            )
*)
         ELSE
            FailParameter('incompatible type being passed to procedure',
                          Call, Param, ProcSym, i)
         END
      END
   END
END CheckParameter ;


(*
   FailParameter - generates an error message indicating that a parameter
                   declaration has failed.

                   The parameters are:

                   CurrentState  - string describing the current failing state.
                   Given         - the token that the source code provided.
                   Expecting     - token or identifier that was expected.
                   ParameterNo   - parameter number that has failed.
                   ProcedureSym  - procedure symbol where parameter has failed.

                   If any parameter is Nul then it is ignored.
*)

PROCEDURE FailParameter (CurrentState : ARRAY OF CHAR;
                         Given        : CARDINAL;
                         Expecting    : CARDINAL;
                         ProcedureSym : CARDINAL;
                         ParameterNo  : CARDINAL) ;
VAR
   First,
   ExpectType,
   GivenType,
   ReturnType: CARDINAL ;
BEGIN
   BeginError ;
   IF CompilingImplementationModule()
   THEN
      WriteString('Error found while compiling the implementation module') ; WriteLn
   ELSIF CompilingProgramModule()
   THEN
      WriteString('Error found while compiling the program module') ; WriteLn
   END ;
   WriteString('problem in parameter') ; WriteCard(ParameterNo, 4) ;
   WriteString(' :   PROCEDURE ') ;
   WriteKey(GetSymName(ProcedureSym)) ;
   WriteString(' (') ;
   IF NoOfParam(ProcedureSym)>=ParameterNo
   THEN
      IF ParameterNo>1
      THEN
         WriteString('.., ')
      END ;
      IF IsVarParam(ProcedureSym, ParameterNo)
      THEN
         WriteString('VAR ')
      END ;
      WriteKey(GetSymName(Expecting)) ;
      WriteString(': ') ;
      First := GetDeclared(GetNthParam(ProcedureSym, ParameterNo)) ;
      ExpectType := GetType(Expecting) ;
      IF IsUnboundedParam(ProcedureSym, ParameterNo)
      THEN
         WriteString('ARRAY OF ') ;
         WriteKey(GetSymName(GetType(ExpectType)))
      ELSE
         WriteKey(GetSymName(ExpectType))
      END ;
      IF ParameterNo<NoOfParam(ProcedureSym)
      THEN
         WriteString('; ...')
      END
   ELSE
      First := GetDeclared(ProcedureSym) ;
      IF NoOfParam(ProcedureSym)>0
      THEN
         WriteString('..')
      END
   END ;
   ReturnType := GetType(ProcedureSym) ;
   IF ReturnType=NulSym
   THEN
      WriteString(') ;')
   ELSE
      WriteString(') : ') ; WriteKey(GetSymName(ReturnType))
   END ;
   WriteLn ;
   WriteString('item being passed is     :   ') ;
   IF IsConstString(Given)
   THEN
      Write("'") ;
      WriteKey(GetSymName(Given)) ;
      Write("'")
   ELSE
      WriteKey(GetSymName(Given))
   END ;
   WriteString(': ') ;
   DisplayType(Given) ;
   WriteLn ;
   IF NOT StrEqual(CurrentState, '')
   THEN
      WriteString(CurrentState)
   END ;
   WriteLn ;
   NearTokens('parameter mismatch', First, GetTokenNo()) ;
   EndError
END FailParameter ;


(*
   DisplayType - displays a symbol, Sym, name and its type.
*)

PROCEDURE DisplayType (Sym: CARDINAL) ;
VAR
   i,
   Low, High,
   Subrange,
   Subscript,
   Type     : CARDINAL ;
BEGIN
   IF IsConstString(Sym)
   THEN
      IF (GetStringLength(Sym) = 1)   (* if = 1 then it maybe treated as a char *)
      THEN
         WriteString('(constant string) or CHAR')
      ELSE
         WriteString('(constant string)')
      END
   ELSIF IsConst(Sym)
   THEN
      WriteString('(constant)')
   ELSIF IsUnknown(Sym)
   THEN
      WriteString('(unknown)')
   ELSE
      Type := GetType(Sym) ;
      IF IsUnbounded(Type)
      THEN
         WriteString('ARRAY OF ') ;
         WriteKey(GetType(Type))
      ELSIF IsArray(Type)
      THEN
         WriteString('ARRAY [') ;
         i := 1 ;
         REPEAT
            Subscript := GetNth(Type, i) ;
            IF Subscript#NulSym
            THEN
               Assert(IsSubscript(Subscript)) ;
               Subrange := GetType(Subscript) ;
               IF NOT IsSubrange(Subrange)
               THEN
                  WriteString('Error with array ') ; WriteKey(GetSymName(Sym)) ;
                  WriteString(' subscript:') ; WriteCard(i, 4) ; WriteLn ;
                  WriteString('No Subrange for this subscript') ; WriteLn ;
                  WriteString('Instead the TYPE given is ') ;
                  WriteKey(GetSymName(Subrange)) ; WriteLn
               END ;
               Assert(IsSubrange(Subrange)) ;
               GetSubrange(Subrange, High, Low) ;
               WriteKey(GetSymName(Low)) ;
               WriteString('..') ;
               WriteKey(GetSymName(High)) ;
            END ;
            INC(i)
         UNTIL Subscript=NulSym ;
         WriteString('] OF ') ;
         WriteKey(GetSymName(GetType(Type)))
      ELSE
         WriteKey(GetSymName(Type))
      END
   END
END DisplayType ;


(*
   ManipulateParameters - manipulates the procedure parameters in
                          preparation for a procedure call.
                          Prepares Boolean, Unbounded and VAR parameters.
                          The Stack:


                          Entry                      Exit

                   Ptr ->
                          +----------------+
                          | NoOfParameters |
                          |----------------|
                          | Param 1        |
                          |----------------|
                          | Param 2        |
                          |----------------|
                          .                .
                          .                .
                          .                .
                          |----------------|
                          | Param #        |
                          |----------------|
                          | ProcSym | Type |         Empty
                          |----------------|
*)

PROCEDURE ManipulateParameters ;
VAR
   ErrorString: ARRAY [0..MaxString] OF CHAR ;
   ParamType,
   NoOfParameters,
   i,
   ProcSym,
   Proc,
   t, Ptr,
   true, false,
   Des        : CARDINAL ;
BEGIN
   PopT(NoOfParameters) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParameters+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF IsVar(ProcSym)
   THEN
      (* Procedure Variable ? *)
      Proc := Stack[Ptr].FalseExit
   ELSE
      Proc := ProcSym
   END ;
   IF NoOfParam(Proc)#NoOfParameters
   THEN
      GetKey(GetSymName(Proc), ErrorString) ;
      StrConCat('attempting to call procedure, ', ErrorString, ErrorString) ;
      StrConCat(ErrorString, ', with an incorrect number of parameters', ErrorString) ;
      NearTokens(ErrorString, GetTokenNo(), GetDeclared(Proc))
   END ;
   i := 1 ;
   WHILE i<=NoOfParameters DO
      IF i>NoOfParam(Proc)
      THEN
         WriteErrorFormat1('parameter not expected for procedure %s', GetSymName(Proc))
      ELSIF IsUnboundedParam(Proc, i)
      THEN
         t := MakeTemporary(OffsetAddr) ;
         PutVar(t, Unbounded) ;
         ParamType := GetType(GetType(GetParam(Proc, i))) ;
         IF IsVarParam(Proc, i)
         THEN
            AssignUnboundedVar(Operand(Ptr+i), t, ParamType)
         ELSE
            AssignUnboundedNonVar(Operand(Ptr+i), t, ParamType)
         END ;
         Stack[Ptr+i].TrueExit := t
      ELSIF IsVarParam(Proc, i)
      THEN
         (*
            Ok must reference by address
            - but we contain the type of the referenced entity
         *)
         t := MakeTemporary(ValueAddr) ;
         PushSize(Address) ;
         PutVarTypeAndSize(t, GetType(Operand(Ptr+i)), PopCard() ) ;
         GenQuad(AddrOp, t, NulSym, Operand(Ptr+i)) ;
         Stack[Ptr+i].TrueExit := t
      ELSIF IsBoolean(Ptr+i)
      THEN
         (* Ok Des will be a boolean type *)
         ConvertBooleanToVariable(Ptr+i)
      END ;
      INC(i)
   END
END ManipulateParameters ;


(*
   AssignUnboundedVar - assigns an Unbounded symbol fields,
                        ArrayAddress and ArrayHigh, from an array symbol.
                        UnboundedSym is not a VAR parameter and therefore
                        this procedure can complete both of the fields.
                        Sym can be a Variable with type Unbounded.
                        Sym can be a Variable with type Array.
                        Sym can be a String Constant.

                        ParamType is the TYPE of the paramater
*)

PROCEDURE AssignUnboundedVar (Sym, UnboundedSym, ParamType: CARDINAL) ;
VAR
   Type: CARDINAL ;
BEGIN
   WriteDebug('Inside AssignUnnboundedVar') ;
   IF IsConst(Sym)
   THEN
      WriteErrorFormat1('cannot pass constant %s as a VAR parameter', GetSymName(Sym))
   ELSIF IsVar(Sym)
   THEN
      Type := GetType(Sym) ;
      IF IsUnbounded(Type)
      THEN
         (* Copy Unbounded Symbol ie. UnboundedSym := Sym *)
         PushT(UnboundedSym) ;
         PushT(Sym) ;
         BuildAssignmentWithoutBounds ;
      ELSIF IsArray(Type) OR (ParamType=Word) OR (ParamType=Byte)
      THEN
         UnboundedVarLinkToArray(Sym, UnboundedSym, ParamType)
      ELSE
         WriteErrorFormat1('illegal type parameter %s: expecting Array or Dynamic Array', GetSymName(Sym)) ;
      END
   ELSE
      WriteErrorFormat1('illegal parameter %s which cannot be passed as: VAR ARRAY OF Type', GetSymName(Sym))
   END
   ; WriteDebug('Exit AssignUnnboundedVar') ;
END AssignUnboundedVar ;


(*
   AssignUnboundedNonVar - assigns an Unbounded symbol fields,
                           The difference between this procedure and
                           AssignUnboundedVar is that this procedure cannot
                           set the Unbounded.Address since the data from
                           Sym will be copied because parameter is NOT a VAR
                           parameter.
                           UnboundedSym is not a VAR parameter and therefore
                           this procedure can only complete the HIGH field
                           and not the ADDRESS field.
                           Sym can be a Variable with type Unbounded.
                           Sym can be a Variable with type Array.
                           Sym can be a String Constant.

                           ParamType is the TYPE of the paramater
*)

PROCEDURE AssignUnboundedNonVar (Sym, UnboundedSym, ParamType: CARDINAL) ;
VAR
   Field,
   AddressField,
   Type        : CARDINAL ;
BEGIN
   IF IsConst(Sym)  (* was IsConstString(Sym) *)
   THEN
      UnboundedNonVarLinkToArray(Sym, UnboundedSym, ParamType)
   ELSIF IsVar(Sym)
   THEN
      Type := GetType(Sym) ;
      IF IsUnbounded(Type)
      THEN
         UnboundedNonVarLinkToArray(Sym, UnboundedSym, ParamType)
      ELSIF IsArray(Type) OR (ParamType=Word) OR (ParamType=Byte)
      THEN
         UnboundedNonVarLinkToArray(Sym, UnboundedSym, ParamType)
      ELSE
         WriteErrorFormat1('illegal type parameter %s: expecting Array or Dynamic Array', GetSymName(Sym)) ;
      END
   ELSE
      WriteErrorFormat1('illegal parameter %s which cannot be passed as: VAR ARRAY OF Type', GetSymName(Sym))
   END
END AssignUnboundedNonVar ;


(*
   UnboundedNonVarLinkToArray - links an array, ArraySym, to an unbounded array,
                                UnboundedSym. The parameter is a NON VAR
                                variety.
*)

PROCEDURE UnboundedNonVarLinkToArray (ArraySym, UnboundedSym, ParamType: CARDINAL) ;
VAR
   t, f,
   AddressField,
   Field       : CARDINAL ;
BEGIN
   (* Unbounded.ArrayAddress := ??? runtime *)
   PushTF(UnboundedSym, GetType(UnboundedSym)) ;
   Field := GetLocalSym(Unbounded, ArrayAddress) ;
   PushTF(Field, GetType(Field)) ;
   PushT(1) ;
   BuildDesignatorRecord ;
   PopT(AddressField) ;
   GenQuad(UnboundedOp, AddressField, NulSym, ArraySym) ;
   (* Unbounded.ArrayHigh := HIGH(ArraySym)   *)
   PushTF(UnboundedSym, GetType(UnboundedSym)) ;
   Field := GetLocalSym(Unbounded, ArrayHigh) ;
   PushTF(Field, GetType(Field)) ;
   PushT(1) ;
   BuildDesignatorRecord ;
   IF (ParamType=Byte) OR (ParamType=Word)
   THEN
      (* SIZE(parameter) DIV TSIZE(ParamType)                   *)
      PushTF(TSize, Cardinal) ;  (* TSIZE(GetType(ArraySym))    *)
      PushT(GetType(ArraySym)) ;
      PushT(1) ;                (* 1 parameter for SIZE()       *)
      BuildFunctionCall ;
      PopTF(t, f) ;
      PushTF(t, Cardinal) ;
      PushT(DivTok) ;           (* Divide by                    *)
      PushTF(TSize, Cardinal) ; (* TSIZE(ParamType)             *)
      PushT(ParamType) ;
      PushT(1) ;                (* 1 parameter for TSIZE()      *)
      BuildFunctionCall ;
      BuildBinaryOp
   ELSE
      PushTF(High, Cardinal) ;  (* HIGH(ArraySym)               *)
      PushT(ArraySym) ;
      PushT(1) ;                (* 1 parameter for HIGH()       *)
      BuildFunctionCall
   END ;
   BuildAssignmentWithoutBounds
END UnboundedNonVarLinkToArray ;


(*
   UnboundedVarLinkToArray - links an array, ArraySym, to an unbounded array,
                             UnboundedSym. The parameter is a VAR variety.
*)

PROCEDURE UnboundedVarLinkToArray (ArraySym, UnboundedSym, ParamType: CARDINAL) ;
VAR
   Field: CARDINAL ;
BEGIN
   (* Unbounded.ArrayAddress := ADR(Sym) *)
   PushTF(UnboundedSym, GetType(UnboundedSym)) ;
   Field := GetLocalSym(Unbounded, ArrayAddress) ;
   PushTF(Field, GetType(Field)) ;
   PushT(1) ;
   BuildDesignatorRecord ;
   PushTF(Adr, Address) ;   (* ADR(ArraySym)                *)
   PushT(ArraySym) ;
   PushT(1) ;               (* 1 parameter for ADR()        *)
   BuildFunctionCall ;
   BuildAssignmentWithoutBounds ;
   (* Unbounded.ArrayHigh := HIGH(ArraySym)   *)
   PushTF(UnboundedSym, GetType(UnboundedSym)) ;
   Field := GetLocalSym(Unbounded, ArrayHigh) ;
   PushTF(Field, GetType(Field)) ;
   PushT(1) ;
   BuildDesignatorRecord ;
   IF (ParamType=Byte) OR (ParamType=Word)
   THEN
      (* SIZE(parameter) DIV TSIZE(ParamType)                   *)
      PushTF(TSize, Cardinal) ;  (* TSIZE(GetType(ArraySym))    *)
      PushT(GetType(ArraySym)) ;
      PushT(1) ;                (* 1 parameter for SIZE()       *)
      BuildFunctionCall ;
      PushT(DivTok) ;           (* Divide by                    *)
      PushTF(TSize, Cardinal) ; (* TSIZE(ParamType)             *)
      PushT(ParamType) ;
      PushT(1) ;                (* 1 parameter for TSIZE()      *)
      BuildFunctionCall ;
      BuildBinaryOp
   ELSE
      PushTF(High, Cardinal) ;  (* HIGH(ArraySym)               *)
      PushT(ArraySym) ;
      PushT(1) ;                (* 1 parameter for HIGH()       *)
      BuildFunctionCall
   END ;
   BuildAssignmentWithoutBounds
END UnboundedVarLinkToArray ;


(*
   BuildPseudoProcedureCall - builds a pseudo procedure call.
                              This procedure does not directly alter the
                              stack, but by calling routines the stack
                              will change in the following way when this
                              procedure returns.

                              The Stack:


                              Entry                      Exit

                       Ptr ->
                              +----------------+
                              | NoOfParam      |
                              |----------------|
                              | Param 1        |
                              |----------------|
                              | Param 2        |
                              |----------------|
                              .                .
                              .                .
                              .                .
                              |----------------|
                              | Param #        |
                              |----------------|
                              | ProcSym | Type |         Empty
                              |----------------|
*)

PROCEDURE BuildPseudoProcedureCall ;
VAR
   NoOfParam,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   PushT(NoOfParam) ;
   (* Compile time stack restored to entry state *)
   IF ProcSym=New
   THEN
      BuildNewProcedure
   ELSIF ProcSym=Dispose
   THEN
      BuildDisposeProcedure
   ELSIF ProcSym=Inc
   THEN
      BuildIncProcedure
   ELSIF ProcSym=Dec
   THEN
      BuildDecProcedure
   ELSIF ProcSym=Incl
   THEN
      BuildInclProcedure
   ELSIF ProcSym=Excl
   THEN
      BuildExclProcedure
   ELSE
      InternalError('pseudo procedure not implemented yet', __FILE__, __LINE__)
   END
END BuildPseudoProcedureCall ;


(*
   IsReallyAPointer - returns true if Sym is really a pointer.
*)

PROCEDURE IsReallyAPointer (Sym: CARDINAL) : BOOLEAN ;
BEGIN
   IF Sym=NulSym
   THEN
      RETURN( FALSE )
   ELSIF IsPointer(Sym)
   THEN
      RETURN( TRUE )
   ELSIF IsVar(Sym) OR IsType(Sym)
   THEN
      RETURN( IsReallyAPointer(GetType(Sym)) )
   ELSE
      RETURN( FALSE )
   END
END IsReallyAPointer ;


(*
   GetItemPointedTo - returns the symbol type that is being pointed to
                      by Sym.
*)

PROCEDURE GetItemPointedTo (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF IsPointer(Sym)
   THEN
      RETURN( GetType(Sym) )
   ELSIF IsVar(Sym) OR IsType(Sym)
   THEN
      RETURN( GetItemPointedTo(GetType(Sym)) )
   END
END GetItemPointedTo ;


(*
   BuildNewProcedure - builds the pseudo procedure call NEW.
                       This procedure is traditionally a "macro" for
                       NEW(x) --> ALLOCATE(x, SIZE(x^))
                       One method of implementation is to emulate a "macro"
                       processor by pushing the relevant input tokens
                       back onto the input stack.
                       However this causes two problems:

                       (i)  Unnecessary code is produced for x^
                       (ii) SIZE must be imported from SYSTEM
                       Therefore we chose an alternative method of
                       implementation;
                       generate quadruples for ALLOCATE(x, SIZE(x^))
                       this, although slightly more efficient,
                       is more complex and circumvents problems (i) and (ii).

                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |
                       |----------------|
                       | Param 1        |
                       |----------------|
                       | Param 2        |
                       |----------------|
                       .                .
                       .                .
                       .                .
                       |----------------|
                       | Param #        |
                       |----------------|
                       | ProcSym | Type |         Empty
                       |----------------|
*)
   
PROCEDURE BuildNewProcedure ;
VAR
   NoOfParam,
   SizeSym,
   PtrSym,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=1
   THEN
      ProcSym := RequestSym(MakeKey('ALLOCATE')) ;
      IF (ProcSym#NulSym) AND IsProcedure(ProcSym)
      THEN
         PtrSym := Operand(Ptr+1) ;
         IF IsReallyAPointer(PtrSym)
         THEN
            (*
               Build macro: ALLOCATE( PtrSym, SIZE(PtrSym^) )
            *)
            PushT(Size) ;            (* Procedure      *)
                                     (* x^             *)
            PushT(GetItemPointedTo(PtrSym)) ;
            PushT(1) ;               (* One parameter  *)
            BuildTSizeFunction ;
            PopT(SizeSym) ;

            PushT(ProcSym) ;         (* ALLOCATE       *)
            PushT(PtrSym) ;          (* x              *)
            PushT(SizeSym) ;         (* TSIZE(x^)      *)
            PushT(2) ;               (* Two parameters *)
            BuildProcedureCall

         ELSE
            WriteError('argument to NEW must be a pointer')
         END
      ELSE
         WriteError('ALLOCATE procedure not found for NEW substitution')
      END
   ELSE
      WriteError('the pseudo procedure NEW only has one parameter')
   END
END BuildNewProcedure ;


(*
   BuildDisposeProcedure - builds the pseudo procedure call DISPOSE.
                           This procedure is traditionally a "macro" for
                           DISPOSE(x) --> DEALLOCATE(x, SIZE(x^))
                           One method of implementation is to emulate a "macro"
                           processor by pushing the relevant input tokens
                           back onto the input stack.
                           However this causes two problems:

                           (i)  Unnecessary code is produced for x^
                           (ii) SIZE must be imported from SYSTEM
                           Therefore we chose an alternative method of
                           implementation;
                           generate quadruples for DEALLOCATE(x, SIZE(x^))
                           this, although slightly more efficient,
                           is more complex and circumvents problems (i)
                           and (ii).

                           The Stack:


                           Entry                      Exit

                    Ptr ->
                           +----------------+
                           | NoOfParam      |
                           |----------------|
                           | Param 1        |
                           |----------------|
                           | Param 2        |
                           |----------------|
                           .                .
                           .                .
                           .                .
                           |----------------|
                           | Param #        |
                           |----------------|
                           | ProcSym | Type |         Empty
                           |----------------|
*)

PROCEDURE BuildDisposeProcedure ;
VAR
   NoOfParam,
   SizeSym,
   PtrSym,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=1
   THEN
      ProcSym := RequestSym(MakeKey('DEALLOCATE')) ;
      IF (ProcSym#NulSym) AND IsProcedure(ProcSym)
      THEN
         PtrSym := Operand(Ptr+1) ;
         IF IsReallyAPointer(PtrSym)
         THEN
            (*
               Build macro: DEALLOCATE( PtrSym, SIZE(PtrSym^) )
            *)
            PushT(Size) ;            (* Procedure      *)
                                     (* x^             *)
            PushT(GetItemPointedTo(PtrSym)) ;
            PushT(1) ;               (* One parameter  *)
            BuildTSizeFunction ;
            PopT(SizeSym) ;

            PushT(ProcSym) ;         (* DEALLOCATE     *)
            PushT(PtrSym) ;          (* x              *)
            PushT(SizeSym) ;         (* TSIZE(x^)      *)
            PushT(2) ;               (* Two parameters *)
            BuildProcedureCall

         ELSE
            WriteError('argument to DISPOSE must be a pointer')
         END
      ELSE
         WriteError('DEALLOCATE procedure not found for DISPOSE substitution')
      END
   ELSE
      WriteError('the pseudo procedure DISPOSE only has one parameter')
   END
END BuildDisposeProcedure ;


(*
   BuildIncProcedure - builds the pseudo procedure call INC.
                       INC is a procedure which increments a variable.
                       It takes one or two parameters:
                       INC(a, b)  or  INC(a)
                       a := a+b   or  a := a+1

                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |
                       |----------------|
                       | Param 1        |
                       |----------------|
                       | Param 2        |
                       |----------------|
                       .                .
                       .                .
                       .                .
                       |----------------|
                       | Param #        |
                       |----------------|
                       | ProcSym | Type |         Empty
                       |----------------|
*)

PROCEDURE BuildIncProcedure ;
VAR
   NoOfParam,
   OperandSym,
   VarSym,
   TempSym,
   ProcSym,
   Ptr       : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   VarSym := Operand(Ptr+1) ;
   IF (NoOfParam=1) OR (NoOfParam=2)
   THEN
      IF IsVar(VarSym)
      THEN
         TempSym := MakeTemporary(OffsetAddr) ;
         PutVar(TempSym, GetType(VarSym)) ;
         IF NoOfParam=2
         THEN
            OperandSym := Operand(Ptr+2)
         ELSE
            OperandSym := MakeConstLit(MakeKey('1'))
         END ;
         GenQuad(BecomesOp, TempSym, NulSym, VarSym) ;

         IF IsPointer(GetType(TempSym))
         THEN
            PushTF(TempSym, Integer)   (* no coorsion for pointers *)
         ELSE
            PushTF(TempSym, GetType(TempSym))
         END ;

         PushT(PlusTok) ;
         PushTF(OperandSym, GetType(OperandSym)) ;
         BuildBinaryOp ;
         PopT(TempSym) ;
         CheckSubrange(VarSym, TempSym) ;
         GenQuad(BecomesOp, VarSym, NulSym, TempSym)
      ELSE
         WriteError('base procedure INC expects a variable as a parameter')
      END
   ELSE
      WriteError('base procedure INC expects 1 or 2 parameters')
   END
END BuildIncProcedure ;


(*
   BuildDecProcedure - builds the pseudo procedure call DEC.
                       DEC is a procedure which decrements a variable.
                       It takes one or two parameters:
                       DEC(a, b)  or  DEC(a)
                       a := a-b   or  a := a-1

                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |
                       |----------------|
                       | Param 1        |
                       |----------------|
                       | Param 2        |
                       |----------------|
                       .                .
                       .                .
                       .                .
                       |----------------|
                       | Param #        |
                       |----------------|
                       | ProcSym | Type |         Empty
                       |----------------|
*)

PROCEDURE BuildDecProcedure ;
VAR
   NoOfParam,
   OperandSym,
   VarSym,
   TempSym,
   ProcSym,
   Ptr       : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   VarSym := Operand(Ptr+1) ;
   IF (NoOfParam=1) OR (NoOfParam=2)
   THEN
      IF IsVar(VarSym)
      THEN
         TempSym := MakeTemporary(OffsetAddr) ;
         PutVar(TempSym, GetType(VarSym)) ;
         IF NoOfParam=2
         THEN
            OperandSym := Operand(Ptr+2)
         ELSE
            OperandSym := MakeConstLit(MakeKey('1'))
         END ;
         GenQuad(BecomesOp, TempSym, NulSym, VarSym) ;

         IF IsPointer(GetType(TempSym))
         THEN
            PushTF(TempSym, Integer)   (* no coorsion for pointers *)
         ELSE
            PushTF(TempSym, GetType(TempSym))
         END ;

         PushT(MinusTok) ;
         PushTF(OperandSym, GetType(OperandSym)) ;
         BuildBinaryOp ;
         PopT(TempSym) ;
         CheckSubrange(VarSym, TempSym) ;
         GenQuad(BecomesOp, VarSym, NulSym, TempSym)
      ELSE
         WriteError('base procedure DEC expects a variable as a parameter')
      END
   ELSE
      WriteError('base procedure DEC expects 1 or 2 parameters')
   END
END BuildDecProcedure ;


(*
   BuildInclProcedure - builds the pseudo procedure call INCL.
                        INCL is a procedure which adds bit b into a BITSET a.
                        It takes two parameters:
                        INCL(a, b)

                        a := a + {b}

                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |
                        |----------------|
                        | Param 1        |
                        |----------------|
                        | Param 2        |
                        |----------------|
                        | ProcSym | Type |         Empty
                        |----------------|
*)

PROCEDURE BuildInclProcedure ;
VAR
   NoOfParam,
   TempOperandSym,
   OperandSym,
   TempVarSym,
   VarSym,
   TempSym,
   ProcSym,
   Ptr           : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   VarSym := Operand(Ptr+1) ;
   IF NoOfParam=2
   THEN
      IF IsVar(VarSym)
      THEN
         IF IsSet(GetType(VarSym))
         THEN
            (* get value into TempVarSym *)
            TempVarSym := MakeTemporary(OffsetAddr) ;
            PutVar(TempVarSym, GetType(VarSym)) ;
            GenQuad(BecomesOp, TempVarSym, NulSym, VarSym) ;
            (* get value into TempOperandSym *)
            OperandSym := Operand(Ptr+2) ;
            TempOperandSym := MakeTemporary(AreConstant(IsConst(OperandSym))) ;
            PutVar(TempOperandSym, GetType(OperandSym)) ;
            GenQuad(BecomesOp, TempOperandSym, NulSym, OperandSym) ;
            (* now build stack for BuildElement *)
            PushTF(TempVarSym, GetType(TempVarSym)) ;
            PushT(TempOperandSym) ;
            BuildElement ;
            (* collect results from stack *)
            PopT(TempOperandSym) ;   (* new bit TempOperandSym *)
            PopT(TempVarSym) ;
            PushTF(TempVarSym, Bitset) ;  (* GetType(TempVarSym)) ; *)
            PushT(LogicalOrTok) ;
            PushTF(TempOperandSym, GetType(TempOperandSym)) ;
            BuildBinaryOp ;
            PopT(TempSym) ;
            GenQuad(BecomesOp, VarSym, NulSym, TempSym)
         ELSE
            WriteError('the first parameter to INCL must be a BITSET variable')
         END
      ELSE
         WriteError('base procedure INCL expects a variable as a parameter')
      END
   ELSE
      WriteError('base procedure INCL expects 2 parameters')
   END
END BuildInclProcedure ;


(*
   BuildExclProcedure - builds the pseudo procedure call EXCL.
                        INCL is a procedure which removes bit b from a BITSET a.
                        It takes two parameters:
                        EXCL(a, b)

                        a := a - {b}

                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |
                        |----------------|
                        | Param 1        |
                        |----------------|
                        | Param 2        |
                        |----------------|
                        | ProcSym | Type |         Empty
                        |----------------|
*)

PROCEDURE BuildExclProcedure ;
VAR
   NoOfParam,
   TempOperandSym,
   OperandSym,
   TempVarSym,
   VarSym,
   TempSym,
   ProcSym,
   Ptr           : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   VarSym := Operand(Ptr+1) ;
   IF NoOfParam=2
   THEN
      IF IsVar(VarSym)
      THEN
         IF IsSet(GetType(VarSym))
         THEN
            (* get value into TempVarSym *)
            TempVarSym := MakeTemporary(OffsetAddr) ;
            PutVar(TempVarSym, GetType(VarSym)) ;
            GenQuad(BecomesOp, TempVarSym, NulSym, VarSym) ;
            (* get value into TempOperandSym *)
            OperandSym := Operand(Ptr+2) ;
            TempOperandSym := MakeTemporary(AreConstant(IsConst(OperandSym))) ;
            PutVar(TempOperandSym, GetType(OperandSym)) ;
            GenQuad(BecomesOp, TempOperandSym, NulSym, OperandSym) ;
            (* now build stack for BuildElement *)
            PushTF(TempVarSym, GetType(TempVarSym)) ;
            PushT(TempOperandSym) ;
            BuildElement ;
            (* collect results from stack *)
            PopT(TempOperandSym) ;   (* new bit TempOperandSym *)
            PopT(TempVarSym) ;
            PushTF(TempVarSym, Bitset) ;  (* GetType(TempVarSym)) ; *)
            PushT(LogicalDifferenceTok) ;
            PushTF(TempOperandSym, GetType(TempOperandSym)) ;
            BuildBinaryOp ;
            PopT(TempSym) ;
            GenQuad(BecomesOp, VarSym, NulSym, TempSym)
         ELSE
            WriteError('the first parameter to EXCL must be a SET variable')
         END
      ELSE
         WriteError('base procedure EXCL expects a variable as a parameter')
      END
   ELSE
      WriteError('base procedure EXCL expects 2 parameters')
   END
END BuildExclProcedure ;


(*
   BuildFunctionCall - builds a function call.
                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |
                       |----------------|
                       | Param 1        |
                       |----------------|
                       | Param 2        |
                       |----------------|
                       .                .
                       .                .
                       .                .
                       |----------------|
                       | Param #        |                        <- Ptr
                       |----------------|         +------------+
                       | ProcSym | Type |         | ReturnVar  |
                       |----------------|         |------------|

*)

PROCEDURE BuildFunctionCall ;
VAR
   NoOfParam,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   PushT(NoOfParam) ;
   (* Compile time stack restored to entry state *)
   IF IsUnknown(ProcSym)
   THEN
      WriteErrorFormat1('function %s undefined', GetSymName(ProcSym))
   ELSIF IsAModula2Type(ProcSym)
   THEN
      BuildTypeCoercion
   ELSIF IsPseudoSystemFunction(ProcSym) OR
         IsPseudoBaseFunction(ProcSym) OR
         IsPseudoMathFunction(ProcSym)
   THEN
      BuildPseudoFunctionCall
   ELSE
      BuildRealFunctionCall
   END
END BuildFunctionCall ;


(*
   BuildTypeCoercion - builds the type coersion.
                       MODULA-2 allows types to be coersed with no runtime
                       penility.
                       It insists that the TSIZE(t1)=TSIZE(t2) where
                       t2 variable := t2(variable of type t1).
                       The ReturnVar on the stack is of type t2.

                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |  
                       |----------------|
                       | Param 1        |  
                       |----------------|
                       | Param 2        |  
                       |----------------|
                       .                .  
                       .                .  
                       .                .  
                       |----------------|
                       | Param #        |                        <- Ptr
                       |----------------|         +------------+
                       | ProcSym | Type |         | ReturnVar  |
                       |----------------|         |------------|

                       Quadruples:

                       CoerceOp  ReturnVar  Type  Param1

                       A type coercion will only be legal if the different
                       types have exactly the same size.
                       Since we can only decide this after M2Eval has processed
                       the symbol table then we create a quadruple explaining
                       the coercion taking place, the code generator can test
                       this assertion and report an error if the type sizes
                       differ.
*)

PROCEDURE BuildTypeCoercion ;
VAR
   NoOfParam,
   ReturnVar,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF NOT IsAModula2Type(ProcSym)
   THEN
      WriteError('coersion expecting a type')
   END ;
   ReturnVar := MakeTemporary(OffsetAddr) ;
   PutVar(ReturnVar, ProcSym) ;  (* Set ReturnVar's TYPE *)
   PutPtr(Ptr) ;
   IF NoOfParam=1
   THEN
      GenQuad(CoerceOp, ReturnVar, ProcSym, Operand(Ptr+1)) ;
      PushTF(ReturnVar, ProcSym)
(*
      ; WriteString('coesing type to ') ; WriteKey(GetSymName(ProcSym)) ;
      ; WriteLn
*)
   ELSE
      WriteError('only one parameter expected in a TYPE coersion')
   END
END BuildTypeCoercion ;


(*
   BuildRealFunctionCall - builds a function call.
                           The Stack:


                           Entry                      Exit

                    Ptr ->
                           +----------------+
                           | NoOfParam      |
                           |----------------|
                           | Param 1        |
                           |----------------|
                           | Param 2        |
                           |----------------|
                           .                .
                           .                .
                           .                .
                           |----------------|
                           | Param #        |                        <- Ptr
                           |----------------|         +------------+
                           | ProcSym | Type |         | ReturnVar  |
                           |----------------|         |------------|

*)

PROCEDURE BuildRealFunctionCall ;
VAR
   NoOfParam,
   i,
   ProcSym,
   Proc,
   ReturnVar,
   t, Ptr,
   true, false,
   Des        : CARDINAL ;
BEGIN
   CheckProcedureParameters ;
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   PushT(NoOfParam) ;  (* Restore stack to origional state *)
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF IsVar(ProcSym)
   THEN
      (* Procedure Variable ? *)
      Proc := Stack[Ptr].FalseExit
   ELSE
      Proc := ProcSym
   END ;
   IF GetType(Proc)=NulSym
   THEN
      WriteErrorFormat1('procedure %s does not have a return value - it is not a function', GetSymName(Proc))
   END ;
   ManipulateParameters ;
   GenQuad(ParamOp, 0, Proc, ProcSym) ;  (* Space for return value *)
   IF PushParametersLeftToRight
   THEN
      i := NoOfParam ;
      WHILE i>0 DO
         GenQuad(ParamOp, i, Proc, Operand(i+Ptr)) ;
         DEC(i)
      END
   ELSE
      i := 1 ;
      WHILE i<=NoOfParam DO
         GenQuad(ParamOp, i, Proc, Operand(i+Ptr)) ;
         INC(i)
      END
   END ;
   GenQuad(CallOp, NulSym, NulSym, ProcSym) ;
   PutPtr(Ptr) ;
   (* ReturnVar - will have the type of the procedure *)
   ReturnVar := MakeTemporary(OffsetAddr) ;
   PutVar(ReturnVar, GetType(Proc)) ;
   GenQuad(FunctValueOp, ReturnVar, NulSym, Proc) ;
   PushTF(ReturnVar, GetType(Proc))
END BuildRealFunctionCall ;


(*
   BuildPseudoFunctionCall - builds the pseudo function
                             The Stack:


                             Entry                      Exit

                      Ptr ->
                             +----------------+
                             | NoOfParam      |  
                             |----------------|
                             | Param 1        |  
                             |----------------|
                             | Param 2        |  
                             |----------------|
                             .                .  
                             .                .  
                             .                .  
                             |----------------|
                             | Param #        |                        <- Ptr
                             |----------------|         +------------+
                             | ProcSym | Type |         | ReturnVar  |
                             |----------------|         |------------|

*)

PROCEDURE BuildPseudoFunctionCall ;
VAR
   NoOfParam,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   PushT(NoOfParam) ;
   (* Compile time stack restored to entry state *)
   IF ProcSym=High
   THEN
      BuildHighFunction
   ELSIF ProcSym=Adr
   THEN
      BuildAdrFunction
   ELSIF ProcSym=Size
   THEN
      BuildSizeFunction
   ELSIF ProcSym=TSize
   THEN
      BuildTSizeFunction
   ELSIF ProcSym=Convert
   THEN
      BuildConvertFunction
   ELSIF ProcSym=Val
   THEN
      BuildValFunction
   ELSIF ProcSym=Chr
   THEN
      BuildChrFunction
   ELSIF ProcSym=Ord
   THEN
      BuildOrdFunction
   ELSIF ProcSym=Trunc
   THEN
      BuildTruncFunction
   ELSIF ProcSym=Float
   THEN
      BuildFloatFunction
   ELSIF IsPseudoMathFunction(ProcSym)
   THEN
      BuildPseudoMathFunction
   ELSE
      InternalError('pseudo Function not implemented yet', __FILE__, __LINE__)
   END ;
END BuildPseudoFunctionCall ;


(*
   BuildHighFunction - checks the stack in preparation for generating
                       quadruples which perform HIGH.
                       This procedure does not alter the stack but
                       determines whether, a, in HIGH(a) is an ArraySym
                       or UnboundedSym.
                       Both cases are different and appropriate quadruple
                       generating routines are called.

                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |  
                       |----------------|
                       | Param 1        |  
                       |----------------|
                       | Param 2        |  
                       |----------------|
                       .                .  
                       .                .  
                       .                .  
                       |----------------|
                       | Param #        |                        <- Ptr
                       |----------------|         +------------+
                       | ProcSym | Type |         | ReturnVar  |
                       |----------------|         |------------|

*)

PROCEDURE BuildHighFunction ;
VAR
   Type,
   NoOfParam,
   Param,
   ReturnVar,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym, Ptr+1 points to Param *)
   Param := Operand(Ptr+1) ;
   Type := GetType(Param) ;
(*
   ; WriteString('Attempting to build HIGH(') ; WriteKey(GetSymName(Param)) ;
   ; WriteString(')') ; WriteLn ;
*)
   (* Restore stack to origional form *)
   PushT(NoOfParam) ;
   IF NoOfParam#1
   THEN
      WriteError('base procedure HIGH expects 1 parameter')
   ELSIF (NOT IsVar(Param)) AND (NOT IsConstString(Param)) AND (NOT (IsConst(Param) AND (GetType(Param)=Char)))
   THEN
      WriteError('base procedure HIGH expects a variable or string constant as its parameter')
   ELSIF IsConst(Param) AND (GetType(Param)=Char)
   THEN
      BuildHighFromChar
   ELSIF IsConstString(Param)
   THEN
      BuildHighFromString
   ELSIF IsArray(Type) OR IsUnbounded(Type)
   THEN
      IF NoOfElements(Type)=1
      THEN
         IF IsArray(Type)
         THEN
            BuildHighFromArray
         ELSIF IsUnbounded(Type)
         THEN
            BuildHighFromUnbounded
         END
      ELSE
         WriteError('not allowed multidemension array parameter for HIGH')
      END
   ELSE
      WriteError('base procedure HIGH expects a variable of type array as its parameter')
   END
END BuildHighFunction ;


(*
   BuildHighFromArray - builds the pseudo function HIGH from an ArraySym.
                        ArraySyms are not dynamic and therefore it can be
                        calculated at compile time since the subscripts
                        are constant.

                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |  
                        |----------------|
                        | Param 1        |  
                        |----------------|
                        | Param 2        |  
                        |----------------|
                        .                .  
                        .                .  
                        .                .  
                        |----------------|
                        | Param #        |                        <- Ptr
                        |----------------|         +------------+
                        | ProcSym | Type |         | ReturnVar  |
                        |----------------|         |------------|

*)

PROCEDURE BuildHighFromArray ;
VAR
   NoOfParam,
   ReturnVar,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ReturnVar := MakeTemporary(ImmediateAddr) ;
   GenQuad(HighOp, ReturnVar, NulSym, Operand(Ptr+1)) ;
   PutPtr(Ptr) ;
   PushT(ReturnVar)
END BuildHighFromArray ;


(*
   BuildHighFromUnbounded - builds the pseudo function HIGH from an
                            UnboundedSym.
                            This is achieved by a hidden record access.
                            HIGH(a) is performed by a._ArrayHigh and therefore
                            the compile time stack is manipulated to be
                            in the form for a record access,
                            BuildDesignatorRecord is called and the stack
                            is finally left in the form shown below.

                            The Stack:


                            Entry                      Exit

                     Ptr ->
                            +----------------+
                            | NoOfParam      |  
                            |----------------|
                            | Param 1        |  
                            |----------------|
                            | Param 2        |  
                            |----------------|
                            .                .  
                            .                .  
                            .                .  
                            |----------------|
                            | Param #        |                        <- Ptr
                            |----------------|         +------------+
                            | ProcSym | Type |         | ReturnVar  |
                            |----------------|         |------------|

*)

PROCEDURE BuildHighFromUnbounded ;
VAR
   NoOfParam,
   Param1,
   ReturnVar,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   Param1 := Operand(Ptr+1) ;
   (*
      Build record access: Param1._ArrayHigh
   *)
   PutPtr(Ptr) ;
   PushTF(Param1, GetType(Param1)) ;
   PushTF(GetLocalSym(Unbounded, ArrayHigh),
          GetType(GetLocalSym(Unbounded, ArrayHigh))) ;
   PushT(1) ;             (* 1 field for Param1._ArrayHigh *)
   BuildDesignatorRecord  (* Now the generate quadruples   *)
END BuildHighFromUnbounded ;


(*
   BuildHighFromString - builds the pseudo function HIGH from an StringSym.
                         Constant Strings are not dynamic and therefore it can
                         be calculated at compile time since the subscripts
                         are constant.

                         The Stack:


                         Entry                      Exit

                  Ptr ->
                         +----------------+
                         | NoOfParam      |  
                         |----------------|
                         | Param 1        |  
                         |----------------|
                         | Param 2        |  
                         |----------------|
                         .                .  
                         .                .  
                         .                .  
                         |----------------|
                         | Param #        |                        <- Ptr
                         |----------------|         +------------+
                         | ProcSym | Type |         | ReturnVar  |
                         |----------------|         |------------|

*)

PROCEDURE BuildHighFromString ;
VAR
   High,
   NoOfParam,
   ReturnVar,
   Ptr      : CARDINAL ;
   a        : ARRAY [0..5] OF CHAR ;  (* up to 999999 constant characters long! *)
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   High := GetStringLength(Operand(Ptr+1)) ;
   IF High>0
   THEN
      DEC(High)
   END ;
   CardToStr(High, 0, a) ;
   ReturnVar := MakeConstLit(MakeKey(a)) ;
   PutPtr(Ptr) ;
   PushT(ReturnVar)
END BuildHighFromString ;


(*
   BuildHighFromChar - builds the pseudo function HIGH from a character constant.
                       Character constants can be passed to ARRAY OF CHAR and thus
                       we need to calculate HIGH when we construct an unbounded sym.
                       This function simply manipulates the stack and places
                       constant 0 on the stack. HIGH(any char)=0

                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |  
                       |----------------|
                       | Param 1        |  
                       |----------------|
                       | Param 2        |  
                       |----------------|
                       .                .  
                       .                .  
                       .                .  
                       |----------------|
                       | Param #        |                        <- Ptr
                       |----------------|         +------------+
                       | ProcSym | Type |         | ReturnVar  |
                       |----------------|         |------------|
*)

PROCEDURE BuildHighFromChar ;
VAR
   NoOfParam,
   ReturnVar,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ReturnVar := MakeConstLit(MakeKey('0')) ;
   PutPtr(Ptr) ;
   PushT(ReturnVar)
END BuildHighFromChar ;


(*
   BuildChrFunction - builds the pseudo procedure call CHR.
                      This procedure is actually a "macro" for
                      CHR(x) --> CONVERT(CHAR, x)
                      However we cannot push tokens back onto the input stack
                      because the compiler is currently building a function
                      call and expecting a ReturnVar on the stack.
                      Hence we manipulate the stack and call
                      BuildConvertFunction.

                      The Stack:


                      Entry                      Exit

               Ptr ->
                      +----------------+
                      | NoOfParam      |
                      |----------------|
                      | Param 1        |
                      |----------------|
                      | Param 2        |
                      |----------------|
                      .                .
                      .                .
                      .                .
                      |----------------|
                      | Param #        |
                      |----------------|
                      | ProcSym | Type |         Empty
                      |----------------|
*)

PROCEDURE BuildChrFunction ;
VAR
   NoOfParam,
   Var,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=1
   THEN
      ProcSym := RequestSym(MakeKey('CONVERT')) ;
      IF (ProcSym#NulSym) AND IsProcedure(ProcSym)
      THEN
         Var := Operand(Ptr+1) ;
         IF IsVar(Var) OR IsConst(Var)
         THEN
            (*
               Build macro: CONVERT( CHAR, Var )
            *)
            PushTF(ProcSym, NulSym) ;
            PushT(Char) ;
            PushT(Var) ;
            PushT(2) ;          (* Two parameters *)
            BuildConvertFunction
         ELSE
            WriteError('argument to CHR must be a variable or constant')
         END
      ELSE
         WriteError('CONVERT procedure not found for CHR substitution')
      END
   ELSE
      WriteError('the pseudo procedure CHR only has one parameter')
   END
END BuildChrFunction ;


(*
   BuildOrdFunction - builds the pseudo procedure call ORD.
                      This procedure is actually a "macro" for
                      ORD(x) --> CONVERT(CARDINAL, x)
                      However we cannot push tokens back onto the input stack
                      because the compiler is currently building a function
                      call and expecting a ReturnVar on the stack.
                      Hence we manipulate the stack and call
                      BuildConvertFunction.

                      The Stack:


                      Entry                      Exit

               Ptr ->
                      +----------------+
                      | NoOfParam      |
                      |----------------|
                      | Param 1        |
                      |----------------|
                      | Param 2        |
                      |----------------|
                      .                .
                      .                .
                      .                .
                      |----------------|
                      | Param #        |
                      |----------------|
                      | ProcSym | Type |         Empty
                      |----------------|
*)

PROCEDURE BuildOrdFunction ;
VAR
   NoOfParam,
   Var,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=1
   THEN
      ProcSym := RequestSym(MakeKey('CONVERT')) ;
      IF (ProcSym#NulSym) AND IsProcedure(ProcSym)
      THEN
         IF IsBoolean(Ptr+1)
         THEN
            ConvertBooleanToVariable(Ptr+1)
         END ;
         Var := Operand(Ptr+1) ;
         IF IsVar(Var) OR IsConst(Var)
         THEN
            (*
               Build macro: CONVERT( CARDINAL, Var )
            *)
            PushTF(ProcSym, NulSym) ;
            PushT(Cardinal) ;
            PushT(Var) ;
            PushT(2) ;          (* Two parameters *)
            BuildConvertFunction
         ELSE
            WriteError('argument to ORD must be a variable or constant')
         END
      ELSE
         WriteError('CONVERT procedure not found for ORD substitution')
      END
   ELSE
      WriteError('the pseudo procedure ORD only has one parameter')
   END
END BuildOrdFunction ;


(*
   BuildValFunction - builds the pseudo procedure call VAL.
                      This procedure is actually a "macro" for
                      VAL(Type, x) --> CONVERT(Type, x)
                      However we cannot push tokens back onto the input stack
                      because the compiler is currently building a function
                      call and expecting a ReturnVar on the stack.
                      Hence we manipulate the stack and call
                      BuildConvertFunction.

                      The Stack:


                      Entry                      Exit

               Ptr ->
                      +----------------+
                      | NoOfParam      |
                      |----------------|
                      | Param 1        |
                      |----------------|
                      | Param 2        |
                      |----------------|
                      .                .
                      .                .
                      .                .
                      |----------------|
                      | Param #        |
                      |----------------|
                      | ProcSym | Type |         Empty
                      |----------------|
*)

PROCEDURE BuildValFunction ;
VAR
   NoOfParam,
   Var, Type,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=2
   THEN
      ProcSym := RequestSym(MakeKey('CONVERT')) ;
      IF (ProcSym#NulSym) AND IsProcedure(ProcSym)
      THEN
         Type := Operand(Ptr+1) ;
         IF IsBoolean(Ptr+2)
         THEN
            ConvertBooleanToVariable(Ptr+2)
         END ;
         Var := Operand(Ptr+2) ;
         IF IsUnknown(Type)
         THEN
            WriteErrorFormat1('undeclared type found in VAL (%s)', GetSymName(Type))
         ELSIF (IsSet(Type) OR IsEnumeration(Type) OR IsSubrange(Type) OR IsType(Type)) AND
               (IsVar(Var) OR IsConst(Var))
         THEN
            (*
               Build macro: CONVERT( Type, Var )
            *)
            PushTF(ProcSym, NulSym) ;
            PushT(Type) ;
            PushT(Var) ;
            PushT(2) ;          (* Two parameters *)
            BuildConvertFunction
         ELSE
            WriteError('arguments to VAL must be (Type, Variable or Constant)')
         END
      ELSE
         WriteError('CONVERT procedure not found for VAL substitution')
      END
   ELSE
      WriteError('the pseudo procedure VAL has 2 parameters, a Type and Variable')
   END
END BuildValFunction ;


(*
   BuildConvertFunction - builds the pseudo function CONVERT.
                          CONVERT( Type, Variable ) ;

                          The Stack:


                          Entry                      Exit

                   Ptr ->
                          +----------------+
                          | NoOfParam      |  
                          |----------------|
                          | Param 1        |  
                          |----------------|
                          | Param 2        |  
                          |----------------|
                          .                .  
                          .                .  
                          .                .  
                          |----------------|
                          | Param #        |                                 <- Ptr
                          |----------------|         +---------------------+
                          | ProcSym | Type |         | ReturnVar | Param1  |
                          |----------------|         |---------------------|

                          Quadruples:

                          ConvertOp  ReturnVar  Param1  Param2

                          Converts variable Param2 into a variable Param1
                          with a type Param1.
*)

PROCEDURE BuildConvertFunction ;
VAR
   Var, Type,
   NoOfParam,
   ProcSym,
   IntegerVar,
   ReturnVar,
   Ptr       : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF NoOfParam#2
   THEN
      WriteError('base procedure CONVERT expects 2 parameters')
   ELSE
      Type := Operand(Ptr+1) ;
      IF IsBoolean(Ptr+2)
      THEN
         ConvertBooleanToVariable(Ptr+2)
      END ;
      Var := Operand(Ptr+2) ;
      IF IsUnknown(Type)
      THEN
         WriteErrorFormat1('undeclared type found in CONVERT (%s)', GetSymName(Type))
      ELSIF IsUnknown(Var)
      THEN
         WriteErrorFormat1('undeclared variable found in CONVERT (%s)', GetSymName(Var))
      ELSIF (IsSet(Type) OR IsEnumeration(Type) OR IsSubrange(Type) OR IsType(Type)) AND
            (IsVar(Var) OR IsConst(Var))
      THEN
         ReturnVar := MakeTemporary(OffsetAddr) ;
         PutVar(ReturnVar, Type) ;
         IF (GetType(Var)#Integer) AND (GetType(Var)#Cardinal) AND
            (GetType(Var)#Word) AND (GetType(Var)#Address) AND
            (IsCoprocessorType(ReturnVar) OR IsCoprocessorType(Var))
         THEN
            (* all coprocessor type conversions must go through INTEGER *)
            IntegerVar := MakeTemporary(OffsetAddr) ;
            PutVar(IntegerVar, Integer) ;
            GenQuad(ConvertOp, IntegerVar, Integer, Var) ;
            GenQuad(ConvertOp, ReturnVar, Type, IntegerVar)
         ELSE
            GenQuad(ConvertOp, ReturnVar, Type, Var)
         END ;
         PutPtr(Ptr) ;
         PushTF(ReturnVar, Type)
      ELSE
         WriteError('base procedure CONVERT expects 2 parameters, a type and a variable')
      END
   END
END BuildConvertFunction ;


(*
   BuildTruncFunction - builds the pseudo procedure call TRUNC.
                        This procedure is actually a "macro" for
                        TRUNC(x) --> CONVERT(CARDINAL, x)
                        However we cannot push tokens back onto the input stack
                        because the compiler is currently building a function
                        call and expecting a ReturnVar on the stack.
                        Hence we manipulate the stack and call
                        BuildConvertFunction.

                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |
                        |----------------|
                        | Param 1        |
                        |----------------|
                        | Param 2        |
                        |----------------|
                        .                .
                        .                .
                        .                .
                        |----------------|
                        | Param #        |
                        |----------------|
                        | ProcSym | Type |         Empty
                        |----------------|
*)

PROCEDURE BuildTruncFunction ;
VAR
   NoOfParam,
   Var,
   ReturnVar,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   Assert(Operand(Ptr)=Trunc) ;
   IF NoOfParam=1
   THEN
      Var := Operand(Ptr+1) ;
      IF IsVar(Var) OR IsConst(Var)
      THEN
         IF (GetType(Var)=LongReal) OR (GetType(Var)=Real)
         THEN
            ReturnVar := MakeTemporary(OffsetAddr) ;
            PutVar(ReturnVar, Cardinal) ;
            GenQuad(MathOp, Operand(Ptr), ReturnVar, Var) ;

            PushTF(ReturnVar, Cardinal)
         ELSE
            WriteError('argument to TRUNC must have type REAL or LONGREAL')
         END
      ELSE
         WriteError('argument to TRUNC must be a variable or constant')
      END
   ELSE
      WriteError('the pseudo procedure TRUNC only has one parameter')
   END
END BuildTruncFunction ;


(*
   BuildFloatFunction - builds the pseudo procedure call FLOAT.
                        This procedure is actually a "macro" for
                        FLOAT(x) --> CONVERT(REAL, x)
                        However we cannot push tokens back onto the input stack
                        because the compiler is currently building a function
                        call and expecting a ReturnVar on the stack.
                        Hence we manipulate the stack and call
                        BuildConvertFunction.

                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |
                        |----------------|
                        | Param 1        |
                        |----------------|
                        | Param 2        |
                        |----------------|
                        .                .
                        .                .
                        .                .
                        |----------------|
                        | Param #        |
                        |----------------|
                        | ProcSym | Type |         Empty
                        |----------------|
*)

PROCEDURE BuildFloatFunction ;
VAR
   NoOfParam,
   Var,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=1
   THEN
      ProcSym := RequestSym(MakeKey('CONVERT')) ;
      IF (ProcSym#NulSym) AND IsProcedure(ProcSym)
      THEN
         Var := Operand(Ptr+1) ;
         IF IsVar(Var) OR IsConst(Var)
         THEN
            (*
               Build macro: CONVERT( REAL, Var )
            *)
            PushTF(ProcSym, NulSym) ;
            PushT(Real) ;
            PushT(Var) ;
            PushT(2) ;          (* Two parameters *)
            BuildConvertFunction
         ELSE
            WriteError('argument to TRUNC must be a variable or constant')
         END
      ELSE
         WriteError('CONVERT procedure not found for TRUNC substitution')
      END
   ELSE
      WriteError('the pseudo procedure TRUNC only has one parameter')
   END
END BuildFloatFunction ;


(*
   BuildPseudoMathFunction - builds the pseudo math functions.

                             The Stack:

                             Entry                      Exit

                      Ptr ->
                             +----------------+
                             | NoOfParam      |
                             |----------------|
                             | Param 1        |
                             |----------------|
                             | Param 2        |
                             |----------------|
                             .                .
                             .                .
                             .                .
                             |----------------|
                             | Param #        |
                             |----------------|
                             | ProcSym | Type |         Empty
                             |----------------|
*)

PROCEDURE BuildPseudoMathFunction ;
VAR
   NoOfParam,
   Var,
   ReturnVar,
   ProcSym,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   PutPtr(Ptr) ;   (* Destroy the compile time procedure stack *)
   IF NoOfParam=1
   THEN
      Var := Operand(Ptr+1) ;
      IF (IsVar(Var) OR IsConst(Var)) AND IsCoprocessorType(Var)
      THEN
         ReturnVar := MakeTemporary(OffsetAddr) ;
         PutVar(ReturnVar, GetType(Var)) ;
         GenQuad(MathOp, Operand(Ptr), ReturnVar, Var) ;

         PushTF(ReturnVar, GetType(Var))
      ELSE
         WriteError('argument to math function must be a variable or constant with type REAL or LONGREAL')
      END
   ELSE
      WriteError('this math pseudo procedure only has one parameter')
   END
END BuildPseudoMathFunction ;


(*
   BuildAdrFunction - builds the pseudo function ADR
                      The Stack:


                      Entry                      Exit

               Ptr ->
                      +----------------+
                      | NoOfParam      |  
                      |----------------|
                      | Param 1        |  
                      |----------------|
                      | Param 2        |  
                      |----------------|
                      .                .  
                      .                .  
                      .                .  
                      |----------------|
                      | Param #        |                        <- Ptr
                      |----------------|         +------------+
                      | ProcSym | Type |         | ReturnVar  |
                      |----------------|         |------------|

*)

PROCEDURE BuildAdrFunction ;
VAR
   NoOfParam,
   ProcSym,
   ReturnVar,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF NoOfParam#1
   THEN
      WriteError('SYSTEM procedure ADR expects 1 parameter')
   ELSIF NOT IsVar(Operand(Ptr+1))
   THEN
      WriteError('SYSTEM procedure ADR expects a variable as its parameter')
   ELSE
      ReturnVar := MakeTemporary(OffsetAddr) ;
      PutVar(ReturnVar, GetType(ProcSym)) ;
      GenQuad(AddrOp, ReturnVar, NulSym, Operand(Ptr+1)) ;
      PutPtr(Ptr) ;
      PushTF(ReturnVar, GetType(ReturnVar))
   END
END BuildAdrFunction ;


(*
   BuildSizeFunction - builds the pseudo function SIZE
                       The Stack:


                       Entry                      Exit

                Ptr ->
                       +----------------+
                       | NoOfParam      |  
                       |----------------|
                       | Param 1        |  
                       |----------------|
                       | Param 2        |  
                       |----------------|
                       .                .  
                       .                .  
                       .                .  
                       |----------------|
                       | Param #        |                        <- Ptr
                       |----------------|         +------------+
                       | ProcSym | Type |         | ReturnVar  |
                       |----------------|         |------------|

*)

PROCEDURE BuildSizeFunction ;
VAR
   Type,
   UnboundedSym,
   Field,
   NoOfParam,
   ProcSym,
   ReturnVar,
   Ptr         : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF NoOfParam#1
   THEN
      WriteError('SYSTEM procedure SIZE expects 1 parameter')
   ELSIF IsVar(Operand(Ptr+1))
   THEN
      Type := GetType(Operand(Ptr+1)) ;
      IF IsUnbounded(Type)
      THEN
         (* eg. SIZE(a)  ; where a is unbounded dereference HIGH and multiply by the TYPE *)
         UnboundedSym := Operand(Ptr+1) ;
         PushTF(UnboundedSym, GetType(UnboundedSym)) ;
         Field := GetLocalSym(Unbounded, ArrayHigh) ;
         PushTF(Field, GetType(Field)) ;
         PushT(1) ;
         BuildDesignatorRecord ;
         PushT(TimesTok) ;
         PushTF(TSize, Cardinal) ; (* TSIZE(GetType(Unbounded))    *)
         PushT(GetType(Type)) ;
         PushT(1) ;                (* 1 parameter for TSIZE()      *)
         BuildFunctionCall ;
         BuildBinaryOp ;
         PopT(ReturnVar)
      ELSE
         ReturnVar := MakeTemporary(ImmediateAddr) ;
         GenQuad(SizeOp, ReturnVar, NulSym, Operand(Ptr+1))
      END
   ELSE
      WriteError('SYSTEM procedure SIZE expects a variable as its parameter')
   END ;
   PutPtr(Ptr) ;
   PushTF(ReturnVar, GetType(ProcSym))
END BuildSizeFunction ;


(*
   BuildTSizeFunction - builds the pseudo function TSIZE
                        The Stack:


                        Entry                      Exit

                 Ptr ->
                        +----------------+
                        | NoOfParam      |  
                        |----------------|
                        | Param 1        |  
                        |----------------|
                        | Param 2        |  
                        |----------------|
                        .                .  
                        .                .  
                        .                .  
                        |----------------|
                        | Param #        |                        <- Ptr
                        |----------------|         +------------+
                        | ProcSym | Type |         | ReturnVar  |
                        |----------------|         |------------|

*)

PROCEDURE BuildTSizeFunction ;
VAR
   NoOfParam,
   ProcSym,
   ReturnVar,
   Ptr      : CARDINAL ;
BEGIN
   PopT(NoOfParam) ;
   GetPtr(Ptr) ;
   DEC(Ptr, NoOfParam+1) ;
   (* Ptr points to the ProcSym *)
   ProcSym := Operand(Ptr) ;
   IF NoOfParam#1
   THEN
      WriteError('SYSTEM procedure TSIZE expects 1 parameter')
   ELSIF IsAModula2Type(Operand(Ptr+1))
   THEN
      ReturnVar := MakeTemporary(ImmediateAddr) ;
      GenQuad(SizeOp, ReturnVar, NulSym, Operand(Ptr+1)) ;
      PutPtr(Ptr) ;
      PushTF(ReturnVar, GetType(ProcSym))
   ELSIF IsVar(Operand(Ptr+1))
   THEN
      ReturnVar := MakeTemporary(ImmediateAddr) ;
      GenQuad(SizeOp, ReturnVar, NulSym, GetType(Operand(Ptr+1))) ;
      PutPtr(Ptr) ;
      PushTF(ReturnVar, GetType(ProcSym))
   ELSE
      WriteError('SYSTEM procedure TSIZE expects a type or variable as its parameter')
   END
END BuildTSizeFunction ;


(*
   BuildProcedureStart - Builds start of the procedure. Creates space for
                         the local variables.
                         The Stack is expected to contain:


                         Entry                   Exit
                         =====                   ====

                 Ptr ->                                       <- Ptr
                        +------------+          +-----------+
                        | ProcSym    |          | ProcSym   |
                        |------------|          |-----------|
                        | Name       |          | Name      |
                        |------------|          |-----------|


                        Quadruples:

                        q   NewLocalVarOp  Line#  _  ProcSym
*)

PROCEDURE BuildProcedureStart ;
VAR
   ProcSym: CARDINAL ;
BEGIN
   PopT(ProcSym) ;
   Assert(IsProcedure(ProcSym)) ;
   PutProcedureStartQuad(ProcSym, NextQuad) ;
   GenQuad(NewLocalVarOp, GetPreviousTokenLineNo(), NulSym, ProcSym) ;
   CurrentProc := ProcSym ;
   IF ReturnQuad#0
   THEN
      WriteError('implementation restriction - nested procedures not allowed')
   END ;
   PushT(ProcSym)
END BuildProcedureStart ;


(*
   BuildProcedureEnd - Builds end of the procedure. Destroys space for
                       the local variables.
                       The Stack is expected to contain:


                       Entry                   Exit
                       =====                   ====

                Ptr ->                                       <- Ptr
                       +------------+          +-----------+
                       | ProcSym    |          | ProcSym   |
                       |------------|          |-----------|
                       | Name       |          | Name      |
                       |------------|          |-----------|


                       Quadruples:

                       q   KillLocalVarOp  _  _  ProcSym
*)

PROCEDURE BuildProcedureEnd ;
VAR
   ProcSym: CARDINAL ;
BEGIN
   PopT(ProcSym) ;
   BackPatch(ReturnQuad, NextQuad) ;
   ReturnQuad := 0 ;
   CurrentProc := NulSym ;
   GenQuad(KillLocalVarOp, NulSym, NulSym, ProcSym) ;
   PutProcedureEndQuad(ProcSym, NextQuad) ;
   GenQuad(ReturnOp, NulSym, NulSym, ProcSym) ;
   CheckFunctionReturn(ProcSym) ;
   PushT(ProcSym)
END BuildProcedureEnd ;


(*
   CheckFunctionReturn - checks to see that a RETURN statement was present in a function.
*)

PROCEDURE CheckFunctionReturn (ProcSym: CARDINAL) ;
VAR
   Op           : QuadOperator ;
   Op1, Op2, Op3,
   Start, End   : CARDINAL ;
BEGIN
   IF GetType(ProcSym)#NulSym
   THEN
      (* yes it is a function *)
      GetProcedureQuads(ProcSym, Start, End) ;
      GetQuad(Start, Op, Op1, Op2, Op3) ;
      WHILE (Start#End) AND (Op#ReturnValueOp) AND (Op#InlineOp) DO
         Start := GetNextQuad(Start) ;
         GetQuad(Start, Op, Op1, Op2, Op3)
      END ;
      IF (Op#ReturnValueOp) AND (Op#InlineOp)
      THEN
         (* an InlineOp can always be used to emulate a RETURN *)
         WriteErrorFormat1('function %s does not RETURN a value', GetSymName(ProcSym))
      END
   END
END CheckFunctionReturn ;


(*
   BuildReturn - Builds the Return part of the procedure.
                 The Stack is expected to contain:


                 Entry                   Exit
                 =====                   ====

         Ptr ->
                 +------------+
                 | e1         |          Empty
                 |------------|
*)

PROCEDURE BuildReturn ;
VAR
   e1, t1,
   t, f,
   Des   : CARDINAL ;
BEGIN
   IF IsBoolean(Ptr-1)
   THEN
      PopBool(t, f) ;
      (* Des will be a boolean type *)
      Des := MakeTemporary(OffsetAddr) ;
      PutVar(Des, Boolean) ;
      PushTF(Des, Boolean) ;
      PushBool(t, f) ;
      BuildAssignmentWithoutBounds ;
      PushTF(Des, Boolean)
   END ;
   PopTF(e1, t1) ;
   IF e1#NulSym
   THEN
      IF GetType(CurrentProc)=NulSym
      THEN
         WriteErrorFormat1('attempting to RETURN a value from a procedure (%s) and not a function', GetSymName(CurrentProc))
      ELSIF NOT IsCompatible(t1, GetType(CurrentProc))
      THEN
         WriteErrorFormat2('attempying to RETURN a value with an incompatible type (%s) to the function (%s)',
                           GetSymName(t1), GetSymName(GetType(CurrentProc)))
      END ;
      GenQuad(ReturnValueOp, e1, NulSym, CurrentProc)
   END ;
   GenQuad(GotoOp, NulSym, NulSym, ReturnQuad) ;
   ReturnQuad := NextQuad-1
END BuildReturn ;


(*
   BuildDesignatorRecord - Builds the record referencing.
                           The Stack is expected to contain:


                           Entry                   Exit
                           =====                   ====

                   Ptr ->
                           +--------------+
                           | n            |
                           |--------------|
                           | fld1 | type1 |
                           |--------------|
                           .              .
                           .              .
                           .              .
                           |--------------|
                           | fldn | typen |                        <- Ptr
                           |--------------|        +-------------+
                           | Sym  | Type  |        | S    | type1|
                           |--------------|        |-------------|
*)

PROCEDURE BuildDesignatorRecord ;
VAR
   n, i,
   Sym, Type,
   adr, Res,
   t1, t2, t3: CARDINAL ;
BEGIN
   PopT(n) ;
   DEC(Ptr, n+1) ;  (* Ptr points to Sym | Type *)
   Sym := Operand(Ptr) ;
   Type := Stack[Ptr].FalseExit ;
   (* adr will be Address type *)
   adr := MakeTemporary(OffsetAddr) ;
   PutVar(adr, Address) ;
   GenQuad(AddrOp, adr, NulSym, Sym) ;
   (* No type for t1 since constant *)
   t1 := MakeTemporary(ImmediateAddr) ;
   GenQuad(OffsetOp, t1, NulSym, Operand(Ptr+1)) ;
   Type := Stack[Ptr+1].FalseExit ;
   IF n>1
   THEN
      FOR i := 2 TO n DO
         (* no type for t2 since constant *)
         t2 := MakeTemporary(ImmediateAddr) ;
         GenQuad(OffsetOp, t2, NulSym, Operand(Ptr+i)) ;
         Type := Stack[Ptr+i].FalseExit ;
         t3 := t1 ;
         (* No type for t1 since constant *)
         t1 := MakeTemporary(ImmediateAddr) ;
         GenQuad(AddOp, t1, t3, t2)
      END
   END ;
   (* Res will be an Address since it is ValueAddr mode *)
   Res := MakeTemporary(ValueAddr) ;
   (*
      Ok must reference by address
      - but we contain the type of the referenced entity
   *)
   PushSize(Address) ;
   PutVarTypeAndSize(Res, Type, PopCard() ) ;
   GenQuad(AddOp, Res, adr, t1) ;
   PushTF(Res, Type)
END BuildDesignatorRecord ;


(*
   BuildDesignatorArray - Builds the array referencing.
                          The compile time stack is unchanged within
                          this procedure.
                          The purpose of this procedure is to work out
                          whether the DesignatorArray is a static or
                          dynamic array.
                          ie. in this compiler terms IsArray or IsUnbounded
                              In which case the appropriate routine is
                              called.
                              However these called routines DO alter
                              the compile time stack.

                          The Stack is expected to contain:


                          Entry                   Exit
                          =====                   ====

                  Ptr ->
                          +--------------+
                          | n            |
                          |--------------|
                          | e1           |
                          |--------------|
                          .              .
                          .              .
                          .              .
                          |--------------|
                          | e2           |                       <- Ptr
                          |--------------|        +------------+
                          | Sym  | Type  |        | S    | T   |
                          |--------------|        |------------|
*)

PROCEDURE BuildDesignatorArray ;
VAR
   IndexPtr,
   Sym,
   n,
   Type    : CARDINAL ;
BEGIN
   PopT(n) ;
   GetPtr(IndexPtr) ;
   DEC(IndexPtr, n+1) ;  (* Ptr points to sym *)
   Sym := GetType(Operand(IndexPtr)) ;
   IF Sym=NulSym
   THEN
      NearToken('type of array is undefined (no such array declared)', GetTokenNo())
   END ;
(*
   ; WriteString('Array is:') ; WriteKey(GetSymName(Operand(IndexPtr))) ; WriteLn ;
*)
   PushT(n) ;       (* Restore stack to origional state *)
   IF IsUnbounded(Sym)
   THEN
      IF n=1
      THEN
         BuildDynamicArray
      ELSE
         WriteError('dynamic array only allowed one indice')
      END
   ELSIF IsArray(Sym)
   THEN
      BuildStaticArray
   ELSE
      WriteError('can only index Static or Dynamic arrays')
   END
END BuildDesignatorArray ;


(*
   BuildStaticArray - Builds the array referencing for static arrays.
                      The Stack is expected to contain:


                      Entry                   Exit
                      =====                   ====

              Ptr ->
                      +--------------+
                      | n            |
                      |--------------|
                      | e1           |
                      |--------------|
                      .              .
                      .              .
                      .              .
                      |--------------|
                      | e2           |                       <- Ptr
                      |--------------|        +------------+
                      | Sym  | Type  |        | S    | T   |
                      |--------------|        |------------|
*)

PROCEDURE BuildStaticArray ;
VAR
   Ptr,
   StartToken,
   Sym,
   i, n,
   OpI,
   Type, Adr,
   Base,
   ti, tj,
   tk, ta: CARDINAL ;
BEGIN
   StartToken := CurrentToken ;
   PutToken(CurrentToken, TokenType) ;
   PopT(n) ;
   GetPtr(Ptr) ;
   DEC(Ptr, n+1) ;  (* Ptr points to sym *)
   Sym := Stack[Ptr].TrueExit ;
   Type := Stack[Ptr].FalseExit ;
   IF NoOfElements(Type)#n
   THEN
      NearTokens('incorrect number of array indices: check [][] with [i, j]',
                 GetDeclared(Type), GetTokenNo())
   END ;
   (* Base has address type since it points to the start of the array in memory *)
   Base := MakeTemporary(OffsetAddr) ;
   PutVar(Base, Address) ;
   GenQuad(BaseOp, Base, Type, Sym) ;
   i := 1 ;
   WHILE i<=n DO
      (* Again ti has no type since constant *)
      ti := MakeTemporary(ImmediateAddr) ;
      GenQuad(ElementSizeOp, ti, Sym, i) ;
      OpI := Operand(Ptr+i) ;
      IF IsConst(OpI)
      THEN
         (* tj has no type since constant *)
         tj := MakeTemporary(ImmediateAddr) ;
         tk := MakeTemporary(ImmediateAddr)
      ELSE
         (* tj has Cardinal type since we have multiplied array indices *)
         (* The problem is that Operand(Ptr+i) might be a CHAR (or any  *)
         (* size < TSIZE(CARDINAL)) so we must coerse.                  *)

         IF GetType(OpI)#Cardinal
         THEN
            PushTF(RequestSym(MakeKey('CONVERT')), NulSym) ;
            PushT(Cardinal) ;
            PushT(OpI) ;
            PushT(2) ;          (* Two parameters *)
            BuildConvertFunction ;
            PopT(OpI)
         END ;
         tj := MakeTemporary(OffsetAddr) ;
         PutVar(tj, MixTypes(GetType(ti), GetType(OpI))) ;
         tk := MakeTemporary(OffsetAddr) ;
         PutVar(tk, MixTypes(GetType(ti), GetType(OpI)))
      END ;
      CheckSubrange(GetNth(Type, i), OpI) ;
      GenQuad(BecomesOp, tj, NulSym, OpI) ;
      GenQuad(MultOp, tk, ti, tj) ;
      PutToken(tk, IdentToken) ;
      INC(i)
   END ;
   (* ta should have type address as it points into memory *)
   ta := MakeTemporary(OffsetAddr) ;
   PutVar(ta, Address) ;
   GetToken ;
   GenQuad(AddOp, ta, MakeConstLit(MakeKey('0')), CurrentToken) ;
   i := 2 ;
   WHILE i<=n DO
      GetToken ;
      GenQuad(AddOp, ta, ta, CurrentToken) ;
      INC(i)
   END ;
   Adr := MakeTemporary(ValueAddr) ;
   (*
      Ok must reference by address
      - but we contain the type of the referenced entity
   *)
   PushSize(Address) ;
   PutVarTypeAndSize(Adr, GetType(Type), PopCard() ) ;

   GenQuad(AddOp, Adr, Base, ta) ;
   PutPtr(Ptr) ;
   PushTF(Adr, GetType(Type)) ;
   GetToken
   ; Assert(StartToken=CurrentToken)
END BuildStaticArray ;


(*
   BuildDynamicArray - Builds the array referencing for dynamic arrays.
                       The Stack is expected to contain:


                       Entry                   Exit
                       =====                   ====

               Ptr ->
                       +--------------+
                       | n            |
                       |--------------|
                       | e1           |
                       |--------------|
                       .              .
                       .              .
                       .              .
                       |--------------|
                       | e2           |                       <- Ptr
                       |--------------|        +------------+
                       | Sym  | Type  |        | S    | T   |
                       |--------------|        |------------|
*)

PROCEDURE BuildDynamicArray ;
VAR
   NewPtr, OldPtr,
   Sym,
   n,
   Type, Adr,
   PtrToBase,
   Base,
   ti, tj, tk    : CARDINAL ;
BEGIN
   PopT(n) ;
   GetPtr(NewPtr) ;
   OldPtr := NewPtr ;
   DEC(NewPtr, n+1) ;  (* NewPtr points to sym *)
   Sym := Stack[NewPtr].TrueExit ;
   Type := Stack[NewPtr].FalseExit ;
   IF NoOfElements(Type)#n
   THEN
      NearTokens('incorrect number of array indices: unbounded arrays have 1 indice',
                 GetDeclared(Type), GetTokenNo())
   END ;
   (*
      Base has type address because
      BuildDesignatorRecord references by address.

      Build a record for retrieving the address of dynamic array.
      BuildDesignatorRecord will generate the required quadruples,
      therefore build set up the stack for BuildDesignatorRecord
      to generate the record access.
   *)
   (*
      Build above current current info needed for array.
      However we do actuclly overwrite, n.
   *)
   PutPtr(OldPtr) ;
   PushTF(Sym, Unbounded) ;
   PushTF(GetLocalSym(Unbounded, ArrayAddress),
          GetType(GetLocalSym(Unbounded, ArrayAddress))) ;
   PushT(1) ;  (* One record field to dereference *)
   BuildDesignatorRecord ;
   PopT(PtrToBase) ;
   (* Now actually copy Unbounded.ArrayAddress into base *)
   Base := MakeTemporary(OffsetAddr) ;
   PutVar(Base, Address) ;           (* has type ADDRESS *)
   GenQuad(CopyOp, Base, NulSym, PtrToBase) ;
   PutPtr(NewPtr) ;  (* Destroy Array info on the stack *)
   IF n#1
   THEN
      (* Only one dimemsion for dynamic arrays *)
      WriteError('dynamic arrays can only have one dimension')
   END ;
   (* ti has no type since constant *)
   ti := MakeTemporary(ImmediateAddr) ;
   GenQuad(ElementSizeOp, ti, Sym, 1) ;
   n := Operand(NewPtr+1) ;
   IF IsConst(n)
   THEN
      (* tj has no type since constant *)
      tj := MakeTemporary(ImmediateAddr) ;
      tk := MakeTemporary(ImmediateAddr)
   ELSE
      (* tj has Cardinal type since we have multiplied array indices *)
      tj := MakeTemporary(OffsetAddr) ;
      (* was:
         PutVar(tj, MixTypes(GetType(ti), GetType(Operand(NewPtr+1)))) ;
         we replace it by: (below). Because we know tj must be a CARDINAL.
         Also mixing types works fine when both are not constants since
         if they are constants then tj is set to NulSym, causing M2EvalSym
         to get upset.

         Although if the index variable is # TSIZE(Cardinal) then BecomesOp
         will rightly complain. Hence we must coerse to a TSIZE(Cardinal) -
         we havent solved type sizes yet so we can only compare against types.
      *)
      IF GetType(n)#Cardinal
      THEN
         PushTF(RequestSym(MakeKey('CONVERT')), NulSym) ;
         PushT(Cardinal) ;
         PushT(n) ;
         PushT(2) ;          (* Two parameters *)
         BuildConvertFunction ;
         PopT(n)
      END ;
      PutVar(tj, Cardinal) ;
      tk := MakeTemporary(OffsetAddr) ;
      PutVar(tk, Cardinal)
   END ;
   CheckDynamicArray(Sym, n) ;
   GenQuad(BecomesOp, tj, NulSym, n) ;
   GenQuad(MultOp, tk, ti, tj) ;
   Adr := MakeTemporary(ValueAddr) ;
   (*
      Ok must reference by address
      - but we contain the type of the referenced entity
   *)
   PushSize(Address) ;
   PutVarTypeAndSize(Adr, GetType(Type), PopCard() ) ;

   GenQuad(AddOp, Adr, Base, tk) ;
   PushTF(Adr, GetType(Type))
END BuildDynamicArray ;


(*
   BuildDesignatorPointer - Builds the record referencing.
                            The Stack is expected to contain:


                            Entry                   Exit
                            =====                   ====

                    Ptr ->                                           <- Ptr
                            +--------------+        +--------------+
                            | Sym1  | Type1|        | Sym2  | Type2|
                            |--------------|        |--------------|
*)

PROCEDURE BuildDesignatorPointer ;
VAR
   Sym1, Type1,
   Sym2, Type2: CARDINAL ;
BEGIN
   PopTF(Sym1, Type1) ;
   IF IsPointer(Type1)
   THEN
      Type2 := GetType(Type1) ;
      Sym2 := MakeTemporary(ValueAddr) ;
      (*
         Ok must reference by address
         - but we contain the type of the referenced entity
      *)
      PushSize(Address) ;
      PutVarTypeAndSize(Sym2, Type2, PopCard() ) ;

      GenQuad(CopyOp, Sym2, NulSym, Sym1) ;
      PushTF(Sym2, Type2)
   ELSE
      WriteErrorFormat1('%s is not a pointer type', GetSymName(Sym1))
   END
END BuildDesignatorPointer ;


(*
   StartBuildWith - performs the with statement.
                    The Stack:

                    Entry                    Exit

                    +------------+
                    | Sym | Type |           Empty
                    |------------|
*)

PROCEDURE StartBuildWith ;
VAR
   Sym, Type,
   adr      : CARDINAL ;
BEGIN
   PopTF(Sym, Type) ;
(* was ...
   (* WriteString('Starting with ') ; WriteKey(GetSymName(Type)) ; WriteLn ; *)
   adr := MakeTemporary(OffsetAddr) ;
   (* adr points to the start of a record therefore give it an Address size *)
   PushSize(Address) ;
   PutVarTypeAndSize(adr, Type, PopCard()) ;
*)
   adr := MakeTemporary(ValueAddr) ;
   PushSize(Address) ;
   PutVarTypeAndSize(adr, Type, PopCard()) ;

   GenQuad(AddrOp, adr, NulSym, Sym) ;
   PushWith(adr, Type) ;
   StartScope(Type)
END StartBuildWith ;


(*
   EndBuildWith - terminates the innermost with scope.
*)

PROCEDURE EndBuildWith ;
BEGIN
   EndScope ;
   PopWith
END EndBuildWith ;


(*
   PushWith - pushes sym and type onto the with stack. It checks for
              previous declaration of this record type.
*)

PROCEDURE PushWith (Sym, Type: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   IF WithPtr>=MaxWithStack
   THEN
      InternalError('MaxWithStack exceeded', __FILE__, __LINE__)
   ELSE
      (*
         actually there is no reason why nested WITH statements
         cannot reference the same record.
         - the outer WITH has presedence.
      *)
      IF Pedantic
      THEN
         i := 0 ;  (* Bottom of the stack *)
         WHILE i<WithPtr DO
            (* Search for other declarations of the with using Type *)
            IF WithStack[i].RecordSym=Type
            THEN
               WriteError('cannot have nested WITH statements referencing the same RECORD')
            END ;
            INC(i)
         END
      END ;
      WITH WithStack[WithPtr] DO
         PtrSym := Sym ;
         RecordSym := Type
      END ;
      INC(WithPtr)
   END
END PushWith ;


PROCEDURE PopWith ;
BEGIN
   IF WithPtr=0
   THEN
      InternalError('with stack underflow', __FILE__, __LINE__)
   ELSE
      DEC(WithPtr)
   END
END PopWith ;


(*
   CheckWithReference - performs the with statement.
                        The Stack:

                        Entry                    Exit

                        +------------+           +------------+
                        | Sym | Type |           | Sym | Type |
                        |------------|           |------------|
*)

PROCEDURE CheckWithReference ;
VAR
   i,
   Sym, Type: CARDINAL ;
   Found    : BOOLEAN ;
BEGIN
   IF WithPtr>0
   THEN
      PopTF(Sym, Type) ;
      i := WithPtr ;   (* Outer WITH always has precidence *)
      Found := FALSE ;
      REPEAT
         DEC(i) ;
         (* WriteString('Checking for a with') ; *)
         WITH WithStack[i] DO
            IF IsRecordField(Sym) AND (Father(Sym)=RecordSym)
            THEN
(*               WriteString(' - found with') ; WriteLn ; *)
               (* Fake a RecordSym.op *)
               PushTF(PtrSym, RecordSym) ;
               PushTF(Sym, Type) ;
               BuildAccessWithField ;
               PopTF(Sym, Type) ;
               Found := TRUE
            END
         END
      UNTIL Found OR (i=0) ;
      PushTF(Sym, Type)
   END
END CheckWithReference ;


(*
   BuildAccessWithField - similar to BuildDesignatorRecord except it
                          does not perform the address operation.
                          The address will have been computed at the
                          beginning of the WITH statement.
                          It also stops the GenQuad procedure from examining the
                          with stack.

                          The Stack

                          Entry

                   Ptr ->
                          +--------------+
                          | Field | Type1|                          <- Ptr
                          |-------|------|          +-------------+
                          | Adr   | Type2|          | Sym  | Type1|
                          |--------------|          |-------------|
*)

PROCEDURE BuildAccessWithField ;
VAR
   OldWithPtr,
   Type1,
   Type2,
   Adr, Res,
   t1, t2  : CARDINAL ;
BEGIN
   OldWithPtr := WithPtr ;
   WithPtr := 0 ;
   (*
      A sneeky trick - now the WITH cannot look at
      the list of outstanding WITH records.
   *)
   DEC(Ptr, 2) ;  (* Ptr points to Adr | Type2 *)
   Adr := Operand(Ptr) ;
   Type2 := Stack[Ptr].FalseExit ;
   t1 := MakeTemporary(ImmediateAddr) ;
   (* No type since t1 is constant *)
   GenQuad(OffsetOp, t1, NulSym, Operand(Ptr+1)) ;
   Type1 := Stack[Ptr+1].FalseExit ;
   Res := MakeTemporary(ValueAddr) ;
   (*
      Ok must reference by address
      - but we contain the type of the referenced entity
   *)
   PushSize(Address) ;
   PutVarTypeAndSize(Res, Type1, PopCard() ) ;
   GenQuad(AddOp, Res, Adr, t1) ;
   PushTF(Res, Type1) ;
   WithPtr := OldWithPtr
   (*
      Now back to previous value.
   *)
END BuildAccessWithField ;


(*
   BuildNulExpression - Builds a nul expression on the stack.
                        The Stack:

                        Entry             Exit

                                                         <- Ptr
                        Empty             +------------+
                                          | NulSym     |
                                          |------------|
*)

PROCEDURE BuildNulExpression ;
BEGIN
   PushT(NulSym)
END BuildNulExpression ;


(*
   BuildBitsetStart - Pushes a Bitset type on the stack.
                      The Stack:

                      Entry             Exit

               Ptr ->                                   <- Ptr

                      Empty             +-------------+
                                        | Bitset      |
                                        |-------------|
*)

PROCEDURE BuildBitsetStart ;
BEGIN
   PushT(Bitset)
END BuildBitsetStart ;


(*
   BuildBitsetEnd - pops the Bitset type from the stack.

                    Entry                   Exit

             Ptr ->
                    +--------------+        Empty
                    | Set Value    |
                    |--------------|
                    | Set Type     |
                    |--------------|
*)

PROCEDURE BuildBitsetEnd ;
VAR
   v, t: CARDINAL ;
BEGIN
   PopT(v) ;
   PopT(t) ;
   PushTF(v, t) ;
   Assert(IsSet(t))
END BuildBitsetEnd ;


(*
   BuildEmptySet - Builds an empty set on the stack.
                   The Stack:

                   Entry             Exit

            Ptr ->                                   <- Ptr
                   +-----------+     +-------------+
                   | SetName   |     | t | SetType |
                   |-----------|     |-------------|

            Quads

            q   t  :=  0
*)

PROCEDURE BuildEmptySet ;
VAR
   Type  : CARDINAL ;
   NulSet: CARDINAL ;
BEGIN
   PopT(Type) ;  (* type of set we are building *)
   Assert(IsSet(Type)) ;
   NulSet := MakeTemporary(ImmediateAddr) ;
   Assert(Type#NulSym) ;
   PutVar(NulSet, Type) ;
   (* Assign with 0 clear all bits *)
   GenQuad(BecomesOp, NulSet, NulSym, MakeConstLit(MakeKey('0'))) ;
   PushTF(Type, Type) ;
   PushTF(NulSet, Type) ;
   IF CompilerDebugging
   THEN
      WriteString('Type = ') ; WriteKey(GetSymName(Type)) ; WriteCard(Type, 4) ; WriteLn ;
      WriteString('Built empty set') ; WriteLn ;
      DisplayStack    (* Debugging info *)
   END
END BuildEmptySet ;


(*
   BuildLogicalOr - performs a logical OR of the top two elements
                    on the stack and then pushes the result.

                    The Stack:

                    Entry             Exit

             Ptr ->              
                    +-----------+
                    | e1  | t1  |                     <- Ptr 
                    |-----------|     +-------------+        
                    | e2  | t2  |     | e3  | t3    |        
                    |-----------|     |-------------|        
            Quads

            q   e3  +  e1  e2
*)

PROCEDURE BuildLogicalOr ;
VAR
   e1, e2   ,
   t1, t2   : CARDINAL ;
   Low, High: CARDINAL ;
BEGIN
   IF CompilerDebugging
   THEN
      WriteString('Build Logical Or') ; WriteLn ;
      DisplayStack   (* Debugging info *)
   END ;
   PopTF(e1, t1) ;
   PopTF(e2, t2) ;   (* t2 contains the set type - we ignore t1 *)
                     (* as it might be an integer ie {1,2,3}    *)
                     (* We push e1 with type t2.                *)
   IF CompilerDebugging
   THEN
      WriteString('e1 = ') ; WriteKey(GetSymName(e1)) ; WriteString('    t1 = ') ; WriteKey(GetSymName(t1)) ; WriteLn ;
      WriteString('e2 = ') ; WriteKey(GetSymName(e2)) ; WriteString('    t2 = ') ; WriteKey(GetSymName(t2)) ; WriteLn
   END ;
   IF t2=NulSym
   THEN
      t2 := Bitset
   END ;
   PushTF(e2, t2) ;
   PushT(LogicalOrTok) ;    (* insist that we perform a logical OR always *)
   PushTF(e1, t2) ;
   BuildBinaryOp ;
   PopTF(e1, t1) ;
   PushTF(e1, t2)
(*  ;    DisplayStack ; *)
END BuildLogicalOr ;


(*
   BuildSetRange - builds a set range.


                          Entry                   Exit
                          =====                   ====


                   Ptr ->                                        <- Ptr
                          +------------+
                          | Element2   |
                          |------------|          +------------+
                          | Element1   |          | BitRange   | 
                          |------------|          |------------|


                   Quadruples Produced

                   q     BitRangeOp  BitRange  Element1 Element2
*)

PROCEDURE BuildSetRange ;
VAR
   BitRange,
   Element1,
   Element2: CARDINAL ;
BEGIN
   PopT(Element2) ;
   PopT(Element1) ;
   BitRange := MakeTemporary(ImmediateAddr) ;
   PutVar(BitRange, Bitset) ;
   GenQuad(BitRangeOp, BitRange, Element1, Element2) ;
   PushTF(BitRange, Bitset)
END BuildSetRange ;


(*
   BuildElement - builds an element.

                          Entry                   Exit
                          =====                   ====


                   Ptr ->                                        <- Ptr
                          +------------+          +------------+
                          | Element    |          | BitNumber  | 
                          |------------|          |------------|
                          | Set|Type   |          | Set|Type   |
                          |------------|          |------------|


                   Quadruples Produced

                   q     BitOp  BitNumber  _  Element
*)

PROCEDURE BuildElement ;
VAR
   High,
   Low,
   BitNumber,
   Set,
   Ignore,
   Type,
   Element  : CARDINAL ;
BEGIN
   IF CompilerDebugging
   THEN
      DisplayStack
   END ;
   PopT(Element) ;
   PopTF(Set, Type) ;
   IF CompilerDebugging
   THEN
      WriteString('Element = ') ; WriteKey(GetSymName(Element)) ; WriteString('    Type = ') ; WriteKey(GetSymName(Type)) ;
      WriteString('       Set     = ') ; WriteKey(GetSymName(Set)) ; WriteLn ;
   END ;
   IF (Type#NulSym) AND IsSet(Type)
   THEN
      IF IsEnumeration(GetType(Type))
      THEN
         (* 
            enumerations start at zero anyway
            and so no need to subtract the base.
            But we will coerse it though.
         *)
      ELSE
         IF IsSubrange(GetType(Type))
         THEN
            GetSubrange(GetType(Type), High, Low) ;   (* must be a subrange type *)
            PushTF(Element, Cardinal) ;  (* must not be a set type since then M2GenCode will perform set difference. *)
            PushT(MinusTok) ; (* Subtract offset from the type ie:       *)
                              (* TYPE  s = SET OF [90..99] ;             *)
            PushTF(Low, NulSym) ;
            BuildBinaryOp ;   (* e = Element - Low *)
            PopTF(Element, Ignore)
         ELSE
            WriteError('type of set is expected to be a subrange type')
         END
      END
   END ;
   BitNumber := MakeTemporary(AreConstant(IsConst(Element))) ;
   PutVar(BitNumber, Bitset) ;
   GenQuad(BitOp, BitNumber, NulSym, Element) ;
   PushTF(Set, Type) ;
   PushTF(BitNumber, Bitset)
END BuildElement ;


(*
   CheckInOp - manipulates the stack so that the IN operator takes
               two bitsets. It converts: number IN bitset to bitset in bitset
               where the first bitset only contains number.


                       Entry                   Exit
                       =====                   ====


               Ptr ->                                         <- Ptr
                       +------------+          +------------+
                       | Set|Type   |          | Set|Type   |
                       |------------|          |------------|
                       | InTok      |          | InTok      |
                       |------------|          |------------|
                       | Element    |          | BitNumber  | 
                       |------------|          |------------|
*)

PROCEDURE CheckInOp ;
VAR
   Op            : CARDINAL ;
   Set, Type1,
   Element, Type2: CARDINAL ;
BEGIN
   PopTF(Set, Type1) ;
   PopT(Op) ;
   PopTF(Element, Type2) ;
   IF Op=InTok
   THEN
      PushTF(Set, Type1) ;
      PushT(Element) ;
      BuildElement ;
      PopTF(Element, Type2) ;
      PopTF(Set, Type1) ;
      Type2 := Type1
   END ;
   (* restore stack *)
   PushTF(Element, Type2) ;
   PushT(Op) ;
   PushTF(Set, Type1)
END CheckInOp ;

   
(*
   RecordOp - Records the operator passed on the stack.
              Checks for AND operator or OR operator
              if either of these operators are found then BackPatching
              takes place.
              The Expected Stack:

              Entry                        Exit

       Ptr ->                                               <- Ptr
              +-------------+               +-------------+
              | OperatorTok |               | OperatorTok |
              |-------------|               |-------------|
              | t    | f    |               | t    | f    |
              |-------------|               |-------------|


              If OperatorTok=AndTok
              Then
                 BackPatch(f, NextQuad)
              Elsif OperatorTok=OrTok
              Then
                 BackPatch(t, NextQuad)
              End
*)

PROCEDURE RecordOp ;
VAR
   Op,
   t, f: CARDINAL ;
BEGIN
   PopT(Op) ;
   IF (Op=AndTok) OR (Op=AmbersandTok)
   THEN
      CheckBooleanId ;
      PopBool(t, f) ;
      BackPatch(t, NextQuad) ;
      PushBool(0, f)
   ELSIF Op=OrTok
   THEN
      CheckBooleanId ;
      PopBool(t, f) ;
      BackPatch(f, NextQuad) ;
      PushBool(t, 0)
   END ;
   PushT(Op)
END RecordOp ;


(*
   CheckForLogicalOperator - returns a logical operator if the operands imply
                             a logical operation should be performed.
*)

PROCEDURE CheckForLogicalOperator (Tok: CARDINAL; e1, t1, e2, t2: CARDINAL) : CARDINAL ;
BEGIN
   IF (Tok=PlusTok) OR (Tok=TimesTok) OR (Tok=MinusTok) OR (Tok=DivTok)
   THEN
      IF (t2#NulSym) AND IsSet(t2)
      THEN
         IF Tok=PlusTok
         THEN
            RETURN( LogicalOrTok )
         ELSIF Tok=MinusTok
         THEN
            RETURN( LogicalDifferenceTok )
         ELSIF Tok=DivTok
         THEN
            RETURN( LogicalXorTok )
         ELSIF Tok=TimesTok
         THEN
            RETURN( LogicalAndTok )
         END
(*
   was
         CASE Tok OF

         PlusTok  : RETURN( LogicalOrTok ) |
         MinusTok : RETURN( LogicalDifferenceTok ) |
         DivTok   : RETURN( LogicalXorTok ) |
         TimesTok : RETURN( LogicalAndTok )

         END
*)
      END
   END ;
   RETURN( Tok )
END CheckForLogicalOperator ;


(*
   BuildBinaryOp   - Builds a binary operation from the quad stack.
                     Be aware that this procedure will check for
                     the overloading of the bitset operators + - \ *.
                     So do NOT call this procedure if you are building
                     a reference to an array which has a bitset type or
                     the address arithmetic will be wrongly coersed into
                     logical ORs.
                     
                     The Stack is expected to contain:


                     Entry                   Exit
                     =====                   ====

              Ptr ->
                     +------------+
                     | Sym1       |
                     |------------|
                     | Operator   |                          <- Ptr
                     |------------|          +------------+
                     | Sym2       |          | Temporary  |
                     |------------|          |------------|


                     Quadruples Produced

                     q     Operator  Temporary  Sym1  Sym2


                OR


                     Entry                   Exit
                     =====                   ====

              Ptr ->
                     +------------+
                     | T1   | F1  |
                     |------------|
                     | OrTok      |                          <- Ptr
                     |------------|          +------------+
                     | T2   | F2  |          | T1+T2| F1  |
                     |------------|          |------------|


                     Quadruples Produced

*)

PROCEDURE BuildBinaryOp ;
VAR
   t1, f1,
   t2, f2,
   e1, e2,
   Tok, t: CARDINAL ;
BEGIN
   Tok := Operand(Ptr-2) ;
   IF Tok=OrTok
   THEN
      CheckBooleanId ;
      PopBool(t1, f1) ;
      PopT(Tok) ;
      PopBool(t2, f2) ;
      Assert(f2=0) ;
      PushBool(Merge(t1, t2), f1)
   ELSIF (Tok=AndTok) OR (Tok=AmbersandTok)
   THEN
      CheckBooleanId ;
      PopBool(t1, f1) ;
      PopT(Tok) ;
      PopBool(t2, f2) ;
      Assert(t2=0) ;
      PushBool(t1, Merge(f1, f2))
   ELSE
      PopTF(e1, t1) ;
      PopT(Tok) ;
      PopTF(e2, t2) ;
      (*
         BinaryOps and UnaryOps only work with immediate and offset addressing.
         which is fine for calculating array and record offsets but we need
         to get the real values to perform normal arithmetic. Not address
         arithmetic.
      *)
      IF GetMode(e1)=ValueAddr
      THEN
         t := MakeTemporary(OffsetAddr) ;
         PutVar(t, t1) ;
         GenQuad(BecomesOp, t, NulSym, e1) ;
         e1 := t
      END ;
      IF GetMode(e2)=ValueAddr
      THEN
         t := MakeTemporary(OffsetAddr) ;
         PutVar(t, t2) ;
         GenQuad(BecomesOp, t, NulSym, e2) ;
         e2 := t
      END ;
      CheckCompatibleTypes(t1, t2) ;
      t := MakeTemporary(AreConstant(IsConst(e1) AND IsConst(e2))) ;
      PutVar(t, MixTypes(t1, t2)) ;  (* this is a trial - can we always put a type for a temporary? *)
      Tok := CheckForLogicalOperator(Tok, e1, t1, e2, t2) ;
      GenQuad(MakeOp(Tok), t, e2, e1) ;
      PushTF(t, GetType(t))
   END
END BuildBinaryOp ;


(*
   BuildUnaryOp   - Builds a unary operation from the quad stack.
                    The Stack is expected to contain:


                    Entry                   Exit
                    =====                   ====

             Ptr ->
                    +------------+
                    | Sym        |
                    |------------|          +------------+
                    | Operator   |          | Temporary  |  <- Ptr
                    |------------|          |------------|


                    Quadruples Produced

                    q     Operator  Temporary  _ Sym

*)

PROCEDURE BuildUnaryOp ;
VAR
   Sym, Tok,
   SymT,  t: CARDINAL ;
BEGIN
   PopT(Sym) ;
   PopT(Tok) ;
   IF Tok=MinusTok
   THEN
      t := MakeTemporary(AreConstant(IsConst(Sym))) ;
      IF NOT IsConst(Sym)
      THEN
         (* ok t needs space *)
         PutVar(t, MixTypes(GetType(Sym), NulSym)) ;
         IF GetMode(Sym)=ValueAddr
         THEN
            SymT := MakeTemporary(OffsetAddr) ;
            PutVar(SymT, GetType(Sym)) ;
            GenQuad(BecomesOp, SymT, NulSym, Sym) ;
            Sym := SymT
         END
      END ;
      GenQuad(NegateOp, t, NulSym, Sym) ;
      PushT(t)
   ELSIF Tok=PlusTok
   THEN
      PushT(Sym)
   ELSE
      WriteErrorFormat1('not expecting this kind of unary operator (%s)', Tok)
   END
END BuildUnaryOp ;


(*
   AreConstant - returns immediate addressing mode if b is true else
                 offset mode is returned. b determines whether the
                 operands are all constant - in which case we can use
                 a constant temporary variable.
*)

PROCEDURE AreConstant (b: BOOLEAN) : ModeOfAddr ;
BEGIN
   IF b
   THEN
      RETURN( ImmediateAddr )
   ELSE
      RETURN( OffsetAddr )
   END
END AreConstant ;


(*
   ConvertBooleanToVariable - carefully converts Stack[i] from a Boolean True|False
                              exit pair into a variable containing the value TRUE or
                              FALSE.
*)

PROCEDURE ConvertBooleanToVariable (i: CARDINAL) ;
VAR
   Des: CARDINAL ;
BEGIN
   Assert(IsBoolean(i)) ;
   (*
      need to convert it to a variable containing the result.
      Des will be a boolean type
   *)
   Des := MakeTemporary(OffsetAddr) ;
   PutVar(Des, Boolean) ;
   PushT(Des) ;
   PushBool(Stack[i].TrueExit, Stack[i].FalseExit) ;
   BuildAssignmentWithoutBounds ;
   WITH Stack[i] DO
      TrueExit := Des ;  (* alter Stack[i] to contain the variable *)
      FalseExit := Boolean ;
      BooleanOp := FALSE (* no longer a Boolean True|False pair    *)
   END
END ConvertBooleanToVariable ;


(*
   BuildRelOpFromBoolean - builds a relational operator sequence of quadruples
                           instead of using a temporary boolean variable.
                           This function can only be used when we perform
                           the following translation:

                           (a=b) # (c=d)  alternatively   (a=b) = (c=d)
                                 ^                              ^

                           it only allows # = to be used as >= <= > < all
                           assume a particular value for TRUE and FALSE.
                           (In which case the user should specify ORD)


                           before

                           q    if r1      op1     op2     t2
                           q+1  Goto                       f2
                           q+2  if r2      op3     op4     t1
                           q+3  Goto                       f1

                           after (in case of =)

                           q    if r1      op1     op2     q+2
                           q+1  Goto                       q+4
                           q+2  if r2      op3     op4     t
                           q+3  Goto                       f
                           q+4  if r2      op3     op4     f
                           q+5  Goto                       t

                           after (in case of #)

                           q    if r1      op1     op2     q+2
                           q+1  Goto                       q+4
                           q+2  if r2      op3     op4     f
                           q+3  Goto                       t
                           q+4  if r2      op3     op4     t
                           q+5  Goto                       f

                           The Stack is expected to contain:


                           Entry                   Exit
                           =====                   ====

                    Ptr ->
                           +------------+
                           | t1 | f1    |
                           |------------|
                           | Operator   |                          <- Ptr
                           |------------|          +------------+
                           | t2 | f2    |          | t    | f   |
                           |------------|          |------------|


*)

PROCEDURE BuildRelOpFromBoolean ;
VAR
   Tok,
   t1, f1,
   t2, f2: CARDINAL ;
BEGIN
   Assert(IsBoolean(Ptr-1) AND IsBoolean(Ptr-3)) ;
   IF Operand(Ptr-2)=EqualTok
   THEN
      (* are the two boolean expressions the same? *)
      PopBool(t1, f1) ;
      PopT(Tok) ;
      PopBool(t2, f2) ;
      (* give the false exit a second chance *)
      BackPatch(t2, t1) ;        (* q    if   _     _    q+2 *)
      BackPatch(f2, NextQuad) ;  (* q+1  if   _     _    q+4 *)
      Assert(NextQuad=f1+1) ;
      WITH Quads[t1] DO
         GenQuad(Operator, Operand1, Operand2, 0)
      END ;
      GenQuad(GotoOp, NulSym, NulSym, 0) ;
      PushBool(Merge(NextQuad-1, t1), Merge(NextQuad-2, f1))
   ELSIF Operand(Ptr-2)=HashTok
   THEN
      (* are the two boolean expressions the different? *)
      PopBool(t1, f1) ;
      PopT(Tok) ;
      PopBool(t2, f2) ;
      (* give the false exit a second chance *)
      BackPatch(t2, t1) ;        (* q    if   _     _    q+2 *)
      BackPatch(f2, NextQuad) ;  (* q+1  if   _     _    q+4 *)
      Assert(NextQuad=f1+1) ;
      WITH Quads[t1] DO
         GenQuad(Operator, Operand1, Operand2, 0)
      END ;
      GenQuad(GotoOp, NulSym, NulSym, 0) ;
      PushBool(Merge(NextQuad-2, f1), Merge(NextQuad-1, t1))
   ELSE
      WriteError('only allowed to use relation operators = # on BOOLEAN expressions as these do not imply a value for TRUE or FALSE')
   END
END BuildRelOpFromBoolean ;


(*
   BuildRelOp   - Builds a relative operation from the quad stack.
                  The Stack is expected to contain:


                  Entry                   Exit
                  =====                   ====

           Ptr ->
                  +------------+
                  | e1         |
                  |------------|                          <- Ptr
                  | Operator   |
                  |------------|          +------------+
                  | e2         |          | t    | f   |
                  |------------|          |------------|


                    Quadruples Produced

                    q     IFOperator  e2  e1  TrueExit    ; e2  e1 since
                    q+1   GotoOp              FalseExit   ; relation > etc
                                                          ; requires order.
*)

PROCEDURE BuildRelOp ;
VAR
   Op,
   t1, t2,
   e1, e2: CARDINAL ;
BEGIN
   IF IsBoolean(Ptr-1) AND IsBoolean(Ptr-3)
   THEN
      (*
         we allow # and = to be used with Boolean expressions.
         we do not allow >  <  >=  <=  though
      *)
      BuildRelOpFromBoolean
   ELSE
      IF IsBoolean(Ptr-1)
      THEN
         ConvertBooleanToVariable(Ptr-1)
      END ;
      IF IsBoolean(Ptr-3)
      THEN
         ConvertBooleanToVariable(Ptr-3)
      END ;
      CheckInOp ;
      PopTF(e1, t1) ;
      PopT(Op) ;
      PopTF(e2, t2) ;
      CheckCompatibleTypes(t1, t2) ;
      GenQuad(MakeOp(Op), e2, e1, 0) ;      (* True  Exit *)
      GenQuad(GotoOp, NulSym, NulSym, 0) ;  (* False Exit *)
      PushBool(NextQuad-2, NextQuad-1)
   END
END BuildRelOp ;


(*
   BuildNot   - Builds a NOT operation from the quad stack.
                The Stack is expected to contain:


                  Entry                   Exit
                  =====                   ====

           Ptr ->                                        <- Ptr
                  +------------+          +------------+
                  | t    | f   |          | f    | t   |
                  |------------|          |------------|
*)

PROCEDURE BuildNot ;
VAR
   t, f: CARDINAL ;
BEGIN
   CheckBooleanId ;
   PopBool(t, f) ;
   PushBool(f, t)
END BuildNot ;


(*
   MakeOp - returns the equalent quadruple operator to a token, t.
*)

PROCEDURE MakeOp (t: CARDINAL) : QuadOperator ;
BEGIN
   IF t=PlusTok
   THEN
      RETURN( AddOp )
   ELSIF t=MinusTok
   THEN
      RETURN( SubOp )
   ELSIF (t=DivTok) OR (t=DivideTok)
   THEN
      RETURN( DivOp )
   ELSIF t=TimesTok
   THEN
      RETURN( MultOp )
   ELSIF t=ModTok
   THEN
      RETURN( ModOp )
   ELSIF t=HashTok
   THEN
      RETURN( IfNotEquOp )
   ELSIF t=LessGreaterTok
   THEN
      RETURN( IfNotEquOp )
   ELSIF t=GreaterEqualTok
   THEN
      RETURN( IfGreEquOp )
   ELSIF t=LessEqualTok
   THEN
      RETURN( IfLessEquOp )
   ELSIF t=EqualTok
   THEN
      RETURN( IfEquOp )
   ELSIF t=LessTok
   THEN
      RETURN( IfLessOp )
   ELSIF t=GreaterTok
   THEN
      RETURN( IfGreOp )
   ELSIF t=InTok
   THEN
      RETURN( IfInOp )
   ELSIF t=LogicalOrTok
   THEN
      RETURN( LogicalOrOp )
   ELSIF t=LogicalAndTok
   THEN
      RETURN( LogicalAndOp )
   ELSIF t=LogicalXorTok
   THEN
      RETURN( LogicalXorOp )
   ELSIF t=LogicalDifferenceTok
   THEN
      RETURN( LogicalDifferenceOp )
   ELSE
      InternalError('binary operation not implemented yet', __FILE__, __LINE__)
   END
END MakeOp ;


(*
   GenQuad - Generate a quadruple with Operation, Op1, Op2, Op3.
*)

PROCEDURE GenQuad (Operation: QuadOperator;
                   Op1, Op2, Op3: CARDINAL) ;
BEGIN
   (* WriteString('Potential Quad: ') ; *)
   IF NextQuad#Head
   THEN
      WITH Quads[NextQuad-1] DO
         Next := NextQuad
      END
   END ;
   PutQuad(NextQuad, Operation, Op1, Op2, Op3) ;
   WITH Quads[NextQuad] DO
      Next := 0 ;
      LineNo := GetLineNo() ;
      TokenNo := GetTokenNo()
   END ;
   (* DisplayQuad(NextQuad) ; *)
   NewQuad(NextQuad)
END GenQuad ;


(*
   DisplayQuadList - Displays all quads in list Head.
*)

PROCEDURE DisplayQuadList (Head: CARDINAL) ;
BEGIN
   WriteString('Quadruples:') ; WriteLn ;
   WHILE Head#0 DO
      DisplayQuad(Head) ;
      Head := Quads[Head].Next
   END
END DisplayQuadList ;


(*
   BackPatch - Makes each of the quadruples on the list pointed to by
               StartQuad, take quadruple Value as a target.
*)

PROCEDURE BackPatch (QuadNo, Value: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   WHILE QuadNo#0 DO
      WITH Quads[QuadNo] DO
         i := Operand3 ;                       (* Next Link along the BackPatch *)
         ManipulateReference(QuadNo, Value)    (* Filling in the BackPatch.     *)
      END ;
      QuadNo := i
   END
 (* ; DisplayQuadList(1) ;  (* Debugging only *) *)
END BackPatch ;


(*
   Merge - joins two quad lists, QuadList2 to the end of QuadList1.
           A QuadList of value zero is a nul list.
*)

PROCEDURE Merge (QuadList1, QuadList2: CARDINAL) : CARDINAL ;
VAR
   i, j: CARDINAL ;
BEGIN
   IF QuadList1=0
   THEN
      RETURN( QuadList2 )
   ELSIF QuadList2=0
   THEN
      RETURN( QuadList1 )
   ELSE
      i := QuadList1 ;
      REPEAT
         j := i ;
         i := Quads[i].Operand3
      UNTIL i=0 ;
      ManipulateReference(j, QuadList2) ;
      RETURN( QuadList1 )
   END
END Merge ;


(*
   DisplayStack - displays the compile time symbol stack.
*)

PROCEDURE DisplayStack ;
VAR
   i: CARDINAL ;
BEGIN
   IF Ptr>0
   THEN
      WriteString('+-------------------+') ; WriteLn ;
      FOR i := Ptr-1 TO 0 BY -1 DO
         WITH Stack[i] DO
            IF BooleanOp
            THEN
               WriteString('| Q') ; WriteCard(TrueExit, 6) ;
               WriteString(' | Q') ; WriteCard(FalseExit, 6) ;
               WriteString(' |') ; WriteLn ;
               WriteString('|---------|---------|') ; WriteLn ;
            ELSE
               WriteString('| ') ; WriteCard(TrueExit, 6) ;
               WriteString('            |') ; WriteLn ;
               WriteString('|-------------------|') ; WriteLn
            END
         END
      END
   END
END DisplayStack ;


(*
   DisplayQuad - displays a quadruple, QuadNo.
*)

PROCEDURE DisplayQuad (QuadNo: CARDINAL) ;
BEGIN
   WriteCard(QuadNo, 4) ;
   WriteString('  ') ;
   WriteQuad(QuadNo) ;
   WriteLn
END DisplayQuad ;


(*
   WriteQuad - Writes out the Quad BufferQuad.
*)

PROCEDURE WriteQuad (BufferQuad: CARDINAL) ;
BEGIN
   WITH Quads[BufferQuad] DO
      WriteOperator(Operator) ;
      WriteString('  [') ;
      WriteCard(NoOfTimesReferenced, 0) ;
      WriteString(']    ') ;
      CASE Operator OF

      BitOp,
      UnboundedVarOp,
      UnboundedOp,
      HighOp,
      ReturnValueOp,
      FunctValueOp,
      OffsetOp,
      NegateOp,
      BecomesOp,
      CopyOp,
      SizeOp,
      AddrOp            : WriteOperand(Operand1) ;
                          WriteString('  ') ;
                          WriteOperand(Operand3) |
      ElementSizeOp,
      IfInOp,
      IfNotInOp,
      IfNotEquOp,
      IfEquOp,
      IfLessOp,
      IfGreOp,
      IfLessEquOp,
      IfGreEquOp        : WriteOperand(Operand1) ;
                          WriteString('  ') ;
                          WriteOperand(Operand2) ;
                          WriteString('  ') ;
                          WriteCard(Operand3, 4) |

      InlineOp,
      LineNumberOp,
      GotoOp            : WriteCard(Operand3, 4) |

      EndFileOp         : WriteKey(GetSymName(Operand3)) |

      ReturnOp,
      CallOp,
      KillLocalVarOp    : WriteOperand(Operand3) |

      EndOp,
      StartOp,
      StartDefFileOp,
      StartModFileOp,
      NewLocalVarOp     : WriteCard(Operand1, 4) ;
                          WriteString('  ') ;
                          WriteKey(GetSymName(Operand3)) |

      ParamOp           : WriteCard(Operand1, 4) ;
                          WriteString('  ') ;
                          WriteOperand(Operand2) ;
                          WriteString('  ') ;
                          WriteOperand(Operand3) |

      BaseOp,
      MathOp,
      LogicalOrOp,
      LogicalAndOp,
      LogicalXorOp,
      LogicalDifferenceOp,
      BitRangeOp,
      CoerceOp,
      ConvertOp,
      AddOp,
      SubOp,
      MultOp,
      ModOp,
      DivOp             : WriteOperand(Operand1) ;
                          WriteString('  ') ;
                          WriteOperand(Operand2) ;
                          WriteString('  ') ;
                          WriteOperand(Operand3) |
      DummyOp,
      CodeOnOp,
      CodeOffOp,
      ProfileOnOp,
      ProfileOffOp,
      OptimizeOnOp,
      OptimizeOffOp     :


      ELSE
         InternalError('quadruple not recognised', __FILE__, __LINE__)
      END
   END
END WriteQuad ;


(*
   WriteOperator - writes the name of the quadruple operator.
*)

PROCEDURE WriteOperator (Operator: QuadOperator) ;
BEGIN
   CASE Operator OF

   MathOp             : WriteString('MathOp       ') |
   LogicalOrOp        : WriteString('Or           ') |
   LogicalAndOp       : WriteString('And          ') |
   LogicalXorOp       : WriteString('Xor          ') |
   LogicalDifferenceOp: WriteString('Diff         ') |
   BitOp              : WriteString('Bit          ') |
   BitRangeOp         : WriteString('BitRange     ') |
   BecomesOp          : WriteString('Becomes      ') |
   CopyOp             : WriteString('Copy         ') |
   BaseOp             : WriteString('Base         ') |
   ElementSizeOp      : WriteString('ElementSize  ') |
   AddrOp             : WriteString('Addr         ') |
   SizeOp             : WriteString('Size         ') |
   OffsetOp           : WriteString('Offset       ') |
   IfInOp             : WriteString('If IN        ') |
   IfNotInOp          : WriteString('If NOT IN    ') |
   IfNotEquOp         : WriteString('If <>        ') |
   IfEquOp            : WriteString('If =         ') |
   IfLessEquOp        : WriteString('If <=        ') |
   IfGreEquOp         : WriteString('If >=        ') |
   IfGreOp            : WriteString('If >         ') |
   IfLessOp           : WriteString('If <         ') |
   GotoOp             : WriteString('Goto         ') |
   DummyOp            : WriteString('Dummy        ') |
   StartDefFileOp     : WriteString('StartDefFile ') |
   StartModFileOp     : WriteString('StartModFile ') |
   EndFileOp          : WriteString('EndFileOp    ') |
   StartOp            : WriteString('Start        ') |
   EndOp              : WriteString('End          ') |
   AddOp              : WriteString('+            ') |
   SubOp              : WriteString('-            ') |
   DivOp              : WriteString('DIV          ') |
   ModOp              : WriteString('MOD          ') |
   MultOp             : WriteString('*            ') |
   NegateOp           : WriteString('Negate       ') |
   ReturnOp           : WriteString('Return       ') |
   ReturnValueOp      : WriteString('ReturnValue  ') |
   FunctValueOp       : WriteString('FunctValue   ') |
   CallOp             : WriteString('Call         ') |
   ParamOp            : WriteString('Param        ') |
   NewLocalVarOp      : WriteString('NewLocalVar  ') |
   KillLocalVarOp     : WriteString('KillLocalVar ') |
   UnboundedOp        : WriteString('Unbounded    ') |
   UnboundedVarOp     : WriteString('UnboundedVar ') |
   CoerceOp           : WriteString('Coerce       ') |
   ConvertOp          : WriteString('Convert      ') |
   HighOp             : WriteString('High         ') |
   CodeOnOp           : WriteString('CodeOn       ') |
   CodeOffOp          : WriteString('CodeOff      ') |
   ProfileOnOp        : WriteString('ProfileOn    ') |
   ProfileOffOp       : WriteString('ProfileOff   ') |
   OptimizeOnOp       : WriteString('OptimizeOn   ') |
   OptimizeOffOp      : WriteString('OptimizeOff  ') |
   InlineOp           : WriteString('Inline       ') |
   LineNumberOp       : WriteString('LineNumber   ')

   ELSE
      InternalError('operator not expected', __FILE__, __LINE__)
   END
END WriteOperator ;


PROCEDURE WriteOperand (Sym: CARDINAL) ;
BEGIN
   WriteKey(GetSymName(Sym)) ;
   Write('[') ; WriteMode(GetMode(Sym)) ; Write(']') ;
   Write('(') ; WriteCard(Sym, 0) ; Write(')')
END WriteOperand ;


PROCEDURE WriteMode (Mode: ModeOfAddr) ;
BEGIN
   CASE Mode OF

   ImmediateAddr: Write('i') |
   NoAddr       : Write('n') |
   OffsetAddr   : Write('o') |
   ValueAddr    : Write('v')

   ELSE
      InternalError('unrecognised mode', __FILE__, __LINE__)
   END
END WriteMode ;


(*
   PushExit - pushes the exit value onto the EXIT stack.
*)

PROCEDURE PushExit (Exit: CARDINAL) ;
BEGIN
   IF ExitPtr>MaxExitStack
   THEN
      InternalError('MaxExitStack exceeded', __FILE__, __LINE__) ;
   ELSE
      ExitStack[ExitPtr] := Exit
   END ;
   INC(ExitPtr)
END PushExit ;


(*
   PopExit - pops the exit value from the EXIT stack.
*)

PROCEDURE PopExit() : CARDINAL ;
BEGIN
   IF ExitPtr>0
   THEN
      DEC(ExitPtr)
   ELSE
      InternalError('exit stack underflow', __FILE__, __LINE__)
   END ;
   RETURN( ExitStack[ExitPtr] )
END PopExit ;


(*
   GetPtr - sets, p, to the value of the quadruple stack pointer.
*)

PROCEDURE GetPtr (VAR p: CARDINAL) ;
BEGIN
   p := Ptr
END GetPtr ;


(*
   PutPtr - sets the quadruple stack pointer to p.
*)

PROCEDURE PutPtr (p: CARDINAL) ;
BEGIN
   Ptr := p
END PutPtr ;


(*
   PopBool - Pops a True and a False exit quad number from the True/False
             stack.
*)

PROCEDURE PopBool (VAR True, False: CARDINAL) ;
BEGIN
   IF Ptr>0
   THEN
      DEC(Ptr)
   ELSE
      InternalError('stack underflow', __FILE__, __LINE__)
   END ;
   WITH Stack[Ptr] DO
      True := TrueExit ;
      False := FalseExit ;
      Assert(BooleanOp)
   END
END PopBool ;


(*
   PushBool - Push a True and a False exit quad numbers onto the
              True/False stack.
*)

PROCEDURE PushBool (True, False: CARDINAL) ;
BEGIN
   IF Ptr>MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         TrueExit := True ;
         FalseExit := False ;
         BooleanOp := TRUE
      END
   END ;
   INC(Ptr) ;
END PushBool ;


(*
   IsBoolean - Returns true is the Stack position Pos contains a Boolean
               Exit. False is returned if an Ident is stored.
*)

PROCEDURE IsBoolean (Pos: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(Stack[Pos].BooleanOp)
END IsBoolean ;


(*
   Operand - Returns the operand value from the Stack at position Pos.
*)

PROCEDURE Operand (Pos: CARDINAL) : CARDINAL ;
BEGIN
   Assert(NOT IsBoolean(Pos)) ;
   RETURN( Stack[Pos].TrueExit )
END Operand ;


(*
   BuildCodeOn - generates a quadruple declaring that code should be
                 emmitted from henceforth.

                 The Stack is unnaffected.
*)

PROCEDURE BuildCodeOn ;
BEGIN
   GenQuad(CodeOnOp, NulSym, NulSym, NulSym)
END BuildCodeOn ;


(*
   BuildCodeOff - generates a quadruple declaring that code should not be
                  emmitted from henceforth.

                  The Stack is unnaffected.
*)

PROCEDURE BuildCodeOff ;
BEGIN
   GenQuad(CodeOffOp, NulSym, NulSym, NulSym)
END BuildCodeOff ;


(*
   BuildProfileOn - generates a quadruple declaring that profile timings
                    should be emmitted from henceforth.

                    The Stack is unnaffected.
*)

PROCEDURE BuildProfileOn ;
BEGIN
   GenQuad(ProfileOnOp, NulSym, NulSym, NulSym)
END BuildProfileOn ;


(*
   BuildProfileOn - generates a quadruple declaring that profile timings
                    should be emmitted from henceforth.

                    The Stack is unnaffected.
*)

PROCEDURE BuildProfileOff ;
BEGIN
   GenQuad(ProfileOffOp, NulSym, NulSym, NulSym)
END BuildProfileOff ;


(*
   BuildOptimizeOn - generates a quadruple declaring that optimization
                     should occur from henceforth.

                     The Stack is unnaffected.
*)

PROCEDURE BuildOptimizeOn ;
BEGIN
   GenQuad(OptimizeOnOp, NulSym, NulSym, NulSym)
END BuildOptimizeOn ;


(*
   BuildOptimizeOff - generates a quadruple declaring that optimization
                      should not occur from henceforth.

                      The Stack is unnaffected.
*)

PROCEDURE BuildOptimizeOff ;
BEGIN
   GenQuad(OptimizeOffOp, NulSym, NulSym, NulSym)
END BuildOptimizeOff ;


(*
   BuildInline - builds an Inline buffer pseudo quadruple operator.
                 The inline buffer, Buf, is stored as the operand
                 to the operator InlineOp.

                 The stack is expected to contain:


                        Entry                   Exit
                        =====                   ====
  
                 Ptr ->
                        +--------------+
                        | Buf          |        Empty
                        |--------------|
*)

PROCEDURE BuildInline ;
VAR
   Buf: CARDINAL ;
BEGIN
   PopT(Buf) ;
   GenQuad(InlineOp, NulSym, NulSym, Buf)
END BuildInline ;


(*
   BuildLineNo - builds a LineNumberOp pseudo quadruple operator.
                 This quadruple indicates which source line has been
                 processed, these quadruples are only generated if we
                 are producing runtime debugging information.

                 The stack is not affected, read or altered in any way.


                        Entry                   Exit
                        =====                   ====
  
                 Ptr ->                              <- Ptr
*)

PROCEDURE BuildLineNo ;
BEGIN
   IF (NextQuad#Head) AND (GenerateLineDebug OR GenerateDebugging)
   THEN
(*
      GenQuad(LineNumberOp, NulSym, NulSym, GetLineNo())
*)
      IF Quads[NextQuad-1].Operator=LineNumberOp
      THEN
         PutQuad(NextQuad-1, LineNumberOp, NulSym, NulSym, GetLineNo())
      ELSE
         GenQuad(LineNumberOp, NulSym, NulSym, GetLineNo())
      END
   END
END BuildLineNo ;


(*
   NoOfDynamicQuads - returns the number of quadruples which have been
                      used by the dynamic code generator.
*)

PROCEDURE NoOfDynamicQuads () : CARDINAL ;
BEGIN
   RETURN( NoOfDynamic )
END NoOfDynamicQuads ;


(*
   IncDynamicQuads - increments the number of dynamic quadruple count
                     by Start..End quadruples.
*)

PROCEDURE IncDynamicQuads (Start, End: CARDINAL) ;
BEGIN
   INC(NoOfDynamic) ;
   WHILE Start#End DO
      Start := GetNextQuad(Start) ;
      INC(NoOfDynamic)
   END
END IncDynamicQuads ;


(*
   PushTF - Push a True and False numbers onto the True/False stack.
            True and False are assumed to contain Symbols or Ident etc.
*)

PROCEDURE PushTF (True, False: CARDINAL) ;
BEGIN
   IF Ptr>MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         TrueExit := True ;
         FalseExit := False ;
         BooleanOp := FALSE
      END
   END ;
   INC(Ptr) ;
END PushTF ;


(*
   PopTF - Pop a True and False number from the True/False stack.
           True and False are assumed to contain Symbols or Ident etc.
*)

PROCEDURE PopTF (VAR True, False: CARDINAL) ;
BEGIN
   IF Ptr>0
   THEN
      DEC(Ptr)
   ELSE
      InternalError('stack underflow', __FILE__, __LINE__)
   END ;
   WITH Stack[Ptr] DO
      True := TrueExit ;
      False := FalseExit ;
      Assert(NOT BooleanOp)
   END
END PopTF ;


(*
   PushT - Push a True number onto the
           True/False stack. The False value will be zero.
*)

PROCEDURE PushT (True: CARDINAL) ;
BEGIN
   IF Ptr>MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         TrueExit := True ;
         FalseExit := 0 ;
         BooleanOp := FALSE
      END ;
      INC(Ptr)
   END
END PushT ;


(*
   PopT - Pops a True number from the True/False
          stack. The False value is ignored.
*)

PROCEDURE PopT (VAR True: CARDINAL) ;
BEGIN
   IF Ptr>0
   THEN
      DEC(Ptr)
   ELSE
      InternalError('stack underflow', __FILE__, __LINE__)
   END ;
   WITH Stack[Ptr] DO
      True := TrueExit ;
      Assert(NOT BooleanOp)
   END ;
END PopT ;



(*
   PushInit - pushes a quadruple number onto a stack size of one!
*)

PROCEDURE PushInit (q: CARDINAL) ;
BEGIN
   IF InitQuad=0
   THEN
      q := InitQuad
   ELSE
      InternalError('InitQuad should be zero', __FILE__, __LINE__)
   END
END PushInit ;


(*
   PopInit - pops a quadruple number from a stack size of one!
*)

PROCEDURE PopInit (VAR q: CARDINAL) ;
BEGIN
   IF InitQuad#0
   THEN
      q := InitQuad
   ELSE
      InternalError('InitQuad should not be zero - problem with module initialization', __FILE__, __LINE__)
   END
END PopInit ;


(*
   Init - initialize the M2Quads module, quadruple stack, with stack
          and quadruples.
*)

PROCEDURE Init ;
BEGIN
   LogicalOrTok := MakeKey('_LOR') ;
   LogicalAndTok := MakeKey('_LAND') ;
   LogicalDifferenceTok := MakeKey('_LDIFF') ;
   LogicalXorTok := MakeKey('_LXOR') ;
   FreeList := 1 ;
   NoOfDynamic := 0 ;
   NewQuad(NextQuad) ;
   Assert(NextQuad=1) ;
   Ptr := 0 ;
   ExitPtr := 0 ;
   Head := 1 ;
   ReturnQuad := 0 ;
   WithPtr := 0 ;
   LastQuadNo := 0 ;
   MustNotCheckBounds := FALSE ;
   InitQuad := 0 ;
   GrowInitialization := 0
END Init ;


BEGIN
   (* WriteString('Initialization of M2Quads') ; WriteLn ; *)
   Init
END M2Quads.
